package nl.knoppel.tools;

/**
 * Utility class to perform mathematical operations
 * 
 * @author Franois L.A. Knoppel -- francois@knoppel.nl
 * 
 */
public class MathTool {

	public static final double INFINITESIMAL_DOUBLE = 1E-99;
	public static final float INFINITESIMAL_FLOAT = 1E-15f;

	/**
	 * Returns the euclidean distance between two points.
	 * 
	 * @param x1
	 * @param y1
	 * @param x2
	 * @param y2
	 * @return
	 */
	public static double distance(float x1, float y1, float x2, float y2) {
		double distance;

		float x_dist = x2 - x1;
		float y_dist = y2 - y1;

		double x_dist_squared = x_dist * x_dist;
		double y_dist_squared = y_dist * y_dist;

		distance = Math.sqrt(x_dist_squared + y_dist_squared);
		return distance;
	}

	/**
	 * Returns the gaussian of x with mean 0 and standard deviation 1
	 * 
	 * @param x
	 * @return
	 */
	public static double gaussian(double x) {
		return Math.exp(-x * x / 2) / Math.sqrt(2 * Math.PI);
	}

	/**
	 * Returns the gaussian of x with mean mu and standard deviation sigma
	 * 
	 * @param x
	 * @param mu
	 * @param sigma
	 * @return
	 */
	public static double gaussian(double x, double mu, double sigma) {
		return gaussian((x - mu) / sigma) / sigma;
	}

	/**
	 * Normalizes an array with newValue = oldValue / totalInputArrayValue.
	 * 
	 * @param inputArray
	 * @return
	 */
	public static double[] normalizeArray(double[] inputArray) {
		double[] returnArray = new double[inputArray.length];

		double total = 0;
		for (double currentvalue : inputArray) {
			total += currentvalue;
		}

		for (int i = 0; i < inputArray.length; i++) {
			returnArray[i] = inputArray[i] / total;
		}

		return returnArray;
	}

	/**
	 * Returns the clockwise angle from north to the line specified by the two
	 * coordinates.
	 * 
	 * @param centerX
	 * @param centerY
	 * @param centerX2
	 * @param centerY2
	 * @return
	 */
	public static double getAngle(float centerX, float centerY, float centerX2,
			float centerY2) {
		float x_dist = centerX2 - centerX + INFINITESIMAL_FLOAT;
		float y_dist = centerY2 - centerY + INFINITESIMAL_FLOAT;

		y_dist = -y_dist;

		double x_dist_squared = x_dist * x_dist;
		double y_dist_squared = y_dist * y_dist;
		double dist = Math.sqrt(x_dist_squared + y_dist_squared);
		double factor_x = x_dist / dist;
		double factor_y = y_dist / dist;

		double angle_radians = Math.atan(factor_y / factor_x);

		float signumX = Math.signum(x_dist);

		angle_radians = 0.5 * Math.PI - angle_radians;
		if (signumX == -1) {
			angle_radians += Math.PI;
		}
		return angle_radians;
	}

	/**
	 * Given the line specified by the coordinates, returns {x,y} with the
	 * euclidean distance of (0,0) --> (x,y) being 1
	 * 
	 * @param from_center_x
	 * @param from_center_y
	 * @param to_center_x
	 * @param to_center_y
	 * @return
	 */
	public static double[] calculateStandardizedVector(int from_center_x,
			int from_center_y, int to_center_x, int to_center_y) {
		int x_dist = from_center_x - to_center_x;
		int y_dist = from_center_y - to_center_y;

		double x_dist_squared = Math.pow(x_dist, 2);
		double y_dist_squared = Math.pow(y_dist, 2);
		double dist = Math.sqrt(x_dist_squared + y_dist_squared);

		double[] factors = new double[2];
		if (dist == 0) {
			factors[0] = 0;
			factors[1] = 1;
		} else {

			double factor_x = x_dist / dist;
			double factor_y = y_dist / dist;
			factors[0] = factor_x;
			factors[1] = factor_y;
		}
		return factors;
	}

	/**
	 * Converts a hexadecimal value to a decimal value.
	 * 
	 * @param hexadecimal
	 * @return
	 */
	public static int convertHexadecimalToDecimal(String hexadecimal) {
		String hexStr = "0123456789ABCDEF";
		int resultValue = 0;

		String transformedHexadecimal = hexadecimal;
		int position = 0;
		while (transformedHexadecimal.length() != 0) {
			int lengthMinusOne = transformedHexadecimal.length() - 1;
			resultValue += hexStr.indexOf(transformedHexadecimal.substring(
					lengthMinusOne).toUpperCase())
					* Math.pow(16, position);
			transformedHexadecimal = transformedHexadecimal.substring(0,
					lengthMinusOne);
			position++;
		}

		return resultValue;
	}

	/**
	 * Converts a decimal value to a hexadecimal value.
	 * 
	 * @param decimal
	 * @return
	 */
	public static String convertDecimalToHexadecimal(int decimal) {
		String hexStr = "0123456789ABCDEF";
		StringBuilder hexResult = new StringBuilder();

		int transformedDecimal = decimal;

		while (transformedDecimal != 0) {
			int currentDigit = transformedDecimal % 16;
			transformedDecimal = (transformedDecimal - currentDigit) / 16;
			hexResult.insert(0, hexStr.charAt(currentDigit));
		}

		return hexResult.toString();
	}
}