package nl.knoppel.tools;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Calendar;

/**
 * Utility class for String manipulation.
 * 
 * @author Franois L.A. Knoppel -- francois@knoppel.nl
 * 
 */
public class StringTool {
	public static ArrayList<Character> lineEndings = new ArrayList<Character>();
	public static ArrayList<Character> punctuation = new ArrayList<Character>();
	public static ArrayList<Character> quotation = new ArrayList<Character>();
	private static Calendar timeFormatter;

	/**
	 * Initializes the timeFormatter and fills the punctuation, lineEndings and
	 * quotation arrays.
	 */
	static {
		timeFormatter = Calendar.getInstance();

		quotation.add('"');
		quotation.add('\'');
		quotation.add('`');

		lineEndings.add('!');
		lineEndings.add('?');
		lineEndings.add('.');
		lineEndings.addAll(quotation);

		punctuation.add(',');
		punctuation.addAll(lineEndings);

	}

	/**
	 * Pads out a string upto padlen with pad chars
	 * 
	 * @param Object
	 *            .toString() to be padded
	 * @param length
	 *            of pad (+ve = pad on right, -ve pad on left)
	 * @param pad
	 *            character
	 */
	public static String pad(Object str, int padlen, String pad) {
		String padding = new String();
		int len = Math.abs(padlen) - str.toString().length();
		if (len < 1)
			return str.toString();
		for (int i = 0; i < len; ++i)
			padding = padding + pad;

		return (padlen < 0 ? padding + str : str + padding);
	}

	public static String stripPunctuation(String input) {
		String output = input;

		for (char currentPunctuationChar : punctuation) {
			output = output.replace(currentPunctuationChar, '\u0000');
		}

		return output;
	}

	/**
	 * Returns whether a word is a line starting word (currently: whether the
	 * first character is uppercase).
	 * 
	 * @param word
	 * @return
	 */
	public static boolean isLineStart(String word) {
		char firstChar = word.charAt(0);
		if (Character.isUpperCase(firstChar)) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Returns whether a word is a line ending.
	 * 
	 * @param word
	 * @return
	 */
	public static boolean isLineEnding(String word) {
		char lastChar = word.charAt(word.length() - 1);
		if (lineEndings.contains(lastChar)) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Counts the number of words in the specified text.
	 * 
	 * @param text
	 * @return
	 */
	public static int getWordCount(String text) {
		String split[] = text.split("\\s");
		return split.length;
	}

	/**
	 * Counts the number of numeric characters in the specified text.
	 * 
	 * @param text
	 * @return
	 */
	public static int getNumberCount(String text) {
		int numberCount = 0;

		char[] chars = text.toCharArray();

		for (char currentChar : chars) {
			if (Character.isDigit(currentChar)) {
				numberCount++;
			}
		}

		return numberCount;
	}

	/**
	 * Counts the number of punctuation characters in the specified text.
	 * 
	 * @param text
	 * @return
	 */
	public static int getPunctuationCount(String text) {
		int punctuationCount = 0;

		char[] chars = text.toCharArray();

		for (char currentChar : chars) {
			if (punctuation.contains(currentChar)) {
				punctuationCount++;
			}
		}

		return punctuationCount;
	}

	/**
	 * Counts the number of non-alphanumeric characters in the specified text.
	 * 
	 * @param text
	 * @return
	 */
	public static int getNonAlphaNumericCount(String text) {
		int alphaNumericCount = 0;

		char[] chars = text.toCharArray();

		for (char currentChar : chars) {
			if (Character.isLetterOrDigit(currentChar)) {
				alphaNumericCount++;
			}
		}

		return text.length() - alphaNumericCount;
	}

	/**
	 * Counts the number of capitals in the specified text.
	 * 
	 * @param text
	 * @return
	 */
	public static int getCapitalsCount(String text) {
		int capitalsCount = 0;

		char[] chars = text.toCharArray();

		for (char currentChar : chars) {
			if (Character.isUpperCase(currentChar)) {
				capitalsCount++;
			}
		}

		return capitalsCount;

	}

	/**
	 * Counts the number of occurrences of the specified character in the
	 * specified text.
	 * 
	 * @param text
	 * @return
	 */
	public static int getCount(char character, String text) {
		int count = 0;

		char[] chars = text.toCharArray();

		for (char currentChar : chars) {
			if (character == currentChar) {
				count++;
			}
		}

		return count;

	}

	/**
	 * Returns whether the specified text has quotation marks as boundaries.
	 * 
	 * @param text
	 * @return
	 */
	public static boolean hasQuoteBoundaries(String text) {
		boolean quoted = false;

		String trimmedText = text.trim();
		if (trimmedText.length() > 1) {
			char firstChar = trimmedText.charAt(0);
			char lastChar = trimmedText.charAt(trimmedText.length() - 1);

			if (quotation.contains(firstChar) && quotation.contains(lastChar)) {
				quoted = true;
			}
		}

		return quoted;
	}

	/**
	 * Returns a string representation of the date associated with the specified
	 * millis count
	 * 
	 * @param time
	 * @return
	 */
	public static String getDateString(long time) {
		timeFormatter.setTimeInMillis(time);
		String dateString = MessageFormat.format("{0}-{1}-{2}", ""
				+ timeFormatter.get(Calendar.YEAR), StringTool.pad(
				timeFormatter.get(Calendar.MONTH), -2, "0"), StringTool.pad(
				timeFormatter.get(Calendar.DAY_OF_MONTH), -2, "0"));
		return dateString;
	}

	/**
	 * Returns getTimeString(time, false)
	 * 
	 * @param time
	 * @return
	 */
	public static String getTimeString(long time) {
		return getTimeString(time, false);
	}

	/**
	 * Returns a string representation of the time associated with the specified
	 * millis count.
	 * 
	 * @param time
	 * @param fileSafe
	 *            specifies whether a format should be returned that can be part
	 *            of a filename.
	 * @return
	 */
	public static String getTimeString(long time, boolean fileSafe) {
		timeFormatter.setTimeInMillis(time);
		String template = "{0}:{1}:{2}.{3}";
		if (fileSafe)
			template = "[{0}{1}h]_[{2}.{3}s]";

		String timeString = MessageFormat.format(template, StringTool.pad(
				timeFormatter.get(Calendar.HOUR_OF_DAY), -2, "0"), StringTool
				.pad(timeFormatter.get(Calendar.MINUTE), -2, "0"), StringTool
				.pad(timeFormatter.get(Calendar.SECOND), -2, "0"), StringTool
				.pad(timeFormatter.get(Calendar.MILLISECOND), -3, "0"));
		return timeString;
	}

	/**
	 * Strips surrounding quotation marks if present.
	 * 
	 * @param text
	 * @return
	 */
	public static String stripSurroundingQuotes(String text) {
		if (hasQuoteBoundaries(text)) {
			text = text.trim();
			text = text.substring(1, text.length() - 2);
		}
		return text;
	}

}