/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.fam;

import java.util.ArrayList;
import java.util.Random;
import java.util.Timer;
import java.util.logging.Logger;

import nl.utwente.ewi.hmi.deira.mmm.EmotionalState;
import nl.utwente.ewi.hmi.deira.mmm.MMM;
import nl.utwente.ewi.hmi.deira.queue.Event;
import nl.utwente.ewi.hmi.deira.queue.EventQueue;

/**
 * Bootstrapping class for FAM part (will be the interface to FAM at a later
 * point).
 * 
 * FIXME: This class combines assumptions on the input (HR) and output module
 * (Visage), functionality should be separated. The FAM translates the emotional
 * state of the agent to facial animation, in which the desired behaviour is
 * essential. Ergo: The FAM should be domain-specific and not output
 * module-specific. The output module-specific elements should be moved there.
 * 
 */
public class HRFacialAnimator extends FacialAnimator {
	private static Logger log = Logger.getLogger("deira.fam");
	/** < Logging instance */
	private Random generator;
	private Timer timer;
	private TimedHeadMotionEvent fte;

	/** < Facial Timer Event (Direct Animation Sending) reference */

	/**
	 * Constructs a new FacialAnimator.
	 * 
	 * @param eq
	 *            reference to the eventqueue.
	 * @param mmm
	 *            reference to the mental model module.
	 */
	public HRFacialAnimator(EventQueue eq, MMM mmm) {
		super("DEIRA HRFAM");

		this.queue = eq;
		this.mmm = mmm;
		this.om = null;
		this.generator = new Random();
		this.timer = new Timer("HRFAM Timer");

		moduleName = "HRFAM";
	}

	/**
	 * Runs the FacialAnimator thread. This will wait for events and attach
	 * facial expression information to the events.
	 */
	public void run() {
		this.terminate = false;

		/*
		 * Set up a timer which occasionally sends head animation independent of
		 * any event occurring in the race.
		 */
		if (this.om != null) {
			this.fte = new TimedHeadMotionEvent(this.om, this.mmm);
			long ftInterval = 1000;

			fte.setRunInterval(ftInterval);
			this.timer.schedule(fte, 18000, ftInterval);
		}

		while (!this.terminate) {
			try {
				this.queue.waitForEvent();
			} catch (Exception e) {
				log.severe("HRFAM: Error waiting for event! " + e);
			}

			Event event = this.queue.peek();

			if (event != null) {
				log.info("HRFAM: Event grabbed (" + event.getId() + ")");
				ArrayList<String> expressions = new ArrayList<String>();
				EmotionalState es = mmm.getEmotionalState();
				String eventtype = event.getType();

				// =- Add look towards user (18s)
				if (eventtype.equals("PRESTART")) {
					expressions.add("look_touser_18s");
				}

				// =- Add look towards user (11s)
				else if (eventtype.equals("POSTFINISH")) {
					expressions.add("look_touser_11s");
					fte.cancel();
				}

				// =- Add look towards user (11s)
				else if (eventtype.equals("POSTFINISHALT")) {
					expressions.add("look_touser_18s");
					fte.cancel();
				}

				// =- Do normal facial animation

				else {
					/* Find out the dominant emotion */
					addDominantEmotions(expressions);

					/* Add 'Direction' left/right Looking */
					if (generator.nextInt(5) == 3) { /*
													 * 20% chance on looking
													 * left or right
													 */
						String LRFile = "";
						if (generator.nextBoolean()) {
							LRFile = "look_right";
						} else {
							LRFile = "look_left";
						}
						expressions.add(LRFile);
					}

					/* Add 'Direction' up/down Looking */
					if (generator.nextInt(10) == 5) { /*
													 * 10% chance on looking up
													 * or down
													 */
						String UDFile = "";
						if (generator.nextBoolean()) {
							UDFile = "look_down";
						} else {
							UDFile = "look_up";
						}
						expressions.add(UDFile);
					}

					/* Add blinking (50% chance) */
					if (generator.nextBoolean()) {
						switch (generator.nextInt(3)) {
						case 0:
							expressions.add("blink_1");
							break;
						case 1:
							expressions.add("blink_2");
							break;
						case 2:
							expressions.add("blink_3");
							break;
						default:
							expressions.add("blink_1");
							break;
						}
					}

				}
				event.tag("expressions", expressions);
				log.info("HRFAM: Event processed (" + event.getId() + ")");
				this.queue.update(event);

			} else {
				log.severe("HRFAM: Woken while nothing was available!");
			}
		}
	}

	/**
	 * Instruct the FacialAnimator thread to stop waiting for events and
	 * terminate.
	 */
	public void close() {
		if (this.fte != null) {
			this.fte.cancel();
		}
		if (this.timer != null) {
			this.timer.cancel();
		}
		this.terminate = true;
	}

}
