/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.iam.riam;

import java.util.ArrayList;

import nl.utwente.ewi.hmi.deira.queue.HREvent;

/**
 * This class represents the Event as described in the event generation rule.
 * When interpreted it returns a RaceEvent.
 */
public class EventGenerationEvent {
	private EventGenerationEventType type = null;
	/** < The type of the event. */
	private EventGenerationEventLabel label = null;
	/** < The label of the event. */
	private EventGenerationEventActors actors = null;
	/** < The actors that are involved in the event. */
	private EventGenerationEventImportance importance = null;
	/** < The basic importance of the event. */
	private EventGenerationEventDecay decay = null;
	/** < The decay rate of the importance of this event, per second. */
	private EventGenerationEventParams params = null;

	/** < Optional additional parameters. */

	/**
	 * Constructor.
	 * 
	 * @param type
	 *            The type of the event.
	 * @param label
	 *            The label of the event.
	 * @param actors
	 *            The actors that are involved in the event.
	 * @param importance
	 *            The basic importance of the event.
	 * @param decay
	 *            The decay rate of the importance of the event, per second.
	 * @param params
	 *            Optional additional parameters (if none, then pass null!).
	 */
	public EventGenerationEvent(EventGenerationEventType type,
			EventGenerationEventLabel label, EventGenerationEventActors actors,
			EventGenerationEventImportance importance,
			EventGenerationEventDecay decay, EventGenerationEventParams params) {
		this.type = type;
		this.label = label;
		this.actors = actors;
		this.importance = importance;
		this.decay = decay;
		this.params = params;
	}

	/**
	 * Interprets the event part of the generation rule, resulting in an actual
	 * event.
	 * 
	 * @param newEventID
	 *            A fresh event id to indicate this newly generated event.
	 * @param raceState
	 *            The (current) raceState for which the Event needs to be
	 *            interpreted.
	 * @param startTime
	 *            The system time at which the race itself started.
	 * @return generated event.
	 */
	public HREvent interpret(int newEventID, ArrayList<Participant> raceState,
			long startTime) {
		String type = this.type.interpret(raceState);
		String label = this.label.interpret(raceState);
		ArrayList<Participant> actors = this.actors.interpret(raceState);
		float importance = this.importance.interpret(raceState);
		float decay = this.decay.interpret(raceState);
		// ArrayList<Integer> params = this.params.interpret(raceState);
		ArrayList<Float> params = this.params.interpret(raceState);

		return new HREvent(newEventID, System.currentTimeMillis() - startTime,
				type, label, actors, importance, decay, params);
	}

	/**
	 * Function to create a String representation of this instance. Mainly for
	 * debugging purposes.
	 * 
	 * @return A String representation of this instance.
	 */
	public String toString() {
		String s = "";
		s += type.toString() + ";";
		s += label.toString() + ";";
		s += actors.toString() + ";";
		s += importance.toString() + ";";
		s += decay.toString() + ";";
		s += params.toString();
		return s;
	}
}
