/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

/* 
 * TODO seperate operator classes for boolean, number and string
 */

package nl.utwente.ewi.hmi.deira.iam.riam;

import java.util.ArrayList;

/**
 * This class is a representation of any (binary) operator in an event generation rule.
 * The operator can be either boolean or mathematical.
 * When interpreted it returns a float value corresponding to the operator and the two involved expressions.
 * 
 * Expr		::= 	(Expr [ + | - | / | * | > | >= | < | <= | == | != | && | || ] Expr)
 *			|		[ #1 | #2 | #3 | #4 ].[ curr | prev ].[ speed | rank | pos ]
 *			|		[ INT | FLOAT ] 
 *
 */
public class EventGenerationOperator {
	private String op;

	/** < String indicating the operator. */

	/**
	 * Constructor.
	 * 
	 * @param op
	 *            String indicating the operator.
	 */
	public EventGenerationOperator(String op) {
		this.op = op;
	}

	/**
	 * Interprets the operator's influence on the two passed on expressions.
	 * 
	 * @param raceState
	 *            The (current) raceState for which the expression needs to be
	 *            interpreted.
	 * @param expr1
	 *            The first expression on which the operator will be executed.
	 * @param expr2
	 *            The second expression on which the operator will be executed.
	 * @return The expression value defined by executing the binary operator on
	 *         the two expressions. If the operator could not be interpreted,
	 *         the result is 0.0f.
	 */
	public float interpret(ArrayList<Participant> raceState, float expr1,
			float expr2) {
		if (op.equals(">")) {
			if (expr1 > expr2)
				return 1.0f;
			else
				return 0.0f;
		} else if (op.equals(">=")) {
			if (expr1 >= expr2)
				return 1.0f;
			else
				return 0.0f;
		} else if (op.equals("<")) {
			if (expr1 < expr2)
				return 1.0f;
			else
				return 0.0f;
		} else if (op.equals("<=")) {
			if (expr1 <= expr2)
				return 1.0f;
			else
				return 0.0f;
		} else if (op.equals("==")) {
			if (expr1 == expr2)
				return 1.0f;
			else
				return 0.0f;
		} else if (op.equals("!=")) {
			if (expr1 != expr2)
				return 1.0f;
			else
				return 0.0f;
		} else if (op.equals("&&")) {
			if (expr1 == 1.0f && expr2 == 1.0f)
				return 1.0f;
			else
				return 0.0f;
		} else if (op.equals("||")) {
			if (expr1 == 1.0f || expr2 == 1.0f)
				return 1.0f;
			else
				return 0.0f;
		} else if (op.equals("+")) {
			return expr1 + expr2;
		} else if (op.equals("-")) {
			return expr1 - expr2;
		} else if (op.equals("/")) {
			return expr1 / expr2;
		} else if (op.equals("*")) {
			return expr1 * expr2;
		} else
			return 0.0f;

	}

	/**
	 * Function to create a String representation of this operator. Mainly for
	 * debugging purposes.
	 * 
	 * @return A String representation of this operator.
	 */
	public String toString() {
		return op;
	}

}
