/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

/*
 * TODO constructor: zit stukje parse in -> alle parser stukjes naar de constructors!
 */

package nl.utwente.ewi.hmi.deira.iam.riam;

import java.util.ArrayList;
import java.util.StringTokenizer;
import java.util.logging.Logger;

/**
 * This class is a representation of any participant expression in the event generation rules.
 * When interpreted it returns a float value corresponding to the expression.
 * 
 * Added in version 2.0: time data.
 * 
 * A participant expression looks like: [ #1 | #2 | #3 | #4 ].[ curr | prev ].[ speed | rank | pos | time]
 */
public class EventGenerationParticipantExpression implements
		EventGenerationExpression {
	private String participant;
	/** < the participant (rank) indicated by the participant expression */
	private String state;
	/** < the state (curr, prev) indicated by the participant expression */
	private String data;
	/** < the data (rank, pos, speed) indicated by the participant expression */

	private static Logger log = Logger.getLogger("deira.ram");

	/** < Logging instance */

	/**
	 * Constructor.
	 * 
	 * @param expr
	 *            String indicating the participant expression.
	 */
	public EventGenerationParticipantExpression(String expr) {
		StringTokenizer st = new StringTokenizer(expr, ".");
		participant = st.nextToken();
		state = st.nextToken();
		data = st.nextToken();
	}

	public EventGenerationParticipantExpression(String participant,
			String state, String data) {
		this.participant = participant;
		this.state = state;
		this.data = data;
	}

	/**
	 * Interprets the expression.
	 * 
	 * @param raceState
	 *            The (current) raceState for which the expression needs to be
	 *            interpreted.
	 * @return The expression value.
	 */
	public float interpret(ArrayList<Participant> raceState) {
		Participant[] actor = new Participant[4];

		boolean noCurrentState = false;
		for (int i = 0; i < raceState.size(); i++) {
			State s = raceState.get(i).getCurrentState();
			if (s != null)
				actor[s.getRank() - 1] = raceState.get(i);
			else {
				noCurrentState = true;
			}
		}
		if (noCurrentState) {
			actor[0] = raceState.get(0);
			actor[1] = raceState.get(1);
			actor[2] = raceState.get(2);
			actor[3] = raceState.get(3);
		}

		Participant p;
		if (participant.equals("#1"))
			p = actor[0];
		else if (participant.equals("#2"))
			p = actor[1];
		else if (participant.equals("#3"))
			p = actor[2];
		else if (participant.equals("#4"))
			p = actor[3];
		else {
			log.warning("Nonexisting participant: " + participant);
			return 0.0f;
		}
		if (p == null) {
			// this should not happen, unless there is an error in the code
			// still, we don't throw an exception, because the rest may still
			// work...
			log.warning("System error: participant for " + participant
					+ " is not defined.");
			return 0.0f;
		}

		State s;
		if (state.equals("curr"))
			s = p.getCurrentState();
		else if (state.equals("prev"))
			s = p.getPreviousState();
		else {
			log.warning("Nonexisting state: " + state);
			return 0.0f;
		}
		if (s == null) {
			// only happens if this code gets called when the participant does
			// not have a current / previous state yet
			// this is currently checked in RaceAnalysis
			log.warning("System error: " + state + " for " + participant
					+ " is not defined.");
			return 0.0f;
		}

		int d;
		if (data.equals("speed")) {
			if (s != null)
				d = s.getSpeed();
			else
				d = 0;
		} else if (data.equals("pos")) {
			if (s != null)
				d = s.getPosition();
			else
				d = 0;
		} else if (data.equals("rank")) {
			if (s != null)
				d = s.getRank();
			else {
				if (participant.equals("#1"))
					d = 1;
				else if (participant.equals("#2"))
					d = 2;
				else if (participant.equals("#3"))
					d = 3;
				else
					d = 4;
			}
		} else if (data.equals("time")) {
			if (s != null) {
				long t = (s.getTimestamp());
				// 1166596016000 -> 6596016000
				// 1166596016000 - 1166596016000/10000000000 = 6596016000 ?
				int ti = new Long(t).intValue();
				// TODO het volgende statement werkt niet, maar wel zoiets nodig
				// :/
				// if (t != ti) log.warning("Time to int overflow");
				// log.info("Long to int: " + t + " -> " + ti);
				d = ti;
			} else
				d = 0;
		} else {
			log.warning("Nonexisting data: " + data);
			return 0.0f;
		}

		return (float) d;
	}

	/**
	 * Function to create a String representation of this expression. Mainly for
	 * debugging purposes.
	 * 
	 * @return A String representation of this expression.
	 */
	public String toString() {
		return participant + "." + state + "." + data;
	}
}
