/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.iam.riam;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.StringTokenizer;

/**
 * Reads in a text file with event generation rule descriptions specified in the
 * events generation rule language. These rules need to be parsed into event
 * generation rule trees.
 * 
 * @deprecated The use of Parser2 is preferred, because of the more steady
 *             implementation.
 */
public class Parser {
	/**
	 * Constructor.
	 */
	public Parser() {
	}

	/**
	 * Parses an entire textfile and returns all event generation rules it
	 * contained.
	 * 
	 * @param textfile
	 *            The file that contains the event generation rules specified in
	 *            the event generation rules language.
	 * @return Event generation rules parsed from the text file.
	 */
	public ArrayList<EventGenerationRule> parse(String textfile) {
		ArrayList<EventGenerationRule> rules = new ArrayList<EventGenerationRule>();
		try {
			InputStream ist = new FileInputStream(textfile);
			BufferedReader istream = new BufferedReader(new InputStreamReader(
					ist));
			String text;
			while ((text = istream.readLine()) != null) {
				// een rule per regel
				text = text.trim();
				if (text.length() > 0 && (text.charAt(0) != '/')) {
					EventGenerationRule rule = parseRule(text);
					// System.out.println("Rule: " + rule);
					// System.out.println();
					rules.add(rule);
				}
			}
		} catch (FileNotFoundException e) {
			System.out.println("Parser could not find " + textfile);
		} catch (IOException e) {
			System.out.println("Parser could not read from " + textfile);
		}
		return rules;
	}

	/*
	 * Expr ::= (Expr [ + | - | / | ] Expr) | [ #1 | #2 | #3 | #4 ].[ curr |
	 * prev ].[ speed | rank | pos ] | [ INT | FLOAT ] Bool_expr ::= (Expr [ > |
	 * >= | < | <= | == | != ] Expr) | (Bool_expr [ && | || ] Bool_expr)
	 */

	/**
	 * Parses a String that should be an expression.
	 * 
	 * @param s
	 *            String representation of an expression.
	 * @return An event generation expression as specified by s.
	 */
	public EventGenerationExpression parseEventExpression(String s) {
		// niet helemaal de nette manier ^^;
		// System.out.println("parse expr: " + s);
		if (s.length() < 1)
			throw new RuntimeException("Parser: expression expected");
		if (s.substring(0, 1).equals("(")) {
			// (Expr [ + | - | / | * ] Expr)
			// (Expr [ > | >= | < | <= | == | != ] Expr)
			// (Bool_expr [ && | || ] Bool_expr)
			// bijbehorende op: 1 haakje open, eerste op
			StringTokenizer st = new StringTokenizer(s.substring(1,
					s.length() - 1), "()", true);
			String prev = "";
			String next = "";
			String token = "";
			boolean opFound = false;
			String op = "";
			int opencount = 1;
			while (st.hasMoreTokens()) {
				token = st.nextToken().trim();
				// System.out.println("parse token " + token);
				if (opFound) {
					next += token;
					// System.out.println("next: " + next);
				} else if (token.equals("(")) {
					opencount++;
					prev += token;
					// System.out.println("prev: " + prev);
				} else if (token.equals(")")) {
					opencount--;
					prev += token;
					// System.out.println("prev: " + prev);
				} else if (opencount == 1) {
					StringTokenizer opst = new StringTokenizer(token.trim(),
							"+-/*><=!&|", true);
					int n = opst.countTokens();
					if (inString(token.substring(0, 1), "+-/*><=!&|")) {
						if (n == 1) {
							// ( (a+b)| - (c+d) ) -> 1 token
							op = token;
							// System.out.println("1 token: " + op);
							opFound = true;
						} else if (n == 2) {
							// System.out.println("2 tokens a");
							// ( (a+b)| - c) -> 2 tokens
							// ( (a+b)| >= (c+d) ) -> 2 token
							op = opst.nextToken().trim();
							opFound = true;
							String op2 = opst.nextToken().trim();
							if (inString(op2, "+-/*><=!&|"))
								op += op2;
							else
								next += op2;
						} else if (n == 3) {
							// ( (a+b)| => c) -> 2 tokens
							op = opst.nextToken().trim();
							op += opst.nextToken().trim();
							next += opst.nextToken().trim();
						}
					} else {
						if (n == 2) {
							// System.out.println("2 tokens b");
							// (| a - (b+c)) -> 2 tokens
							prev += opst.nextToken().trim();
							op = opst.nextToken().trim();
							opFound = true;
						} else if (n == 3) {
							// (| a - b ) -> 3 tokens
							// (| a >= (b+c)) -> 3 tokens
							prev += opst.nextToken().trim();
							op = opst.nextToken().trim();
							String op2 = opst.nextToken().trim();
							if (inString(op2, "+-/*><=!&|"))
								op += op2;
							else
								next += op2;
							opFound = true;
							// System.out.println("3 tokens: " + prev + ", " +
							// op + ", " + next);
						} else if (n == 4) {
							// >=, <=, ==, !=, &&, ||
							// (| a >= b ) -> 4 tokens
							prev += opst.nextToken().trim();
							op = opst.nextToken().trim();
							op += opst.nextToken().trim();
							opFound = true;
							next += opst.nextToken().trim();
							// System.out.println("4 tokens: " + prev + ", " +
							// op + ", " + next);
						}
					}
				} else {
					prev += token;
					// System.out.println("prev: " + prev);
				}
			}
			EventGenerationExpression expr1 = parseEventExpression(prev);
			EventGenerationExpression expr2 = parseEventExpression(next);
			EventGenerationOperator operator = new EventGenerationOperator(op);
			// System.out.println("expr1 for " + s + ": " + expr1);
			// System.out.println("operator for " + s + ": " + operator);
			// System.out.println("expr2 for " + s + ": " + expr2);
			// er zit geen feitelijk verschil tussen Math en Boolean Expression
			// ^^;
			return new EventGenerationMathExpression(expr1, operator, expr2);
		} else if (s.substring(0, 1).equals("#")) {
			// [ #1 | #2 | #3 | #4 ].[ curr | prev ].[ speed | rank | pos ]
			return new EventGenerationParticipantExpression(s);
		} else if (Character.isDigit(s.substring(0, 1).toCharArray()[0])) {
			// [ INT | FLOAT ]
			return new EventGenerationNumberExpression(s);
		} else
			return null;
	}

	/**
	 * Parses a String that should represent a list of the actors involved in
	 * the event.
	 * 
	 * @param s
	 *            String representation of a list of actors by rank.
	 * @return An Event Generation Event Actors class containing a list of these
	 *         actors.
	 */
	public EventGenerationEventActors parseEventActors(String s) {
		// < #1, #2, #3, #4 >
		ArrayList<Integer> actors = new ArrayList<Integer>();
		StringTokenizer st = new StringTokenizer(
				s.substring(1, s.length() - 1), ",");

		while (st.hasMoreTokens()) {
			String token = st.nextToken().trim();
			if (token.length() > 1 && token.charAt(0) == '#') {
				int p = Integer.parseInt(token.trim().substring(1, 2));
				actors.add(p);
			}
		}

		return new EventGenerationEventActors(actors);
	}

	/**
	 * Parses a String that should represent an event to be generated.
	 * 
	 * @param s
	 *            String representation of an event.
	 * @return An EventGenerationEvent that can be interpreted to a RaceEvent.
	 */
	public EventGenerationEvent parseEvent(String s) { // Type; Label; Actors;
														// Importance;
														// Decay_rate; Params
		StringTokenizer st = new StringTokenizer(s, ";", true);

		EventGenerationEventType type;
		EventGenerationEventLabel label;
		EventGenerationEventActors actors;
		EventGenerationEventImportance importance;
		EventGenerationEventDecay decay;
		EventGenerationEventParams params = null; // TODO iteration 2 ??

		String token = st.nextToken().trim();
		// System.out.println("type token: " + token);
		if (token.equals(";"))
			type = new EventGenerationEventType("");
		else if (token.equals("")) {
			type = new EventGenerationEventType("");
			st.nextToken(); // ;
		} else {
			type = new EventGenerationEventType(token);
			st.nextToken(); // ;
		}

		token = st.nextToken().trim();
		// System.out.println("label token: " + token);
		if (token.equals(";"))
			label = new EventGenerationEventLabel("");
		else if (token.equals("")) {
			label = new EventGenerationEventLabel("");
			st.nextToken(); // ;
		} else {
			label = new EventGenerationEventLabel(token);
			st.nextToken(); // ;
		}

		token = st.nextToken().trim();
		// System.out.println("actors token: " + token);
		if (token.equals(";"))
			actors = parseEventActors("<>");
		else if (token.equals("")) {
			actors = parseEventActors("<>");
			st.nextToken(); // ;
		} else {
			actors = parseEventActors(token);
			st.nextToken(); // ;
		}

		token = st.nextToken().trim();
		// System.out.println("importance token: " + token);
		if (token.equals(";"))
			importance = new EventGenerationEventImportance(
					parseEventExpression("0"));
		else if (token.equals("")) {
			importance = new EventGenerationEventImportance(
					parseEventExpression("0"));
			st.nextToken(); // ;
		} else {
			importance = new EventGenerationEventImportance(
					parseEventExpression(token));
			st.nextToken(); // ;
		}

		token = st.nextToken().trim();
		// System.out.println("decay token: " + token);
		if (token.equals(";"))
			decay = new EventGenerationEventDecay(parseEventExpression("0"));
		else if (token.equals("")) {
			decay = new EventGenerationEventDecay(parseEventExpression("0"));
			st.nextToken(); // ;
		} else {
			decay = new EventGenerationEventDecay(parseEventExpression(token));
			st.nextToken(); // ;
		}

		return new EventGenerationEvent(type, label, actors, importance, decay,
				params);
	}

	/**
	 * Parses an event generation rule.
	 * 
	 * @param s
	 *            String representation of an event generation rule.
	 * @return the event generation rule as represented by the String.
	 */
	public EventGenerationRule parseRule(String s) {
		// StringTokenizer st = new StringTokenizer(s, "->");
		String[] part = s.split("->");
		EventGenerationExpression b = parseEventExpression(part[0].trim());
		EventGenerationEvent e = parseEvent(part[1].trim());
		return new EventGenerationRule(new EventGenerationRequirement(b), e);
	}

	/**
	 * Help function to say whether c can be found in s.
	 * 
	 * @param c
	 *            String of one Character.
	 * @param s
	 *            String containing some possibilities.
	 * @return true if c could be found in s.
	 */
	public boolean inString(String c, String s) {
		for (int i = 0; i < s.length(); i++) {
			if (c.equals(s.substring(i, i + 1)))
				return true;
		}
		return false;
	}

}
