/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

/* 
 * TODO evt ook parse methodes in de klasses zelf plaatsen? (net als interpret)
 */

package nl.utwente.ewi.hmi.deira.iam.riam;

import java.util.ArrayList;
import java.util.logging.Logger;

/**
 * Reads in a text file with event generation rule descriptions specified in the
 * events generation rule language. These rules need to be parsed into event
 * generation rule trees.
 */
public class Parser2 {

	private static Logger log = Logger.getLogger("deira.ram");
	/** < Logging instance. */

	ArrayList<ArrayList<Token>> lines;
	/** Lines as series (ArrayLists) of Tokens. */
	int linenr;
	/** Currently parsing line (number). */
	int tokennr;
	/** Currently parsing token (number). */
	Tokenizer tokenizer;

	/** Tokenizer to split up the script into Tokens to facilitate parsing */

	/**
	 * Constructor.
	 */
	public Parser2() {
		tokenizer = new Tokenizer();
	}

	/**
	 * Parses an entire textfile and returns all event generation rules it
	 * contained.
	 * 
	 * @param textfile
	 *            The file that contains the event generation rules specified in
	 *            the event generation rules language.
	 * @return Event generation rules parsed from the text file.
	 */
	public ArrayList<EventGenerationRule> parse(String textfile) {
		tokenizer.load(textfile);
		lines = tokenizer.getLines();
		ArrayList<EventGenerationRule> rules = new ArrayList<EventGenerationRule>();
		linenr = 0;
		for (ArrayList<Token> line : lines) {
			tokennr = 0;
			EventGenerationRule rule = parseRule(line);
			rules.add(rule);
			// log.info(rule.toString());
			linenr++;
		}

		return rules;
	}

	/*
	 * Expr ::= (Expr [ + | - | / | ] Expr) | [ #1 | #2 | #3 | #4 ].[ curr |
	 * prev ].[ speed | rank | pos ] | [ INT | FLOAT ] Bool_expr ::= (Expr [ > |
	 * >= | < | <= | == | != ] Expr) | (Bool_expr [ && | || ] Bool_expr)
	 */
	/**
	 * Parses a String that should be an expression.
	 * 
	 * @param s
	 *            String representation of an expression.
	 * @return An event generation expression as specified by s.
	 */
	public EventGenerationExpression parseEventExpression() {
		ArrayList<Token> line = lines.get(linenr);

		if (line.get(tokennr) == null)
			log.severe("Expression expected on line " + linenr);
		if (peek(Token.LPAREN)) {
			take(Token.LPAREN);
			EventGenerationExpression e1 = parseEventExpression();
			// take any: + - / * > >= < <= == != && ||
			EventGenerationOperator op = new EventGenerationOperator(take(null));
			EventGenerationExpression e2 = parseEventExpression();
			take(Token.RPAREN);
			return new EventGenerationMathExpression(e1, op, e2);
		} else if (peek(Token.NUMBER)) {
			String s = take(Token.NUMBER);
			return new EventGenerationNumberExpression(s);
		} else {
			// [ #1 | #2 | #3 | #4 ].[ curr | prev ].[ speed | rank | pos ]
			String p = take(Token.IDENTIFIER);
			take(Token.DOT);
			String s = take(Token.IDENTIFIER);
			take(Token.DOT);
			String d = take(Token.IDENTIFIER);
			return new EventGenerationParticipantExpression(p, s, d);
		}
	}

	/**
	 * Parses a String that should represent a list of the actors involved in
	 * the event.
	 * 
	 * @param s
	 *            String representation of a list of actors by rank.
	 * @return An Event Generation Event Actors class containing a list of these
	 *         actors.
	 */
	public EventGenerationEventActors parseEventActors() {
		ArrayList<Integer> actors = new ArrayList<Integer>();
		// < #1, #2, #3, #4 >
		if (!peek(Token.LT))
			return new EventGenerationEventActors(null);

		take(Token.LT);
		while (!peek(Token.GT)) {
			int p = Integer.parseInt(""
					+ lines.get(linenr).get(tokennr).toString().charAt(1));
			actors.add(p);
			take(Token.IDENTIFIER);
			// niet 100% correct... maar werkt wel
			if (peek(Token.COMMA))
				take(Token.COMMA);
		}
		take(Token.GT);

		return new EventGenerationEventActors(actors);
	}

	/**
	 * Parses event parameters (at the current line and column).
	 * 
	 * @return the parsed event generation event parameters.
	 */
	public EventGenerationEventParams parseEventParams() {
		// < x, y, z >
		if (!peek(Token.LT))
			return new EventGenerationEventParams(null);

		ArrayList<EventGenerationExpression> params = new ArrayList<EventGenerationExpression>();
		take(Token.LT);
		while (!peek(Token.GT)) {
			params.add(parseEventExpression());
			// niet 100% correct... maar werkt wel
			if (peek(Token.COMMA))
				take(Token.COMMA);
		}
		take(Token.GT);
		return new EventGenerationEventParams(params);
	}

	/**
	 * Parses a String that should represent an event to be generated.
	 * 
	 * @param s
	 *            String representation of an event.
	 * @return An EventGenerationEvent that can be interpreted to a RaceEvent.
	 */
	private EventGenerationEvent parseEvent() { // Type; Label; Actors;
		// Importance; Decay_rate;
		// Params

		EventGenerationEventType type = new EventGenerationEventType(
				take(Token.IDENTIFIER));
		take(Token.SEMICOLON);
		EventGenerationEventLabel label = new EventGenerationEventLabel("");
		if (!peek(Token.SEMICOLON))
			label = new EventGenerationEventLabel(take(null)); // IDENTIFIER or
		// NUMBER !
		take(Token.SEMICOLON);
		EventGenerationEventActors actors = parseEventActors();
		take(Token.SEMICOLON);
		EventGenerationEventImportance importance = new EventGenerationEventImportance(
				parseEventExpression());
		take(Token.SEMICOLON);
		EventGenerationEventDecay decay = new EventGenerationEventDecay(
				parseEventExpression());
		take(Token.SEMICOLON);
		EventGenerationEventParams params = parseEventParams();

		return new EventGenerationEvent(type, label, actors, importance, decay,
				params);
	}

	/**
	 * For the RaceAnalysis module, to create 'home-made' events (in particular:
	 * prestart, and start).
	 * 
	 * @param s
	 *            Text representation of the event, in event generation script
	 *            format.
	 * @return The EventGenerationEvent represented by the text.
	 */
	public EventGenerationEvent parseEventText(String s) {
		lines = new ArrayList<ArrayList<Token>>();
		ArrayList<Token> line = tokenizer.tokenize(s);
		lines.add(line);
		linenr = 0;
		tokennr = 0;
		EventGenerationEvent e = parseEvent();
		return e;
	}

	/**
	 * Parses an event generation rule.
	 * 
	 * @param s
	 *            String representation of an event generation rule.
	 * @return the event generation rule as represented by the String.
	 */
	public EventGenerationRule parseRule(ArrayList<Token> line) {
		EventGenerationRequirement b = new EventGenerationRequirement(
				parseEventExpression());
		take(Token.RESULTS);
		EventGenerationEvent e = parseEvent();

		return new EventGenerationRule(b, e);
	}

	/**
	 * Takes the next token, if the type matches (or always if type == null).
	 * 
	 * @param type
	 *            The type the token to take should have. To take any type, use
	 *            null.
	 * @return The string representation of the token taken.
	 */
	private String take(String type) {
		if (linenr >= lines.size()) {
			log.severe("Line expected");
			return "";
		}
		ArrayList<Token> line = lines.get(linenr);

		if (tokennr >= line.size()) {
			log.severe("Token expected");
			return "";
		}
		Token token = line.get(tokennr);

		log.finest("Take: \t" + token.toString() + "\t" + type);

		if (type == null) { // take any
			tokennr++;
			return token.toString();
		} else if (token.getType().equals(type)) {
			tokennr++;
			return token.toString();
		} else {
			log.severe(type + " expected on line " + (linenr + 1) + " after "
					+ lines.get(linenr).get(tokennr - 1));
			// throw new RuntimeException("oi!");
			// TODO wat liever? doorgaan met foute rule, of stoppen?
			return "";
		}
	}

	/**
	 * Function to check whether the upcoming Token is of a certain type.
	 * 
	 * @param type
	 *            The type which the next Token is checked for.
	 * @return Whether the next Token is of type type.
	 */
	private boolean peek(String type) {
		if (linenr >= lines.size())
			return false;
		ArrayList<Token> line = lines.get(linenr);

		if (tokennr >= line.size())
			return false;
		Token token = line.get(tokennr);

		if (token.getType().equals(type))
			return true;
		else
			return false;
	}

}
