/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.iam.riam;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.logging.Logger;

import nl.utwente.ewi.hmi.deira.db.BackgroundInformation;
import nl.utwente.ewi.hmi.deira.generic.DeiraThread;
import nl.utwente.ewi.hmi.deira.generic.Module;
import nl.utwente.ewi.hmi.deira.om.OutputtedEvent;
import nl.utwente.ewi.hmi.deira.queue.Event;
import nl.utwente.ewi.hmi.deira.queue.HREvent;

/**
 * Module that receives race state information from the HRSS. This information
 * is pushed through the event generation rules. Generated events are passed on
 * to the MMM.
 */
public class RaceAnalysis extends DeiraThread implements OutputtedEvent, Module {
	// variables for UDP connection
	private byte[] data = new byte[1024];
	private DatagramSocket socket;
	/** UDP connection socket */
	private DatagramPacket packet;
	/** Datagram packet for information over the UDP connection */
	private ByteBuffer byteBuffer;
	/** Buffer for information over the UDP connection. */
	private String hrsshost = "localhost";
	/** HRSS server. */
	private int hrssport = 2002;
	/** Port for the HRSS, to set up the UDP connection. */
	private int ramport = 2001;
	/** Port for the RIAM, to set up the UDP connection. */
	private final static char start = 's';
	/** Start signal to be sent to the HRSS to start the race. */

	// race state
	private ArrayList<Participant> raceState = new ArrayList<Participant>();
	/**
	 * The race state consisting of all the participants, which in their turn
	 * know their own current and previous states.
	 */
	private long startTime;
	/** Time at which the race started (system time). */
	private int eventCounter = 0;
	/** Event counter to maintain a unique event id for dispatch. */
	private nl.utwente.ewi.hmi.deira.mmm.MMM mmm;
	/** Mental model module to which the events need to be passed on. */
	private Parser2 parser;
	/** Parser to parse the rules script into event generation rules. */
	private ArrayList<EventGenerationRule> rules;
	/** Event generation rules as resulted from the parser. */
	private Comparator<Object> comparator = new ParticipantComparator<Object>();
	/** Comparator which determines the order/ranks of the participants. */
	private long generation_of_last_event;
	private boolean prestartDone;
	public boolean startRace;
	/** < Set to true when the race can be started */
	private boolean run = true;

	private static Logger log = Logger.getLogger("deira.riam");

	/** < Logging instance */

	/**
	 * Constructor. Uses as rule script 'test.rls'.
	 * 
	 * @param mmm
	 *            The mental model module to pass the events to.
	 */
	public RaceAnalysis(nl.utwente.ewi.hmi.deira.mmm.MMM mmm) {
		this.mmm = mmm;
		this.prestartDone = false;

		raceState.add(new Participant(BackgroundInformation.PARTICIPANT[0]));
		raceState.add(new Participant(BackgroundInformation.PARTICIPANT[1]));
		raceState.add(new Participant(BackgroundInformation.PARTICIPANT[2]));
		raceState.add(new Participant(BackgroundInformation.PARTICIPANT[3]));

		parser = new Parser2();
		rules = parser.parse("test.rls");

		this.setName("DEIRA RIAM");
	}

	/**
	 * Sets up an UDP connection with the HRSS
	 */
	public void setupConnection() {
		try {
			// datagram socket receiving from port
			socket = new DatagramSocket(ramport);
			packet = new DatagramPacket(data, data.length);
			byteBuffer = ByteBuffer.wrap(data);
		} catch (Exception e) {
			log.severe(e.toString());
		}
	}

	/**
	 * Sends start signal to the HRSS ('s'). Adds start states to the
	 * participants (position 0, speed 0 is not passed on by the HRSS !). Adds
	 * start event to queue. TODO pass start event to the MMM Start checking for
	 * new data from the HRSS (-> receiveData()).
	 */
	public void startRace() {

		byte[] startBytes = String.valueOf(start).getBytes();
		try {
			DatagramPacket dp = new DatagramPacket(startBytes, 0,
					startBytes.length, InetAddress.getByName(hrsshost),
					hrssport);
			socket.send(dp);
			log.info("Sent start signal to HRSS");
		} catch (java.net.UnknownHostException e) {
			log.severe("Could not resolve hostname for HRSS: " + hrsshost);
		} catch (java.io.IOException e) {
			log.severe("Could not send start signal to HRSS.");
		}

		startTime = System.currentTimeMillis();

		raceState.get(0)
				.addState(
						new nl.utwente.ewi.hmi.deira.iam.riam.State(startTime,
								0, 0, 1));
		raceState.get(1)
				.addState(
						new nl.utwente.ewi.hmi.deira.iam.riam.State(startTime,
								0, 0, 2));
		raceState.get(2)
				.addState(
						new nl.utwente.ewi.hmi.deira.iam.riam.State(startTime,
								0, 0, 3));
		raceState.get(3)
				.addState(
						new nl.utwente.ewi.hmi.deira.iam.riam.State(startTime,
								0, 0, 4));

		// dispatch start event
		mmm.processEvent(parser.parseEventText(
				"START;;<>;" + HREvent.START_IMPORTANCE + ";"
						+ HREvent.START_DECAY + ";").interpret(eventCounter,
				raceState, startTime));
		getNewEventID();

		// start up the 'ticks' for iam.riam
		receiveData();

	}

	/**
	 * Keeps receiving data via the UDP connection from the HRSS, until the
	 * connection is broken or socket is closed. Analyzes this data to update
	 * the current race state. Generates events based on the event rules and
	 * race state. Passes the generated events on to the mmm
	 */
	public void receiveData() {
		try {
			while (!socket.isClosed()) {
				// receive data from HRSS
				socket.receive(packet);
				int timestamp = byteBuffer.getInt(0);
				log.fine("HRSS time: " + timestamp);
				int loc[] = new int[4];
				int spd[] = new int[4];
				// int rnk[] = new int[4];
				// rnk[0] = 0; rnk[1] = 0; rnk[2] = 0; rnk[3] = 0;

				loc[0] = byteBuffer.getInt(4);
				spd[0] = byteBuffer.getInt(8);

				loc[1] = byteBuffer.getInt(12);
				spd[1] = byteBuffer.getInt(16);

				loc[2] = byteBuffer.getInt(20);
				spd[2] = byteBuffer.getInt(24);

				loc[3] = byteBuffer.getInt(28);
				spd[3] = byteBuffer.getInt(32);

				// update raceState
				Participant[] participant = { raceState.get(0),
						raceState.get(1), raceState.get(2), raceState.get(3) };

				long currTime = System.currentTimeMillis();
				participant[0]
						.addState(new nl.utwente.ewi.hmi.deira.iam.riam.State(
								currTime - startTime, loc[0], spd[0], 0));
				participant[1]
						.addState(new nl.utwente.ewi.hmi.deira.iam.riam.State(
								currTime - startTime, loc[1], spd[1], 0));
				participant[2]
						.addState(new nl.utwente.ewi.hmi.deira.iam.riam.State(
								currTime - startTime, loc[2], spd[2], 0));
				participant[3]
						.addState(new nl.utwente.ewi.hmi.deira.iam.riam.State(
								currTime - startTime, loc[3], spd[3], 0));

				// determine ranks
				Object[] a = raceState.toArray();
				Arrays.sort(a, comparator);
				for (int i = 0; i < a.length; i++) {
					Participant p = (Participant) a[i];
					p.getCurrentState().setRank(i + 1);
				}

				log.fine("raceState " + raceState.get(0) + ": "
						+ raceState.get(0).getCurrentState());
				log.fine("raceState " + raceState.get(1) + ": "
						+ raceState.get(1).getCurrentState());
				log.fine("raceState " + raceState.get(2) + ": "
						+ raceState.get(2).getCurrentState());
				log.fine("raceState " + raceState.get(3) + ": "
						+ raceState.get(3).getCurrentState());

				// pass race state through event rules for event generation
				int nevents = 0;

				if (participant[0].stateSize() >= 2) {
					for (int i = 0; i < rules.size(); i++) {
						HREvent event = rules.get(i).interpret(eventCounter,
								raceState, startTime);
						if (event != null) {
							getNewEventID();
							// queue.add(event);
							if (event.getType().equals("POSTFINISHALT")) {
								this.close();
							}
							mmm.processEvent(event);
							this.generation_of_last_event = System
									.currentTimeMillis();
							nevents++;
						}
					}

					long background_falloff = 6000;
					long time_since_last_event = System.currentTimeMillis()
							- this.generation_of_last_event;
					// generate background event if no other event occurred and
					// the set time has passed since the last event.
					if (nevents == 0
							&& (time_since_last_event > background_falloff)) {
						getNewEventID();
						mmm.processEvent(parser.parseEventText(
								"BACKGROUND;;<#1,#2,#3,#4>;"
										+ HREvent.BACKGROUND_IMPORTANCE + ";"
										+ HREvent.BACKGROUND_DECAY + ";")
								.interpret(eventCounter, raceState, startTime));
					}

				}
			}

		} catch (IOException se) {
			log.info("Socket closed.");
		}

	}

	/**
	 * Returns a fresh unused event id. (increases eventCounter by 1 to do this)
	 * 
	 * @return a fresh unused event id.
	 */
	public int getNewEventID() {
		eventCounter++;
		return eventCounter;
	}

	/**
	 * Closes the RaceAnalysis, which mainly consists of closing the socket,
	 * which in turn results in stopping receiving data.
	 */
	public void close() {
		run = false;
		log.info("Closing socket...");
		if (socket != null)
			socket.close();
	}

	public void OnOutputtedEvent(Event event) {
		this.prestartDone = true;
	}

	/**
	 * Starts the race and receiving data.
	 */
	public void run() {
		// =-- Set up the connection with the HRSS
		setupConnection();

		// =-- Wait for all threads to be ready TODO: Beter een signaaltje van
		// alle threads dan arbitrary wait.
		log.info("RIAM: Pausing to let all threads start");
		try {
			synchronized (this) {
				this.wait(1000);
			}
		} catch (InterruptedException e) {
		}

		// =-- Wait for start race signal!
		try {
			synchronized (this) {
				while (run && !this.startRace) {
					this.wait(1000);
				}
			}
		} catch (InterruptedException e) {
		}

		// =-- Generate pre-start event
		// Request to be notified on destruction of the event.
		HREvent event = parser.parseEventText(
				"PRESTART;;<>;" + HREvent.PRESTART_IMPORTANCE + ";"
						+ HREvent.PRESTART_DECAY + ";").interpret(eventCounter,
				raceState, System.currentTimeMillis());
		event.tag("outputtedNotify", this);
		mmm.processEvent(event);
		getNewEventID();

		// =-- Wait for OM to send ready to go!
		try {
			log
					.info("RIAM: Waiting for permission to start the race (from OM)");
			while (!this.prestartDone && !socket.isClosed()) {
				synchronized (this) {
					this.wait(1000);
				}
			}
		} catch (InterruptedException e) {
		}

		if (socket.isClosed()) {
			log.info("RIAM: Socket Link Lost!");
			this.prestartDone = false;
		} else {
			log.info("RIAM: Got Permission: Starting!");

			// =-- Start the race!
			startRace();
		}
	}

	public String getModuleName() {
		return "RIAM";
	}

}
