/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.iam.riam;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.logging.Logger;

/**
 * Class that breaks up event generation script text into 'tokens'
 */
public class Tokenizer {

	private static Logger log = Logger.getLogger("deira.ram");
	/** < Logging instance */

	private int linenr = 0;
	/** Current line that is tokenized */

	private ArrayList<ArrayList<Token>> rules = new ArrayList<ArrayList<Token>>();

	/** The lines converted to a series of Tokens */

	/**
	 * Constructor
	 */
	public Tokenizer() {
	}

	/**
	 * Reads in and tokenizes textfile.
	 * 
	 * @param textfile
	 *            the rls file to be tokenized.
	 */
	public void load(String textfile) {
		try {
			InputStream ist = new FileInputStream(textfile);
			BufferedReader istream = new BufferedReader(new InputStreamReader(
					ist));
			String text;
			while ((text = istream.readLine()) != null) {
				linenr++;
				// one rule per line
				text = text.trim();
				if (text.length() > 0 && (text.charAt(0) != '/')) {
					// iow: there has to be something in the rule
					// and commentary should be skipped
					ArrayList<Token> rule = tokenize(text);
					rules.add(rule);
				}
			}
		} catch (FileNotFoundException e) {
			log.severe("Tokenizer could not find " + textfile);
		} catch (IOException e) {
			log.severe("Tokenizer could not read from " + textfile);
		}
	}

	/**
	 * Returns the tokenized lines which were the result of loading and parsing
	 * the given textfile.
	 * 
	 * @return The tokenized lines as parsed when loading the given textfile.
	 */
	public ArrayList<ArrayList<Token>> getLines() {
		return rules;
	}

	/**
	 * Splits a string (typically one rule line) into tokens and returns the
	 * resulting Token ArrayList.
	 * 
	 * @param s
	 *            A string to be tokenized
	 * @return The series of Tokens resulting from tokenizing s.
	 */
	public ArrayList<Token> tokenize(String s) {
		char[] chars = s.toCharArray();
		ArrayList<Token> tokens = new ArrayList<Token>();
		String id = "";
		String num = "";

		for (int i = 0; i < chars.length; i++) {
			String cs = String.valueOf(chars[i]);

			String peek = ";";
			if (i + 1 < chars.length)
				peek = String.valueOf(chars[i + 1]);

			// TODO string literals ??
			// log.info(cs);
			if (Character.isWhitespace(cs.charAt(0)))
				; // ignore
			else if (Token.LPAREN.equals(cs)) {
				if (id != "")
					tokens.add(new Token(id, Token.IDENTIFIER, linenr, i + 1));
				else if (num != "")
					tokens.add(new Token(num, Token.NUMBER, linenr, i + 1));
				id = "";
				num = "";
				tokens.add(new Token(cs, Token.LPAREN, linenr, i + 1));
			} else if (Token.RPAREN.equals(cs)) {
				if (id != "")
					tokens.add(new Token(id, Token.IDENTIFIER, linenr, i + 1));
				else if (num != "")
					tokens.add(new Token(num, Token.NUMBER, linenr, i + 1));
				id = "";
				num = "";
				tokens.add(new Token(cs, Token.RPAREN, linenr, i + 1));
			} else if (Token.EQ.equals(cs + peek)) {
				if (id != "")
					tokens.add(new Token(id, Token.IDENTIFIER, linenr, i + 1));
				else if (num != "")
					tokens.add(new Token(num, Token.NUMBER, linenr, i + 1));
				id = "";
				num = "";
				tokens.add(new Token(cs + peek, Token.EQ, linenr, i + 1));
				i++;
			} else if (Token.NE.equals(cs + peek)) {
				if (id != "")
					tokens.add(new Token(id, Token.IDENTIFIER, linenr, i + 1));
				else if (num != "")
					tokens.add(new Token(num, Token.NUMBER, linenr, i + 1));
				id = "";
				num = "";
				tokens.add(new Token(cs + peek, Token.NE, linenr, i + 1));
				i++;
			} else if (Token.GTE.equals(cs + peek)) {
				if (id != "")
					tokens.add(new Token(id, Token.IDENTIFIER, linenr, i + 1));
				else if (num != "")
					tokens.add(new Token(num, Token.NUMBER, linenr, i + 1));
				id = "";
				num = "";
				tokens.add(new Token(cs + peek, Token.GTE, linenr, i + 1));
				i++;
			} else if (Token.LTE.equals(cs + peek)) {
				if (id != "")
					tokens.add(new Token(id, Token.IDENTIFIER, linenr, i + 1));
				else if (num != "")
					tokens.add(new Token(num, Token.NUMBER, linenr, i + 1));
				id = "";
				num = "";
				tokens.add(new Token(cs + peek, Token.LTE, linenr, i + 1));
				i++;
			} else if (Token.GT.equals(cs)) {
				if (id != "")
					tokens.add(new Token(id, Token.IDENTIFIER, linenr, i + 1));
				else if (num != "")
					tokens.add(new Token(num, Token.NUMBER, linenr, i + 1));
				id = "";
				num = "";
				tokens.add(new Token(cs, Token.GT, linenr, i + 1));
			} else if (Token.LT.equals(cs)) {
				if (id != "")
					tokens.add(new Token(id, Token.IDENTIFIER, linenr, i + 1));
				else if (num != "")
					tokens.add(new Token(num, Token.NUMBER, linenr, i + 1));
				id = "";
				num = "";
				tokens.add(new Token(cs, Token.LT, linenr, i + 1));
			} else if (Token.RESULTS.equals(cs + peek)) {
				if (id != "")
					tokens.add(new Token(id, Token.IDENTIFIER, linenr, i + 1));
				else if (num != "")
					tokens.add(new Token(num, Token.NUMBER, linenr, i + 1));
				id = "";
				num = "";
				tokens.add(new Token(cs + peek, Token.RESULTS, linenr, i + 1));
				i++;
			} else if (Token.AND.equals(cs + peek)) {
				if (id != "")
					tokens.add(new Token(id, Token.IDENTIFIER, linenr, i + 1));
				else if (num != "")
					tokens.add(new Token(num, Token.NUMBER, linenr, i + 1));
				id = "";
				num = "";
				tokens.add(new Token(cs + peek, Token.AND, linenr, i + 1));
				i++;
			} else if (Token.OR.equals(cs + peek)) {
				if (id != "")
					tokens.add(new Token(id, Token.IDENTIFIER, linenr, i + 1));
				else if (num != "")
					tokens.add(new Token(num, Token.NUMBER, linenr, i + 1));
				id = "";
				num = "";
				tokens.add(new Token(cs + peek, Token.OR, linenr, i + 1));
				i++;
			} else if (Token.PLUS.equals(cs)) {
				if (id != "")
					tokens.add(new Token(id, Token.IDENTIFIER, linenr, i + 1));
				else if (num != "")
					tokens.add(new Token(num, Token.NUMBER, linenr, i + 1));
				id = "";
				num = "";
				tokens.add(new Token(cs, Token.PLUS, linenr, i + 1));
			} else if (Token.MINUS.equals(cs)) {
				if (id != "")
					tokens.add(new Token(id, Token.IDENTIFIER, linenr, i + 1));
				else if (num != "")
					tokens.add(new Token(num, Token.NUMBER, linenr, i + 1));
				id = "";
				num = "";
				tokens.add(new Token(cs, Token.MINUS, linenr, i + 1));
			} else if (Token.MULT.equals(cs)) {
				if (id != "")
					tokens.add(new Token(id, Token.IDENTIFIER, linenr, i + 1));
				else if (num != "")
					tokens.add(new Token(num, Token.NUMBER, linenr, i + 1));
				id = "";
				num = "";
				tokens.add(new Token(cs, Token.MULT, linenr, i + 1));
			} else if (Token.DIV.equals(cs)) {
				if (id != "")
					tokens.add(new Token(id, Token.IDENTIFIER, linenr, i + 1));
				else if (num != "")
					tokens.add(new Token(num, Token.NUMBER, linenr, i + 1));
				id = "";
				num = "";
				tokens.add(new Token(cs, Token.DIV, linenr, i + 1));
			} else if (Token.SEMICOLON.equals(cs)) {
				if (id != "")
					tokens.add(new Token(id, Token.IDENTIFIER, linenr, i + 1));
				else if (num != "")
					tokens.add(new Token(num, Token.NUMBER, linenr, i + 1));
				id = "";
				num = "";
				tokens.add(new Token(cs, Token.SEMICOLON, linenr, i + 1));
			} else if (Token.COMMA.equals(cs)) {
				if (id != "")
					tokens.add(new Token(id, Token.IDENTIFIER, linenr, i + 1));
				else if (num != "")
					tokens.add(new Token(num, Token.NUMBER, linenr, i + 1));
				id = "";
				num = "";
				tokens.add(new Token(cs, Token.COMMA, linenr, i + 1));
			} else if (Token.DOT.equals(cs)) {
				if (num != "")
					num += cs;
				else {
					if (id != "")
						tokens.add(new Token(id, Token.IDENTIFIER, linenr,
								i + 1));
					else if (num != "")
						tokens.add(new Token(num, Token.NUMBER, linenr, i + 1));
					id = "";
					num = "";
					tokens.add(new Token(cs, Token.DOT, linenr, i + 1));
				}
			} else if (Character.isDigit(cs.charAt(0))) {
				if (id != "")
					id += cs;
				else
					num += cs;
			} else {
				id += cs;
			}
		}

		return tokens;
	}

}
