/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.iam.vsiam;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.text.MessageFormat;
import java.util.Iterator;
import java.util.logging.Logger;

import natlang.rdg.documentplanner.DocumentPlanner;
import natlang.rdg.lexicalchooser.LexicalChooser;
import natlang.rdg.microplanner.MicroPlanner;
import natlang.rdg.model.RSGraph;
import natlang.rdg.surfacerealizer.SurfaceRealizer;
import nl.utwente.ewi.hmi.deira.generic.DeiraThread;
import nl.utwente.ewi.hmi.deira.generic.Module;
import nl.utwente.ewi.hmi.deira.mmm.MMM;
import nl.utwente.ewi.hmi.deira.queue.VSEvent;

public class VSAnalysis extends DeiraThread implements Module {

	private int eventCounter = 0;
	/** Event counter to maintain a unique event id for dispatch. */
	private nl.utwente.ewi.hmi.deira.mmm.MMM mmm;
	public boolean startStory;
	private boolean run = true;
	/** Mental model module to which the events need to be passed on. */

	private static Logger log = Logger.getLogger("deira.vsiam");

	/** < Logging instance */

	/**
	 * @param mmm
	 */
	public VSAnalysis(MMM mmm) {
		super("DEIRA VSIAM");
		this.mmm = mmm;
	}

	/**
	 * Returns a fresh unused event id. (increases eventCounter by 1 to do this)
	 * 
	 * @return a fresh unused event id.
	 */
	public int getNewEventID() {
		eventCounter++;
		return eventCounter;
	}

	/**
     */
	public void close() {
		run = false;
	}

	/**
	 */
	public void run() {
		runLoop: while (run) {

			while (!this.startStory) {
				synchronized (this) {
					try {
						this.wait(1000);
						if (!run) {
							break runLoop;
						}
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
			}

			//=-- FK TODO: Temporary switch --> should be GUI-driven.
			int mode = 1;
			
			switch (mode) {
			case 0:
				this.runNarratorProcessing();
				break;
			case 1:
				this.runFileProcessing("anansi_berijdt_tijger.txt");
				break;
				
			}

			this.startStory = false;
		}
	}

	private void runNarratorProcessing() {
		boolean dpresult = true;
		DocumentPlanner documentplanner = new DocumentPlanner();
		MicroPlanner microplanner = new MicroPlanner();

		long time = 0;
		float importance = 0.6f;
		float decay = -1.0f; //=-- The longer the event is in the queue, the more important it becomes.

		// =-- Get next paragraph from documentplanner
		dpresult = documentplanner.transform();
		if (dpresult) {
			RSGraph dpoutput = documentplanner.getGraph();

			// =-- Do microplanning for paragraph
			microplanner.setGraph(dpoutput);
			RSGraph mcoutput = microplanner.transform();

			// =-- Add OntModel for surfacerealizer
			com.hp.hpl.jena.ontology.OntModel ontmodel = documentplanner.getModel();

			LexicalChooser lc = microplanner.getLexicalChooser();
			SurfaceRealizer sr = new SurfaceRealizer(ontmodel);

			// =-- Initialize parameters for surface realization
			sr.setGraph(mcoutput);
			sr.setLexicalChooser(lc);

			// =-- Generate events
			synchronized (this) {
				String currentitem;
				try {
					if (sr.transform()) {
						Iterator<String> it = sr.getResult();
						while (it.hasNext() && run) {
							currentitem = it.next();
							time = System.currentTimeMillis();
							VSEvent sentence = new VSEvent(currentitem, eventCounter, time, "sentence", "", importance, decay);
							mmm.processEvent(sentence);
							this.wait(currentitem.length() * 50);
						}
					} else {
						log.severe("VSIAM: Surface transformation failed!");
					}
				} catch (Exception e) {
					log.severe("VSIAM: Surface transformation failed!");
					e.printStackTrace();
				}
			}

		}
	}

	private void runFileProcessing(String inputFile) {
		long time = 0;
		float importance = 0.6f;
		float decay = -1.0f; // =-- The longer the event is in the queue, the more important it becomes.

		// =-- Generate events
		String currentitem;
		synchronized (this) {

			try {
				BufferedReader reader = new BufferedReader(new FileReader(inputFile));
				currentitem = reader.readLine();
				while (currentitem != null && run) {
					if (!currentitem.equals("")) {
						time = System.currentTimeMillis();
						VSEvent sentence = new VSEvent(currentitem, eventCounter, time, "sentence", "", importance, decay);
						mmm.processEvent(sentence);
						this.wait(currentitem.length() * 50);
					}

					currentitem = reader.readLine();
				}
			} catch (IOException e) {
				log.severe(MessageFormat.format("VSIAM: Read from file [{0}] failed!", inputFile));
			} catch (InterruptedException e) {
				log.severe("VSIAM: Thread interrupted!");
			}
		}
	}

	public String getModuleName() {
		return "VSIAM";
	}

}
