package nl.utwente.ewi.hmi.deira.iam.vvciam.match;


import java.io.BufferedReader;
import java.io.IOException;
import java.util.Observable;
import java.util.Observer;

import nl.utwente.ewi.hmi.deira.iam.vvciam.Situation;


/**
 * Een Robosoccermatch tussen 2 teams.
 * Van een match wordt bijgehouden: welke tijd er op de teller staat en wat de stand is.
 * @version
 * @author Buursma, Hendriksen, Ten Hoeve, Van der Kooi
 */
public class MatchTracker extends Observable implements Observer {

	private Team team1;
	private Team team2;
	private int currenttime;
	private int[] score={0,0};
	private int framenumber;
	private int lastTeamScored;

    /** Constructor van de klasse Match.
     * Teaminfo (naam, namenspelers) wordt uitgelezen uit het gegeven bestand. doelpunten en tijd worden op 0 geinitialiseerd.
     * @param filename naam (pad+bestandsnaam) van een correct geformateerde MatchInfo bestand. 
     * @throws IOException Bij I/O problemen met het uitlezen van het MatchInfo bestand.
     */
	public MatchTracker(String filename) throws IOException{
	    	readFile(filename);
		currenttime = 0;
		framenumber = 0;
		lastTeamScored = -1;
	}
	
    private void readFile(String filename) throws IOException{
        try{
            BufferedReader in = null;
            in = new BufferedReader(new java.io.FileReader(filename));
            Team[] teams=new Team[2];
            for(int t=0; t< 2; t++){
                String teamName = in.readLine();
                Robot[] robots=new Robot[5];
                for(int i=0; i < 5; i++){
                    String str=in.readLine();
                    int robotId=Integer.valueOf(str.substring(0, str.indexOf(',')));
                    String robotName=str.substring(str.indexOf(',')+1);
                    robots[i]=new Robot(robotName ,robotId);
                }
                teams[t]= new Team(teamName, robots[0], robots[1], robots[2], robots[3],robots[4]);
            }
            this.team1=teams[0];
            this.team2=teams[1];
            in.close();
        } catch (IOException e){
            throw new IOException("Matchfile IOException: " + e.toString());    
        }    
    }
		
	/**
	 * Methode om de tijd bij te houden
	 * als er een frame wordt verwerkt dan kijkt deze methode, hoeveel frames er al zijn geweest.
	 */
	public void updateTime(){
		framenumber++;
		if (framenumber == nl.utwente.ewi.hmi.deira.iam.vvciam.FileReader.FPS){
			currenttime++;
			framenumber = 0;
			setChanged();
			notifyObservers(null);
		}
	}
	
	/**
	 * Er wordt een doelpunt gemaakt en dus moet de score bijgewerkt worden
	 * @param i, het scorende team, als i = 0 dan is het team a die heeft gescoord, bij i = 1, dan heeft teamb gescoord.
	 * @ensure i = 0 || i ==1
	 */	
	public void goal(int i){
        	score[i] += 1;
		lastTeamScored = i;		 
        	setChanged();
	    	notifyObservers(null);

	}
	
	/** 
	 * De methode dient aangeroepen te worden als het laatste doelpunt afgekeurd is.
	 * De laatste goal zal weer van de stand afgetrokken worden.
	 */
	public void goalDisallowed() {
	    if(lastTeamScored != -1){
    		score[lastTeamScored] -= 1;
            	setChanged();
	        notifyObservers(null);
	    }
	}
	
	/** Geeft aantal goals van het eerste team.
	 */	
	public int getScore1(){
	  return score[0];
	}

	/** Geeft aantal goals van het tweede team.
	 */		
	public int getScore2(){	
	  return score[1];	
	}	

	/** Geeft het aantal seconden sinds de start van het commentaar.
	 */		
	public int getTime(){	
	  return currenttime;	
	}
	
	/** Geeft de instatie van Team die het eerste team vertegenwoordigt.
	 */	
	public Team getTeam1(){
	  return team1;	
	}
	
	/** Geeft de instatie van Team die het tweede team vertegenwoordigt.
	 */		
	public Team getTeam2(){
	  return team2;
	}
	
	/** Geeft de instatie van Team gegeven door het team nummer.
	 * @param 0 is eerste team. 1 is tweede team.
	 */		
	public Team getTeam(int teamnr) {
		if (teamnr == 0)
		    return team1;
		else if (teamnr == 1)
		    return team2;
		else 
		    return null;
	}
	
	/**
	 * levert het aantal doelpunten verschil tussen de beide teams op, als gegeven was dat teamScored er nog 1 gescoord heeft.
	 * @param teamScored het team dat net een doelpunt gemaakt heeft
	 */
	public int getGoalDifference(int teamScored) {
	    int tscore[] = {score[0],score[1]};
	    tscore[teamScored]++;
	    return Math.abs(tscore[0]-tscore[1]);
	}
	
	/** Geeft het huidige doelpunten verschil tussen beide teams.
	*/
	public int getGoalDifference() {
	    return Math.abs(score[0]-score[1]);
	}
	
	/** Deze functie dient alleen via het Observer/Observable pattern aangeroepen te worden door een Analyzer.FileReader.
	 * @param obj Situation object met de huidige Situation.
	*/	
	public void update(Observable o, Object obj){		
	  if(obj instanceof Situation){
	    updateTime();
 	  } 			
	}
	
	/** Geeft de huidige score in spreektaal.
	 * @return voorbeeld: "1 nil"
	 */
	public String getScore() {
		String stand = (score[0] > score[1]) ? (score[0] + " " + (score[1] == 0 ? "nil" : score[1])) : ((score[1] == 0 ? "nil" : score[1]) + " " + (score[0] == 0 ? "nil" : score[0]));
		return stand;
	}
	
	/** Geeft de teamnaam van het team met de meeste doelpunten.
	 * @return bij gelijk aantal doelpunten wordt het tweede team gegeven.
	*/
	public String getWinningTeam() {
		return ((score[0] > score[1]) ? team1.getName() : team2.getName());
	}
	
	/** Geeft de teamnaam van het team met de minste doelpunten.
	* @return bij gelijk aantal doelpunten wordt het eerste team gegeven.
	*/	
	public String getLosingTeam() {
		return ((score[0] > score[1]) ? team2.getName() : team1.getName());
	}
}
