/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.mmm;

import java.util.HashMap;
import java.util.Map.Entry;

/**
 * Class to represent the emotional characteristics of a particular reporter.
 * Each emotional aspect has a factor by which emotions are multiplied.
 */
public class EmotionalCharacteristics {
	/** < Emotional factors, represented as doubles ranging from 0 to 5. */
	private HashMap<String, Double> emotionalfactors = new HashMap<String, Double>();

	/**
	 * @param emotionalfactors
	 */
	public EmotionalCharacteristics(HashMap<String, Double> emotionalfactors) {
		super();
		this.emotionalfactors = emotionalfactors;
	}

	/**
	 * Gives the dominant emotional factor
	 * 
	 * @return
	 */
	public String getDominantEmotionalFactor() {
		// =-- Initialize variables
		String dominantfactor = "";
		double max = 0;
		double valueholder;

		// =-- Iterate emotions
		for (Entry<String, Double> currentvalue : this.emotionalfactors
				.entrySet()) {
			valueholder = currentvalue.getValue().doubleValue();
			if (valueholder > max) {
				// =-- Update dominant emotion
				max = valueholder;
				dominantfactor = currentvalue.getKey();
			}
		}
		return dominantfactor;
	}

	/**
	 * Applies the factors to a map of emotions
	 * 
	 * @param emotions
	 * @return the updated emotions
	 */
	public HashMap<String, Double> applyFactors(HashMap<String, Double> emotions) {
		// =-- Initialize return variable
		HashMap<String, Double> updatedemotions = new HashMap<String, Double>();

		// =-- Iterate input map and store new values
		for (Entry<String, Double> currentemotion : emotions.entrySet()) {
			String key = currentemotion.getKey();
			Double emotionalFactor = emotionalfactors.get(key);
			if (emotionalFactor == null) {
				emotionalFactor = 1.0;
			}
			updatedemotions.put(key, currentemotion.getValue()
					* emotionalFactor);
		}

		// =-- Return new values
		return updatedemotions;
	}

	/**
	 * Returns the value for a specific emotion
	 * 
	 * @param emotion
	 * @return
	 */
	public double getEmotionalFactor(String emotion) {
		double emotionalfactor = -1;

		// =-- Check existense of emotion
		if (this.emotionalfactors.containsKey(emotion)) {

			// =-- Set return value to actual value
			emotionalfactor = this.emotionalfactors.get(emotion).doubleValue();
		}

		return emotionalfactor;
	}

	/**
	 * @return Returns double Tension factor.
	 * @deprecated
	 */
	public double getTensionFactor() {
		return this.getEmotionalFactor("tension");
	}

	/**
	 * @return Returns double Surprise factor.
	 * @deprecated
	 */
	public double getSurpriseFactor() {
		return this.getEmotionalFactor("surprise");
	}

	/**
	 * @return Returns double Amusement factor.
	 * @deprecated
	 */
	public double getAmusementFactor() {
		return this.getEmotionalFactor("amusement");
	}

	/**
	 * @return Returns double Pity factor.
	 * @deprecated
	 */
	public double getPityFactor() {
		return this.getEmotionalFactor("pity");
	}

}