/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.mmm;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Set;
import java.util.Map.Entry;

/**
 * Class to represent an emotional state of a reporter at a particular
 * timestamp.
 */
public class EmotionalState {
	private static final int history_window = 5;
	/**
	 * Emotional levels, represented as doubles ranging from 0 to 1.
	 */
	private HashMap<String, Double> emotions = new HashMap<String, Double>();

	private HashMap<String, ArrayList<Double>> emotions_history = new HashMap<String, ArrayList<Double>>();

	// -----constructors------
	/**
	 * Constructor for EmotionalState objects.
	 * 
	 * @param eventID
	 *            the id of the last processed Event.
	 * 
	 */
	public EmotionalState(HashMap<String, Double> emotions) {
		this.emotions = emotions;
	}

	/**
	 * Constructor of default EmotionalState.
	 */
	public EmotionalState() {
	}

	// ------useful methods-------
	/**
	 * Method to retrieve the most dominant emotion of this EmotionalState.
	 * 
	 * @return the dominant emotion.
	 */
	public String getDominantEmotion() {
		// =-- Initialize variables
		String dominantemotion = "";
		double max = 0;
		double valueholder;

		// =-- Iterate emotions
		Set<Entry<String, Double>> emotionalvalues = this.emotions.entrySet();
		for (Entry<String, Double> currentvalue : emotionalvalues) {
			valueholder = currentvalue.getValue().doubleValue();
			if (valueholder > max) {
				// =-- Update dominant emotion
				max = valueholder;
				dominantemotion = currentvalue.getKey();
			}
		}
		return dominantemotion;
	}

	// ------only general methods from here--------
	/**
	 * Returns the value for a specific emotion
	 * 
	 * @param emotion
	 * @return
	 */
	public double getEmotionalLevel(String emotion) {
		double emotionallevel = -1;

		// =-- Check existense of emotion
		if (this.emotions.containsKey(emotion)) {

			// =-- Set return value to actual value
			emotionallevel = this.emotions.get(emotion).doubleValue();
		}

		return emotionallevel;
	}

	/**
	 * Returns the value for a specific emotion
	 * 
	 * @param emotion
	 * @return
	 */
	public double getAveragedEmotionalLevel(String emotion) {
		double emotionallevel = -1;

		// =-- TODO Iterate relevant list
		double sum = 0;
		ArrayList<Double> emotion_history = emotions_history.get(emotion);

		if (emotion_history != null) {
			for (Double value : emotion_history) {
				sum += value;
			}
			emotionallevel = sum / emotion_history.size();
		}

		return emotionallevel;
	}

	/**
	 * @return Returns double Tension level.
	 * @deprecated
	 */
	public double getTension() {
		return this.getEmotionalLevel("tension");
	}

	/**
	 * @return Returns double Surprise level.
	 * @deprecated
	 */
	public double getSurprise() {
		return this.getEmotionalLevel("surprise");
	}

	/**
	 * @return Returns double Amusement level.
	 * @deprecated
	 */
	public double getAmusement() {
		return this.getEmotionalLevel("amusement");
	}

	/**
	 * @return Returns double Pity level.
	 * @deprecated
	 */
	public double getPity() {
		return this.getEmotionalLevel("pity");
	}

	public void updateState(HashMap<String, Double> aggregateemotions) {

		this.emotions = aggregateemotions;
		for (Entry<String, Double> entry : aggregateemotions.entrySet()) {
			updateHistory(entry.getKey(), entry.getValue());
		}

	}

	private void updateHistory(String key, Double value) {
		ArrayList<Double> list = emotions_history.get(key);
		if (list == null) {
			list = new ArrayList<Double>();
			emotions_history.put(key, list);
		}

		if (list.size() >= history_window) {
			// =-- Remove oldest element
			list.remove(0);
		}
		list.add(value);
	}

	public HashMap<String, Double> getEmotionalLevels() {

		return emotions;
	}
}