/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.mmm;

import java.util.HashMap;
import java.util.Set;
import java.util.Map.Entry;

/**
 * Class to represent an emotion for a particular Event.
 */
public class EventEmotion {
	private int eventID;
	/** < A unique ID of the corresponding Event (generated by the RAM). */
	private long timeStamp;
	/**
	 * < A timestamp of the moment of creation of this EventEmotion in
	 * milliseconds.
	 */
	private int duration;
	/**
	 * < The duration of this EventEmotion. The number of milliseconds in which
	 * the EventEmotion decays.
	 */
	private HashMap<String, Double> emotions /**
	 * < Emotional levels, represented
	 * as doubles ranging from 0 to 1.
	 */
	;

	// -----constructors------

	/**
	 * EventEmotion constructor for 1st iteration.
	 * 
	 * @param eventID
	 *            the id of the corresponding Event.
	 * @param duration
	 *            the number of milliseconds in which the EventEmotion decays.
	 * @param tension
	 *            the tension level (range from 0 to 1).
	 */
	public EventEmotion(int eventID, int duration,
			HashMap<String, Double> emotions) {
		this.eventID = eventID;
		this.timeStamp = System.currentTimeMillis();
		this.duration = duration;
		this.emotions = emotions;

	}

	/**
	 * @return Returns int corresponding eventID.
	 */
	public int getEventID() {
		return this.eventID;
	}

	/**
	 * @return Returns the timestamp of this EmotionalState.
	 */
	public long getTimeStamp() {
		return this.timeStamp;
	}

	/**
	 * @return Returns int Duration.
	 */
	public int getDuration() {
		return this.duration;
	}

	// ------useful methods-------
	/**
	 * Method to retrieve the most dominant emotion of this EmotionalState.
	 * 
	 * @return the dominant emotion.
	 */
	public String getDominantEmotion() {
		// =-- Initialize variables
		String dominantemotion = "";
		double max = 0;
		double valueholder;

		// =-- Iterate emotions
		Set<Entry<String, Double>> emotionalvalues = this.emotions.entrySet();
		for (Entry<String, Double> currentvalue : emotionalvalues) {
			valueholder = currentvalue.getValue().doubleValue();
			if (valueholder > max) {
				// =-- Update dominant emotion
				max = valueholder;
				dominantemotion = currentvalue.getKey();
			}
		}
		return dominantemotion;
	}

	/**
	 * Returns the value for a specific emotion
	 * 
	 * @param emotion
	 * @return
	 */
	public double getEmotionalLevel(String emotion) {
		double emotionallevel = -1;

		// =-- Check existense of emotion
		if (this.emotions.containsKey(emotion)) {

			// =-- Set return value to actual value
			emotionallevel = this.emotions.get(emotion).doubleValue();
		}

		return emotionallevel;
	}

	public HashMap<String, Double> getEmotions() {
		return emotions;
	}

	/**
	 * @return Returns double Tension level.
	 * @deprecated
	 */
	public double getTension() {
		return this.getEmotionalLevel("tension");
	}

	/**
	 * @return Returns double Surprise level.
	 * @deprecated
	 */
	public double getSurprise() {
		return this.getEmotionalLevel("surprise");
	}

	/**
	 * @return Returns double Amusement level.
	 * @deprecated
	 */
	public double getAmusement() {
		return this.getEmotionalLevel("amusement");
	}

	/**
	 * @return Returns double Pity level.
	 * @deprecated
	 */
	public double getPity() {
		return this.getEmotionalLevel("pity");
	}

}