/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.mmm;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map.Entry;
import java.util.logging.Logger;

import nl.utwente.ewi.hmi.deira.db.PersonalityDB;
import nl.utwente.ewi.hmi.deira.queue.Event;
import nl.utwente.ewi.hmi.deira.queue.EventQueue;
import nl.utwente.ewi.hmi.deira.queue.HREvent;
import nl.utwente.ewi.hmi.deira.queue.VSEvent;
import nl.utwente.ewi.hmi.deira.queue.VVCEvent;

/**
 * Class to represent the mental model module.
 */
public class MMM {
	private PersonalityDB pdb;
	/**
	 * < The personality database of the reporter. From this, the emotional
	 * characteristics will be retrieved whenever needed.
	 */
	private EventQueue eq;
	/** < The EventQueue of the system. */
	private int lastProcessedEventID;
	/** < The last eventID processed by the MMM. */
	private EmotionalState es;
	/** < The emotional state of the reporter. */
	private ArrayList<EventEmotion> emotions;
	/** < A ArrayList with all generated EventEmotion objects. */
	private ArrayList<EventEmotion> activeEmotions;
	/** < A ArrayList with all currently active EventEmotion objects. */
	private static Logger log = Logger.getLogger("deira.mmm");
	/** < Logging instance */
	private MMMEvent mmmEvent;

	/** < MMM Event callback */

	// -----constructor------
	/**
	 * Constructor for Mental Model Module.
	 * 
	 * @param pdb
	 *            reference to the personality database.
	 * @param eq
	 *            reference to the eventqueue.
	 */
	public MMM(PersonalityDB pdb, EventQueue eq, MMMEvent mmmEvent) {
		this.pdb = pdb;
		this.eq = eq;
		this.mmmEvent = mmmEvent;

		this.es = new EmotionalState();
		this.emotions = new ArrayList<EventEmotion>();
		this.activeEmotions = new ArrayList<EventEmotion>();
	}

	// ------public methods--------
	/**
	 * Method to be called by Race Analysis Module (RAM).
	 * 
	 * @param re
	 *            the new RaceEvent.
	 */
	public void processEvent(Event re) {
		// process the RaceEvent (calculate and store EventEmotion)
		EventEmotion emo = this.generateEmotion(re);
		this.emotions.add(emo);
		this.activeEmotions.add(emo);

		// put the RaceEvent with the generated EventEmotion in the EventQueue
		re.tag("emotion", emo);
		this.eq.add(re);

		// store last processed eventID
		this.lastProcessedEventID = re.getId();

		// update and backup internal emotional state
		this.updateEmotionalState();
	}

	/**
	 * Returns the id of the last processed Event.
	 * 
	 * @return the id of the last processed Event.
	 */
	public int getLastProcessedEventID() {
		return this.lastProcessedEventID;
	}

	/**
	 * Returns an EventEmotion for a particular eventID.
	 * 
	 * @param eventID
	 *            intvalue eventID.
	 * @return The EventEmotion for the Event specified by eventID.
	 * @return null if the emotion is not found.
	 */
	public EventEmotion getEmotion(int eventID) {
		// search for emotion in this.emotions with corresponding eventID
		Iterator<EventEmotion> it = this.emotions.iterator();
		while (it.hasNext()) {
			EventEmotion emo = (EventEmotion) it.next();
			if (emo.getEventID() == eventID)
				return emo;
		}
		// if not found
		return null;
	}

	/**
	 * @return the current EmotionalState.
	 */
	public EmotionalState getEmotionalState() {
		return this.es;
	}

	/**
	 * Calculates an EventEmotion for an Event.
	 * 
	 * @param event
	 *            the Event.
	 * @return EventEmotion for Event.
	 */
	private EventEmotion generateEmotion(Event event) {

		// =-- Variable declaration
		int duration;
		HashMap<String, Double> emotions;
		EmotionalTagger tagger;

		// =-- Determine tagger class
		if (event instanceof HREvent) {
			tagger = new HREmotionalTagger();
		} else if (event instanceof VSEvent) {
			tagger = new VSEmotionalTagger();
		} else if (event instanceof VVCEvent) {
			tagger = new VVCEmotionalTagger();
		} else {
			tagger = new EmotionalTagger();
		}

		// =-- Create emotional values
		tagger.processEvent(event);
		emotions = tagger.getGeneratedEmotions();
		duration = tagger.getDuration();

		// =-- Apply emotional disposition
		EmotionalCharacteristics ec = pdb.getEmotionalCharacteristics();
		emotions = ec.applyFactors(emotions);

		// =-- Make sure emotions aren't out of bounds
		emotions = doEmotionCutoff(emotions);

		log.info("MMM.generateEmotion(" + event.toString() + ")");

		return new EventEmotion(event.getId(), duration, emotions);
	}

	/**
	 * Cuts off emotions at 0 and 1.
	 * 
	 * @param emotions
	 * @return
	 */
	private HashMap<String, Double> doEmotionCutoff(
			HashMap<String, Double> emotions) {
		// =-- Initialize return variable
		HashMap<String, Double> updatedemotions = new HashMap<String, Double>();

		// =-- Iterate input map and store new values
		for (Entry<String, Double> currentemotion : emotions.entrySet()) {
			String key = currentemotion.getKey();
			Double value = currentemotion.getValue();
			if (value.doubleValue() > 1) {
				value = new Double(1);
			} else if (value.doubleValue() < 0) {
				value = new Double(0);
			}
			updatedemotions.put(key, value);
		}

		// =-- Return new values
		return updatedemotions;
	}

	/**
	 * Updates the internal EmotionalState of the MMM.
	 */
	private void updateEmotionalState() {
		// calculate the EmotionalState as follows:
		// To calculate the emotional state of a reporter at a particular
		// moment,
		// all emotions that were generated so far during the race are iterated.
		// (or only the active ones)
		// The levels of all emotional aspects are decreased relatively to the
		// duration time of the emotion.
		// The duration is the time in which the entire EventEmotion completely
		// decays.
		// After considering the duration, the remainders are added together and
		// result in the current emotional state.

		HashMap<String, Double> aggregateemotions = new HashMap<String, Double>();

		// iterate over all currently active Emotions
		Iterator<EventEmotion> it = this.activeEmotions.iterator();
		while (it.hasNext()) {
			EventEmotion emo = (EventEmotion) it.next();
			int duration = emo.getDuration();
			HashMap<String, Double> emotions = emo.getEmotions();

			long stamp = emo.getTimeStamp();
			long current = System.currentTimeMillis();
			long diff = current - stamp;
			double decreaseFactor = ((double) diff) / ((double) duration);

			if (decreaseFactor > 1) {
				// remove this decayed EventEmotion
				it.remove();
			} else {
				// adding it all up
				for (Entry<String, Double> currentemotion : emotions.entrySet()) {
					String key = currentemotion.getKey();
					double value = currentemotion.getValue();
					value -= value * decreaseFactor;
					if (aggregateemotions.containsKey(key)) {
						value += aggregateemotions.get(key);
					}
					aggregateemotions.put(key, value);
				}

			}
		}

		// adding a lot of things up could make a particular level > 100% so we
		// should ceil to 1
		aggregateemotions = doEmotionCutoff(aggregateemotions);

		log.info("MMM.updateEmotionalState()");

		this.es.updateState(aggregateemotions);
		if (this.mmmEvent != null) {
			this.mmmEvent.OnEmotionStateUpdate(es);
		}
	}

}
