/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.om;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.Socket;
import java.nio.CharBuffer;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.logging.Logger;

/**
 * Establishes and maintains an output link with a VisageLink.
 */
public class HaptekOutputLink {
	private static int MAX_READ_LENGTH = 4000;
	/** < Buffer size used for reading out server responses */
	private Socket socket;
	/** < Low-level socket connection to the server */
	private BufferedReader input;
	/** < Input stream attached to the socket */
	private BufferedWriter output;
	/** < Output stream attached to the socket */
	private ArrayList<String> inputStrings;
	/** < Pending input strings read from the socket */
	private String animationDir;
	private HashMap<String, Integer> voiceMap = new HashMap<String, Integer>();
	private static Logger log = Logger.getLogger("deira.om");
	/** < Logging instance */
	private static HashMap<String, String> animationTranslationTable = new HashMap<String, String>();
	private static HashMap<String, String> customAnimations = new HashMap<String, String>();

	private double min_speed = 0.8;
	private double max_speed = 1.4;
	private double min_pitch = 100.0;
	private double max_pitch = 240.0;
	private double min_volume = 60;
	private double max_volume = 200;

	{
		initAnimationTables();
	}

	/**
	 * Sets up and initializes a link to an HaptekPlayer. Note: Remember to call
	 * close() if you are done with the OutputLink.
	 * 
	 * @param host
	 *            Hostname (usually localhost or 127.0.0.1).
	 * @param port
	 *            Port to connect to (HaptekPlayer default is 50011).
	 */
	HaptekOutputLink(String host, int port, String animationDir)
			throws Exception {
		/* Open Connection, Set up reader/writer */
		try {
			log.info("Opening Connection to '" + host + ":" + port + "'");
			this.socket = new Socket(host, port);

			/*
			 * There is one reader and two types of writers. One for writing
			 * text and one for writing pure binary. The latter is used for
			 * sending WAV files over the socket.
			 */
			log.finer("Creating buffered reader & writers");
			this.input = new BufferedReader(new InputStreamReader(socket
					.getInputStream()));
			this.output = new BufferedWriter(new OutputStreamWriter(socket
					.getOutputStream()));
			this.inputStrings = new ArrayList<String>();
			this.animationDir = animationDir;

			loadVoiceMap();

		} catch (Exception e) {
			log.severe("Error setting up connection: " + e);
			close(); // Force a close, to clean-up, also sets the link state
			throw e;
		}
	}

	private void loadVoiceMap() {
		try {
			BufferedReader voiceMapReader = new BufferedReader(new FileReader(
					"hapteklink\\voice_map.ini"));

			String currentline;
			String name;
			int index;
			while (voiceMapReader.ready()) {
				currentline = voiceMapReader.readLine();
				int indexOfEqualSign = currentline.indexOf('=');
				name = currentline.substring(0, indexOfEqualSign).trim();
				index = Integer.parseInt(currentline.substring(
						indexOfEqualSign + 1).trim());
				voiceMap.put(name, index);
			}

		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
		}

	}

	private static void initAnimationTables() {
		animationTranslationTable.put("pity", "pity.hap");
		animationTranslationTable.put("pity_intense", "pity_intense.hap");

		animationTranslationTable.put("anger", "anger.hap");
		animationTranslationTable.put("anger_intense", "anger_intense.hap");

		animationTranslationTable.put("sadness", "sadness.hap");
		animationTranslationTable.put("sadness_intense", "sadness_intense.hap");

		animationTranslationTable.put("surprise", "surprise.hap");
		animationTranslationTable.put("surprise_short", "surpriseshort.hap");
		animationTranslationTable
				.put("surprise_intense", "surpriseintense.hap");
		animationTranslationTable.put("surprise_intense_short",
				"surpriseintenseshort.hap");

		animationTranslationTable.put("happy", "hap_happy.hap");
		animationTranslationTable.put("smile_open_1s", "smile_open_1s.hap");
		animationTranslationTable.put("smile_open_3s", "smile_open_3s.hap");
		animationTranslationTable.put("smile_closed_1s", "smile_closed_1s.hap");
		animationTranslationTable.put("smile_closed_3s", "smile_closed_3s.hap");

		animationTranslationTable.put("tension_low", "tension_level1-2.hap");
		animationTranslationTable.put("tension_medium", "tension_level2-4.hap");
		animationTranslationTable.put("tension_high", "tension_level4-5.hap");

		animationTranslationTable.put("look_touser_18s", "looktouser18s.hap");
		animationTranslationTable.put("look_touser_11s", "looktouser11s.hap");

		animationTranslationTable.put("look_left", "lookleft.hap");
		animationTranslationTable.put("look_right", "lookright.hap");
		animationTranslationTable.put("look_up", "lookup.hap");
		animationTranslationTable.put("look_down", "lookdown.hap");

		animationTranslationTable.put("blink_1", "blink.hap");
		animationTranslationTable.put("blink_2", "blink2.hap");
		animationTranslationTable.put("blink_3", "blink3.hap");

		animationTranslationTable.put("headmotion_verylow", "headmotion1.hap");
		animationTranslationTable.put("headmotion_low", "headmotion2.hap");
		animationTranslationTable.put("headmotion_medium", "headmotion3.hap");
		animationTranslationTable.put("headmotion_high", "headmotion4.hap");

		animationTranslationTable.put("glance_up_right", "glance_up_right.hap");
		animationTranslationTable.put("glance_up_middle",
				"glance_up_middle.hap");
		animationTranslationTable.put("glance_up_left", "glance_up_left.hap");

		animationTranslationTable.put("glance_down_right",
				"glance_down_right.hap");
		animationTranslationTable.put("glance_down_middle",
				"glance_down_middle.hap");
		animationTranslationTable.put("glance_down_left",
				"glance_down_left.hap");

		double base = 0.1;
		double attack = 0.5;
		double decay = 0.7;
		String crowdLookString = "[NAC]\\SetSwitch [switch= lookright state= a f0= {0} f1= {1} f2= {2}]\n";

		// =-- FIXME Somewhat obsolete due to parameter functionality.
		customAnimations.put("look_right5", MessageFormat.format(
				crowdLookString, (base * 5.0), attack, decay));
		customAnimations.put("look_right4", MessageFormat.format(
				crowdLookString, (base * 4.0), attack, decay));
		customAnimations.put("look_right3", MessageFormat.format(
				crowdLookString, (base * 3.0), attack, decay));
		customAnimations.put("look_right2", MessageFormat.format(
				crowdLookString, (base * 2.0), attack, decay));
		customAnimations.put("look_right1", MessageFormat.format(
				crowdLookString, (base * 1.0), attack, decay));
		customAnimations.put("look_center", MessageFormat.format(
				crowdLookString, (base * 0.0 + 0.001), attack, decay));
		customAnimations.put("look_left1", MessageFormat.format(
				crowdLookString, (base * -1.0), attack, decay));
		customAnimations.put("look_left2", MessageFormat.format(
				crowdLookString, (base * -2.0), attack, decay));
		customAnimations.put("look_left3", MessageFormat.format(
				crowdLookString, (base * -3.0), attack, decay));
		customAnimations.put("look_left4", MessageFormat.format(
				crowdLookString, (base * -4.0), attack, decay));
		customAnimations.put("look_left5", MessageFormat.format(
				crowdLookString, (base * -5.0), attack, decay));

		customAnimations
				.put(
						"glance_stop",
						"[NAC]\\clock [t= {duration}] \\SetSwitch [switch= lookdown state= off f1= 0.1]\n[NAC]\\clock [t= {duration}] \\SetSwitch [switch= lookleft state= off f1= 0.1]\n");
		customAnimations
				.put(
						"glance_stop2",
						"[NAC]\\SetSwitch [switch= lookdown state= off f1= 0.1]\n[NAC]\\SetSwitch [switch= lookup state= off f1= 0.1]\n[NAC]\\SetSwitch [switch= lookleft state= off f1= 0.1]\n");

	}

	/**
	 * Sends basic setup information to HaptekLink.
	 * 
	 * @param string
	 * 
	 * @return
	 */

	public boolean sendSetup(String face, String voice) throws Exception {
		assert this.socket != null; /*
									 * If this fails close() probably has been
									 * called
									 */
		boolean success;

		// =---- Language IDs
		// Dutch (NL) 5.1 = 413
		// Dutch (BE) 5.1 = 813
		// English (UK) 5.1 = 809
		// English (US) 5.1 = 409

		// Dutch (NL) 4.0 = 1043
		// Dutch (BE) 4.0 =
		// English (UK) 4.0 = 2057
		// English (US) 4.0 = 1033

		String loadAnimationDir = MessageFormat.format(
				"[NAC]\\PathAdd [name= [{0}]]\n", animationDir);
		log.info("HapOM: Loading base animation: " + loadAnimationDir); // Debug
		this.output.write(loadAnimationDir);

		String modelFile = "torso_male.htr";
		if (face.equals("male")) {
			// modelFile = "torso_male/torso_maleStartup.hap";
			modelFile = "body_male/body_maleStartup.hap";
		} else if (face.equals("female")) {
			modelFile = "torso_female/torso_femaleStartup.hap";
		} else if (face.equals("leno")) {
			modelFile = "leno/Lenostartup.hap";
		}

		// =-- TODO retrieve index based on language, SAPI version and gender.
		// --> Use SSML instead?
		// Real sapi 5.0 index = Sapi 4.0 total indexes + Sapi 5.0 index
		// Claire, Ellen en Laura: 26, 28, 29
		int voiceIndex = voiceMap.get(voice);

		String clearSceneCommand = "[NAC]\\Load [file= [/data/standard/emptyScene.hap]]\n";
		log.info("HapOM: Loading model: " + clearSceneCommand); // Debug
		this.output.write(clearSceneCommand);

		String loadModelCommand = MessageFormat.format(
				"[NAC]\\Load [file= [{0}]]\n", modelFile);
		log.info("HapOM: Loading model: " + loadModelCommand); // Debug
		this.output.write(loadModelCommand);

		String selectVoiceCommand = MessageFormat.format(
				"[NAC]\\SapiTTSLoad [i0= {0}]\n", voiceIndex);
		// String selectVoiceCommand =
		// "[TTS]<voice gender=\"male\" language=\"en-us\" />";
		log.info("HapOM: Selecting voice: " + selectVoiceCommand); // Debug
		this.output.write(selectVoiceCommand);

		String baseAnimationScript = "baseAnimation.hap";
		String loadBaseAnimation = MessageFormat.format(
				"[NAC]\\Load [file= [{0}]]\n", baseAnimationScript);
		log.info("HapOM: Loading base animation: " + loadBaseAnimation); // Debug
		this.output.write(loadBaseAnimation);

		// =-- Load a background
		String background = "stage.jpg";
		String selectBackgroundCommand = MessageFormat.format(
				"[HAP]\\loadbackgrnd [file= {0}]]\n", background);
		log.info("HapOM: Selecting background: " + selectBackgroundCommand); // Debug
		this.output.write(selectBackgroundCommand);

		log.info("HapOM: Flushing stream..."); // 
		this.output.flush();

		success = waitSingleNotify();

		return success;
	}

	/**
	 * Sends a new play order OutputLink.
	 * 
	 * @param text
	 *            The text to be sent over the socket.
	 * @param animations
	 *            ArrayList with animation filenames.
	 * @param speed
	 *            Speech rate
	 * @param pitch
	 *            Pitch of speech
	 * @param volume
	 *            Loudness of speech
	 * @throws Exception
	 */
	public boolean sendAndWait(String text, ArrayList<String> animations,
			int speed, int pitch, int volume) throws Exception {
		assert this.socket != null; /*
									 * If this fails close() probably has been
									 * called
									 */

		String commandText = "";

		// =-- Process animations
		for (String animation : animations) {

			String animationScript = animationTranslationTable.get(animation);
			if (animationScript != null) {

				if (customAnimations.containsKey(animation)) {
					commandText += customAnimations.get(animation);
				} else {
					commandText += MessageFormat.format(
							"[NAC]\\Load [file= [{0}]]\n", animationScript);
				}
			} else {
				log
						.info(MessageFormat
								.format(
										"HapOM: Couldn`t find animation script for animation [{0}]",
										animation));
			}
		}

		double realspeed = min_speed + ((double) speed / 10)
				* (max_speed - min_speed);

		double realpitch = min_pitch + ((double) pitch / 10)
				* (max_pitch - min_pitch);

		double realvolume = min_volume + ((double) volume / 10)
				* (max_volume - min_volume);

		String speedString = "" + realspeed;
		String pitchString = "" + realpitch + "Hz";
		String volumeString = "" + realvolume;

		// =-- Process speech parameters
		text = MessageFormat
				.format(
						"<prosody rate=\"{0}\" pitch=\"{1}\" volume=\"{2}\">{3}</prosody>",
						speedString, pitchString, volumeString, text);

		commandText += MessageFormat.format("[TTS]{0}\n", text);

		log.info("OM: Sending: " + commandText);
		this.output.write(commandText);

		/*
		 * NOTE: This wasn't here before, but generally has a good effect, so
		 * we'll keep it for now
		 */
		// this.output.write(0); // Null terminator for previously send string
		log.info("OM: Flushing stream..."); // 
		this.output.flush();

		return waitSingleNotify();
	}

	public boolean sendAndWaitAnimationsOnly(ArrayList<String> animations)
			throws Exception {
		return this.sendAndWaitAnimationsOnly(animations,
				new HashMap<String, Double>());
	}

	public boolean sendAndWaitAnimationsOnly(ArrayList<String> animations,
			HashMap<String, Double> parameters) throws Exception {
		assert this.socket != null;

		String commandText = "";

		// =-- Process animations
		for (String animation : animations) {

			String animationScript = animationTranslationTable.get(animation);
			if (animationScript != null) {
				commandText += MessageFormat.format(
						"[NAC]\\Load [file= [{0}]]\n", animationScript);

			} else if (customAnimations.containsKey(animation)) {
				String command = customAnimations.get(animation);
				command = this.replaceParameters(command, parameters);
				commandText += command;
			} else {

				log
						.info(MessageFormat
								.format(
										"HapOM: Couldn`t find animation script for animation [{0}]",
										animation));
			}
		}

		log.info("OM: Sending: " + commandText);
		this.output.write(commandText);

		/*
		 * NOTE: This wasn't here before, but generally has a good effect, so
		 * we'll keep it for now
		 */
		// this.output.write(0); // Null terminator for previously send string
		log.info("OM: Flushing stream..."); // 
		this.output.flush();

		return true;
	}

	private String replaceParameters(String command,
			HashMap<String, Double> parameters) {

		boolean foundParameter = true;
		while (foundParameter) {
			int paramBeginIndex = command.indexOf('{');
			int paramEndIndex = command.indexOf('}');

			if (paramBeginIndex == -1 || paramEndIndex == -1) {
				foundParameter = false;
			} else {

				String param = command.substring(paramBeginIndex + 1,
						paramEndIndex);
				Double paramValue = parameters.get(param);
				if (paramValue != null) {
					command = command.replaceAll("[{]" + param + "[}]", ""
							+ paramValue);
				}
			}
		}
		return command;

	}

	/**
	 * Wait for a DONE notification from the server.
	 * 
	 * @return True if a success notification was received.
	 * @throws Exception.
	 */
	public boolean waitSingleNotify() throws Exception {
		/******* Waiting part ********/
		log.info("OM: Attempting to read notification from server");

		if (inputStrings.isEmpty()) {
			updateFromSocket();
			if (inputStrings.isEmpty()) {
				// The reason this can't happen is that reading from the socket
				// blocks until there is something to be read ...
				log
						.severe("OM: No notifications pending on socket, while one was requested and indicated as available. This should never happen!");
			}
		}

		// Retrieve and remove the first element of the inputString vector
		String response = inputStrings.get(0).trim();
		inputStrings.remove(0);
		log.info("OM: Notification read and trimmed: '" + response + "'");
		return response.equals("DONE");
	}

	/**
	 * Retrieves whether the OutputLink has pending notifications to be read.
	 * 
	 * @return True if there are pending notifications, False otherwise.
	 */
	public boolean hasNotifications() throws Exception {
		try {
			return (this.input.ready() || !inputStrings.isEmpty());
		} catch (Exception e) {
			log.severe("Error determining notification availability: " + e);
			throw e;
		}
	}

	/**
	 * Reads all pending XML notification from the sockets and puts the in the
	 * inputStrings list of strings waiting to be parsed.
	 * 
	 * NOTE: Theoretically this approach won't work if there is too much data in
	 * the buffer (exceeding MAX_READ_LENGTH characters) In such cases the input
	 * is actually cut off and we'd need multiple reads. The chance of this
	 * occurring is very low, so it is not currently handled.
	 * 
	 * @throws Exception
	 *             On socket read error.
	 */
	private void updateFromSocket() throws Exception {
		assert this.socket != null; /*
									 * If this fails close() probably has been
									 * called
									 */

		CharBuffer buffer = CharBuffer
				.allocate(HaptekOutputLink.MAX_READ_LENGTH);
		input.read(buffer.array(), 0, HaptekOutputLink.MAX_READ_LENGTH);

		String string = "";
		while (buffer.remaining() > 0) {
			char c = buffer.get();

			if (c == '\n') {
				inputStrings.add(new String(string));
				string = "";
			} else {
				string = string + c;
			}
		}
	}

	/**
	 * Closes all connections.
	 * 
	 * Note: Calling this method after all transactions are completed is a
	 * necessity!
	 */
	public void close() throws Exception {
		log.info("Closing connection");

		if (this.output != null) {
			this.output.close();
			this.output = null;
		}
		if (this.input != null) {
			this.input.close();
			this.input = null;
		}
		if (this.socket != null) {
			this.socket.close();
			this.socket = null;
		}
	}

}