/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.om;

/*
 * TODO: It might be nicer to be able to enable/disable logging facilities
 * programmatically (or even by using the 'normal' logging facilities and
 * redirection)
 */

import java.util.ArrayList;
import java.util.HashMap;
import java.util.logging.Logger;

import nl.utwente.ewi.hmi.deira.db.PersonalityDB;
import nl.utwente.ewi.hmi.deira.generic.DeiraThread;
import nl.utwente.ewi.hmi.deira.generic.Module;
import nl.utwente.ewi.hmi.deira.queue.Event;
import nl.utwente.ewi.hmi.deira.queue.EventQueue;
import nl.utwente.ewi.hmi.deira.tgm.TextFragment;

/**
 * Bootstrapping class for Outputter part.
 */
public abstract class Outputter extends DeiraThread implements Module {
	protected PersonalityDB pdb;
	/**
	 * < The personality database of the reporter. From this, the face to use
	 * will be obtained
	 */
	protected static Logger log = Logger.getLogger("deira.om");
	/**
	 * < Logging instance
	 */
	protected EventQueue queue;
	protected boolean terminate = false;
	protected boolean isOperational = false;
	protected OutputEvent outputEvent;

	/**
	 * Creates a new Outputter.
	 * 
	 * @param pdb
	 *            PersonalityDB reference to the personality database.
	 * @param queue
	 *            The event queue to remove the events from.
	 * @param outputevent
	 *            Event that is triggered when textual output is sent.
	 */
	public Outputter(PersonalityDB pdb, EventQueue queue,
			OutputEvent outputEvent) {
		super();
		this.setName("DEIRA OM");

		this.pdb = pdb;
		this.queue = queue;
		this.outputEvent = outputEvent;
	}

	/**
	 * Output module main loop. This removes events of the queue and sends the
	 * associated relevant data to Visage over the link.
	 */
	public void run() {

		do {
			try {
				this.queue.waitForEvent();
			} catch (Exception e) {
				log.severe("OM: Error waiting for event! " + e);
			}

			Event event = this.queue.poll();
			if (event != null) {
				processEvent(event);
				outputNotification(event);
			} else {
				log.severe("OM: Woken while nothing was available!");
			}

		} while (!this.terminate);
	}

	private void outputNotification(Event event) {
		if (event.hasTag("outputtedNotify")) {
			OutputtedEvent obj = (OutputtedEvent) event
					.getTag("outputtedNotify");

			/* Trigger event for whatever requested it */
			synchronized (obj) {
				obj.OnOutputtedEvent(event);
			}
		}
	}

	protected abstract void processEvent(Event event);

	public abstract void sendDirectAnimations(ArrayList<String> animations);

	// =-- FIXME: hack (results will probably be detrimental if this method
	// actually called)
	public void sendDirectAnimations(ArrayList<String> animations,
			HashMap<String, Double> parameters) {
		this.sendDirectAnimations(animations);
	}

	/**
	 * Instructs the OM to stop waiting for events and exit.
	 * 
	 */
	public void close() {
		this.terminate = true;
	}

	/**
	 * Retrieves the text of a race event.
	 * 
	 * @param event
	 *            Event to retrieve the text from.
	 * @return Text retrieved from the event (or null if there was no text).
	 */
	protected String getTextOfEvent(Event event) {
		TextFragment fragment = (TextFragment) event.getTag("TextFragment");
		if (fragment != null) {
			String text = "";
			for (String speechPart : fragment.getText()) {
				text += speechPart + " ";
			}
			return text.trim(); /* Remove trailing whitespace */
		} else {
			return null;
		}
	}

	/**
	 * Retrieves the speech characteristics for a specific event.
	 * 
	 * @param event
	 *            Event to retrieve the speech characteristics for.
	 * @return ArrayList containing the speech characteristics (speed, pitch and
	 *         volume in that order)
	 */
	protected HashMap<String, Integer> getSpeechCharacteristicsForEvent(
			Event event) {
		HashMap<String, Integer> speechChars = null;

		Object tagInfo = event.getTag("SpeechParameters");
		speechChars = (HashMap<String, Integer>) tagInfo;

		return speechChars;
	}

	/**
	 * Retrieves a list of facial animations used for the head. NOTE: This
	 * function was actually created for post processing, but most of this stuff
	 * is in FacialAnimator itself now (where it should be). Hence we can remove
	 * this function in due time.
	 * 
	 * @param event
	 *            Event to get the facial expression for.
	 * @return List of facial expressions (as references to '.fba' files).
	 */
	protected ArrayList<String> getFacialAnimationsForEvent(Event event) {
		return (ArrayList<String>) event.getTag("expressions");
	}

	protected String generateAnimationsString(ArrayList<String> animations) {
		String strAnimations = "";
		for (String animation : animations) {
			strAnimations = strAnimations + animation + " ,";
		}
		if (strAnimations.length() > 0) {
			strAnimations = strAnimations.substring(0,
					strAnimations.length() - 2);
			// Cut off last ' ,'
		}
		return strAnimations;
	}

	public boolean isOperational() {

		return isOperational;
	}

	public String getModuleName() {
		return "OM";
	}

	/**
	 * Retrieve the output event that was set (if any).
	 * 
	 * @return The output event that was set.
	 */
	public OutputEvent getOutputEvent() {
		return outputEvent;
	}

}