/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.om;

/*
 * TODO: It might be nicer to be able to enable/disable logging facilities
 * programmatically (or even by using the 'normal' logging facilities and
 * redirection)
 */

import java.util.ArrayList;
import java.util.HashMap;

import nl.utwente.ewi.hmi.deira.db.PersonalityDB;
import nl.utwente.ewi.hmi.deira.queue.Event;
import nl.utwente.ewi.hmi.deira.queue.EventQueue;


/**
 * Bootstrapping class for Outputter part.
 */
public class VisageOutputter extends Outputter {
	VisageOutputLink outputLink;
	
	/**
	 * Creates a new Outputter.
	 * 
	 * @param pdb
	 *            PersonalityDB reference to the personality database.
	 * @param queue
	 *            The event queue to remove the events from.
	 * @param outputevent
	 *            Event that is triggered when textual output is sent.
	 */
	public VisageOutputter(PersonalityDB pdb, EventQueue queue,
			OutputEvent outputEvent) {
		super(pdb, queue, outputEvent);

		this.outputLink = null;		
		
	}

	/**
	 * Sets up the connection between this Outputter and the VisageClient.
	 * 
	 * @throws Exception
	 *             If the connection could not be established.
	 */
	public void setupConnection() throws Exception {
		try {
			log.info("Connecting to Server");
			outputLink = new VisageOutputLink("127.0.0.1", 9999);

			String basePath = System.getProperty("user.dir");
			String modelFile = "";
			String voice = "";
			if (true || pdb.getFace().equals("male")) {
				modelFile = basePath+"\\data\\model\\kenneth\\kenneth.afm"; 
			}
			
			if (true || pdb.getVoice().equals("male")) {
				voice = "ScanSoft Tom_Full_22kHz"; 
			}

			outputLink.sendSetup(modelFile, voice, basePath+"\\bin\\visagelink\\p2v.cfg", basePath+"\\data\\animations\\");
		} catch (Exception e) {
			outputLink = null;
			throw e;
		}

		isOperational = true;
	}

	@Override
	protected void processEvent(Event event) {
		if (outputLink == null) {
			try {
				setupConnection();
			} catch (Exception e) {
				log.severe("Can not establish connection with Player!");
			}
		}

		if (event != null) {

			long delayLow = Long.MAX_VALUE;
			long delayHigh = Long.MIN_VALUE;

			log.info("VisOM: Event grabbed (" + event.getId() + ")");
			String text = getTextOfEvent(event);

			if (text != null) {
				try {
					
					//=-- Retrieve animations
					ArrayList<String> animations = getFacialAnimationsForEvent(event);

					//=-- Retrieve speech characteristics
					HashMap<String, Integer> speechChars = getSpeechCharacteristicsForEvent(event);

					int speed = (int) Math.round(1.0 + (float)speechChars.get("speed")/5);
					int pitch = (int) Math.round(0.0 + speechChars.get("pitch") * 0.8); 
					int volume = (int) Math.round(5.0 + speechChars.get("volume")*0.5);
					
					//=-- Notify event outputted observer --> Generally the GUI 
					if (this.outputEvent != null) {
						
						//=-- Prepare animation string
						String strAnimations = generateAnimationsString(animations);

						//=-- Calculate delay
						long delay = System.currentTimeMillis()
								- event.getCreateSystemTime();
						if (delay < delayLow) {
							delayLow = delay;
						} else if (delay > delayHigh) {
							delayHigh = delay;
						}
						
						//=-- Call event outputted actions
						this.outputEvent.OnOutputEvent(text, speed, pitch, volume, strAnimations,
								delay, delayLow, delayHigh);

						// =-- Write text to log (enable the 'write' function
						// and the creation / clean-up of log-file to use this)
						// this.write(text);
					}
					
					
					//=-- Handle actually outputting the event.
					log.info("VisOM: Waiting for outputlink to become available");
					if (outputLink == null) { /* Something is wrong! */
						log
								.severe("VisOM: The outputlink is gone! Terminating ...");
						this.terminate = true;
						return;
					}
					synchronized (outputLink) {
						log
								.info("VisOM: Outputlink available, locking it and continuing");

						if (!outputLink.sendAndWait(text, animations,
								speed, pitch,
								volume)) {
							log.warning("Received error from VisageLink");
						}
						log.info("VisOM: Event processed (" + event.getId() + ")");

					}
					log.info("VisOM: Outputlink unlocked");

				} catch (Exception e) {
					log.severe("Exception: " + e);
					e.printStackTrace();
					this.terminate = true;
				}
			} else {
				log.severe("VisOM: Text of event is null!");
			}
		} else {
			log.severe("VisOM: Woken while nothing was available!");
		}

	}

	/**
	 * Directly sends animations through the outputlink (bypassing the output
	 * module event loop).
	 * 
	 * @param animations
	 *            Animations to send (vector of fba filenames).
	 */
	public void sendDirectAnimations(ArrayList<String> animations) {
		if (outputLink == null) {
			log.severe("VisOM(DA): The outputlink is gone!");
			return;
		}
	
		try {
			log.info("VisOM(DA): Waiting for outputlink to become available");
			synchronized (outputLink) {
				log
						.info("VisOM(DA): Outputlink available, locking it and continuing");
				if (!outputLink.sendAndWaitAnimationsOnly(animations)) {
					log
							.warning("Received error from VisageInterface sending animations only");
				} else {
					outputLink.notifyAll();
				}
			}
			log.info("VisOM(DA): Outputlink unlocked");
		} catch (java.net.SocketException e) {
			log.severe("Connection Error: " + e);
			shutdownLink();
		} catch (Exception e) {
			log.severe("Exception: " + e);
		}
	}	
	
	public void close() {

		shutdownLink();
		super.close();
	}

	/**
	 * Shuts down the link.
	 */
	public void shutdownLink() {
		log.info("Disconnecting from Server");
		try {
			outputLink.close();
		} catch (Exception e) {
			log.severe("Exception occured: " + e);
		}
		outputLink = null;
	}

}