/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.om;

/*
 * TODO: It might be nicer to be able to enable/disable logging facilities
 * programmatically (or even by using the 'normal' logging facilities and
 * redirection)
 */

import java.util.ArrayList;
import java.util.HashMap;

import javax.speech.AudioException;
import javax.speech.Central;
import javax.speech.EngineException;
import javax.speech.EngineList;
import javax.speech.EngineStateError;
import javax.speech.synthesis.Synthesizer;
import javax.speech.synthesis.SynthesizerModeDesc;
import javax.speech.synthesis.SynthesizerProperties;
import javax.speech.synthesis.Voice;

import nl.utwente.ewi.hmi.deira.db.PersonalityDB;
import nl.utwente.ewi.hmi.deira.queue.Event;
import nl.utwente.ewi.hmi.deira.queue.EventQueue;

/**
 * Bootstrapping class for Outputter part.
 */
public class VoiceOutputter extends Outputter {
	private OutputEvent outputEvent;
	private Voice voice;
	private Synthesizer synth;
	private SynthesizerProperties props;

	/**
	 * Creates a new Outputter.
	 * 
	 * @param pdb
	 *            PersonalityDB reference to the personality database.
	 * @param queue
	 *            The event queue to remove the events from.
	 * @param outputevent
	 *            Event that is triggered when textual output is sent.
	 */
	public VoiceOutputter(PersonalityDB pdb, EventQueue queue,
			OutputEvent outputEvent) {
		super(pdb, queue, outputEvent);

		initVoice();
	}

	private void initVoice() {
		String requestedVoice = pdb.getVoice();

		SynthesizerModeDesc desc = null;

		EngineList list = Central.availableSynthesizers(null);
		outerLoop: for (Object descobj : list) {
			desc = (SynthesizerModeDesc) descobj;
			Voice[] voices = desc.getVoices();
			for (Voice currentvoice : voices) {
				voice = currentvoice;
				if (currentvoice.getName().equals(requestedVoice)) {
					break outerLoop;
				}
			}
		}

		if (voice == null) {
			log.severe("Unable to find a suitable SAPI5 voice!");
		} else {

			try {
				synth = Central.createSynthesizer(desc);
				synth.allocate();
				synth.resume();
				synth.waitEngineState(Synthesizer.ALLOCATED);
			} catch (IllegalArgumentException e) {
				e.printStackTrace();
			} catch (EngineException e) {
				e.printStackTrace();
			} catch (AudioException e) {
				e.printStackTrace();
			} catch (EngineStateError e) {
				e.printStackTrace();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			System.out.println("Using voice " + voice);
			props = synth.getSynthesizerProperties();
			props.setVoice(voice);
		}
	}

	/**
	 * Retrieve the output event that was set (if any).
	 * 
	 * @return The output event that was set.
	 */
	public OutputEvent getOutputEvent() {
		return outputEvent;
	}

	@Override
	protected void processEvent(Event event) {
		if (event != null) {

			log.info("OM: Event grabbed (" + event.getId() + ")");
			String text = getTextOfEvent(event);

			if (text != null) {
				HashMap<String, Integer> speechChars = getSpeechCharacteristicsForEvent(event);

				float min_pitch = 100.0f;
				float max_pitch = 240.0f;
				float pitch = min_pitch
						+ ((float) speechChars.get("pitch") / 10)
						* (max_pitch - min_pitch);
				props.setPitch(pitch);

				float min_speed = 100;
				float max_speed = 200;
				float speed = min_speed
						+ ((float) speechChars.get("speed") / 10)
						* (max_speed - min_speed);
				props.setSpeakingRate(speed);

				float min_volume = 0.0f;
				float max_volume = 1.0f;
				float volume = min_volume
						+ ((float) speechChars.get("volume") / 10)
						* (max_volume - min_volume);
				props.setVolume(volume);

				synth.speakPlainText(text, null);
				try {
					synth.waitEngineState(Synthesizer.QUEUE_EMPTY);
				} catch (IllegalArgumentException e) {
					e.printStackTrace();
				} catch (InterruptedException e) {
					e.printStackTrace();
				}

			} else {
				log.severe("OM: Text of event is null!");
			}
		} else {
			log.severe("OM: Woken while nothing was available!");
		}

	}

	public void sendDirectAnimations(ArrayList<String> animations) {
		// Does nothing
	}

	public void close() {
		try {
			synth.waitEngineState(Synthesizer.QUEUE_EMPTY);
			synth.deallocate();
			synth.waitEngineState(Synthesizer.DEALLOCATED);
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
		} catch (InterruptedException e) {
			e.printStackTrace();
		} catch (EngineException e) {
			e.printStackTrace();
		} catch (EngineStateError e) {
			e.printStackTrace();
		}

		super.close();
	}

}