/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.om.xface;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.logging.Logger;

/**
 * Representation of FAP file in the system. This class can load and save FAP
 * files as well as mutate them.
 */
public class FAPFile {
	private static Logger log = Logger.getLogger("deira.fam"); /**< Logging instance */
	
	private InputStream stream = null; /**< Stream that is used (file handle) **/
	private BufferedReader reader = null; /**< Reader that is used (read from file handle) **/
	
	private Integer FPS; /**< Frames per second of this FAP file */
	private Integer FAPCount; /**< Number of FAP lines in this file */
	
	private ArrayList<ArrayList<Integer>> FAPMasks; /**< All FAP mask lines in the file **/
	private ArrayList<ArrayList<Float>> FAPs; /**< All actual FAPs in the file **/
	
	/* Quickly hacked accessor functions */
	
	public int getFPS() {
		return FPS;
	}
	public int getFAPCount() {
		return FAPCount;
	}
	
	/**
	 * Creates a new FAPFile from an InputStream. This automatically reads
	 * and parses the data in the stream.
	 * 
	 * @param stream Stream to read from.
	 * @throws Exception Thrown on read errors.
	 */
	public FAPFile(InputStream stream) throws Exception
	{
		this.stream = stream;
		this.reader = new BufferedReader(new InputStreamReader(this.stream));
		this.FAPMasks = new ArrayList<ArrayList<Integer>>();
		this.FAPs = new ArrayList<ArrayList<Float>>();
		
		parseHeader();
		parseLines();
	}
	
	/**
	 * Creates a new FAPFile from direct data (used internally mostly).
	 * 
	 * @param FPS FPS to maintain
	 * @param masks FAP Masks
	 * @param faps FAP values
	 */
	public FAPFile(int FPS, ArrayList<ArrayList<Integer>> masks, ArrayList<ArrayList<Float>> faps)
	{
		this.FPS = FPS;
		this.FAPCount = faps.size();
		this.FAPMasks = masks;
		this.FAPs = faps;
	}
	
	/**
	 * Parses the header of a text FAP stream.
	 * 
	 * @throws Exception On read error.
	 */
	private void parseHeader() throws Exception
	{
		log.fine("Reading FAP File Header");
		
		String headerLine = this.reader.readLine();
		String headers[] = headerLine.split(" ");
		
		if (headers.length != 4) {
			log.warning("Unexpected Header Length " + headers.length + " instead of 4");
		}
		this.FPS = Integer.parseInt(headers[2]);
		this.FAPCount = Integer.parseInt(headers[3]);
		
		log.fine("FPS: " + this.FPS + ", FAPCount: " + this.FAPCount);
	}
	
	/**
	 * Parses all the mask and fap lines of a text FAP stream.
	 * NOTE: This does not 'deal' with the expression and viseme FAPs yet
	 * (those have multiple FAP values, respectively 4 and 6).
	 * 
	 * @throws Exception On read error.
	 */
	private void parseLines() throws Exception
	{
		log.fine("Reading FAP File Lines");
		
		int line = 1;
		while (this.reader.ready()) {
			String maskLine = this.reader.readLine();
			String masks[] = maskLine.split(" ");
			
			String specLine = this.reader.readLine();
			String specs[] = specLine.split(" ");
			
			if (masks.length != 68) {
				log.warning("Unexpected Mask Length " + masks.length + " instead of 68");
			}
			
			ArrayList<Integer> parsedMasks = new ArrayList<Integer>();
			for (String m: masks) {
				int val = Integer.parseInt(m);
				parsedMasks.add(val);
			}
			
			ArrayList<Float> parsedSpecs = new ArrayList<Float>();
			for (String s: specs) {
				float val = Float.parseFloat(s);
				parsedSpecs.add(val);
			}
			parsedSpecs.remove(0); /* Remove the first element containing the frame number */

			/* Do a quick sanity check, Number of positive masks should equal to number of FAP specs */
			int count = 0;
			for (int i: parsedMasks) {
				if (i == 1) {
					count++; 
				}
			}			
			
			if (count != parsedSpecs.size()) {
				log.warning("Line " + line + " of FAP file : Mask specifies there should be " + count + " FAPs, but there are " + parsedSpecs.size());
			}
			
			/* 
			 * NOTE: This removes zero entries, While interesting (and resulting in
			 * a reduction of processing requirements further down the road). Some of the
			 * zeroes do seem to be required to get the face back to a neutral pose for
			 * whatever reason. Probably this is an XFace issue. Anyway, this code may
			 * be interesting for the future ..
			 *  
			int j = 0;
			for (int i = 0; i < parsedMasks.size(); i++) {
				Integer mask = parsedMasks.get(i);
				
				if (mask == 1) {
					Float fap = parsedSpecs.get(j);
				
					if (Math.rint(fap) == fap && fap.intValue() == 0) {
						parsedMasks.set(i, 0); // Zero the mask
						parsedSpecs.removeElementAt(j);
					} else {
						j++;
					}
				}
			}

			if (parsedSpecs.size() < count) {
				log.info("Zero-value reduced faps from " + count + " to " + parsedSpecs.size());
			}
			*/
			
			FAPMasks.add(parsedMasks);
			FAPs.add(parsedSpecs);
			
			line++;
		}
	}
	
	/**
	 * Writes the masks and values to a new FAP file.
	 * 
	 * @param destination File to write to.
	 * @throws Exception On write error.
	 */
	public void write(String destination) throws Exception
	{
		Writer out = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(destination)));
		
		out.write("1.0 DEIRA-PROCESSED " + this.FPS + " " + this.FAPCount + "\r\n");
		
		for (int i = 0; i < this.FAPMasks.size(); i++) {
			ArrayList<Integer> masks = FAPMasks.get(i);
			ArrayList<Float> specs = FAPs.get(i);
			
			for (int m: masks) {
				out.write(m + " ");
			}
			out.write("\r\n");

			out.write(String.valueOf(i) + " "); /* Frame Index number */
			for (Float s: specs) {
				/* Only write the fractional part if necessary */
				if (Math.rint(s) == s) {
					out.write(String.valueOf(s.intValue()));
				} else {
					out.write(String.valueOf(s));
				}
				out.write(" ");
			}
			out.write("\r\n");
		}
		out.flush();
		out.close();
	}
	
	/**
	 * Appends the other FAP File to the end of this one.
	 * This requires the FPS in both files to be equal.
	 * 
	 * @param other The other FAP file.
	 * @return A new FAP file with the contents of this FAP file + that of the other one.
	 */
	public FAPFile append(FAPFile other)
	{
		assert other.FPS == this.FPS; /* Need equal FPS for now */
		
		ArrayList<ArrayList<Integer>> newFAPMasks = new ArrayList<ArrayList<Integer>>();
		ArrayList<ArrayList<Float>> newFAPs = new ArrayList<ArrayList<Float>>();
		
		/* Copy FAPs from this file */
		for (int i = 0; i < this.FAPCount; i++) {
			newFAPMasks.add(new ArrayList<Integer>(this.FAPMasks.get(i)));
			newFAPs.add(new ArrayList<Float>(this.FAPs.get(i)));
		}
		
		/* Copy FAPs from the other file */
		for (int i = 0; i < other.FAPCount; i++) {
			newFAPMasks.add(new ArrayList<Integer>(other.FAPMasks.get(i)));
			newFAPs.add(new ArrayList<Float>(other.FAPs.get(i)));
		}
		
		return new FAPFile(this.FPS, newFAPMasks, newFAPs);
	}
	
	/**
	 * Merges this FAP file with another. The other fap file needs to have the same
	 * FPS and number of FAPs for this to work!
	 * 
	 * @param other FAPFile to merge this one with.
	 * @return A new FAPFile containing the contents of this one merged with the specified other.
	 */
	public FAPFile merge(FAPFile other)
	{
		assert other.FPS == this.FPS; /* Need equal FPS for now */
		assert other.FAPCount < this.FAPCount; /* Need equal number of FAPs for now */
		
		ArrayList<ArrayList<Integer>> newFAPMasks = new ArrayList<ArrayList<Integer>>();
		ArrayList<ArrayList<Float>> newFAPs = new ArrayList<ArrayList<Float>>();
		
		for (int i = 0; i < this.FAPCount; i++) {
			ArrayList<Integer> masks = this.FAPMasks.get(i);
			ArrayList<Float> specs = this.FAPs.get(i);
			
			ArrayList<Integer> newMask = null;
			ArrayList<Float> newSpecs = null;
			
			if (i < other.FAPCount) {
				ArrayList<Integer> otherMasks = other.FAPMasks.get(i);
				ArrayList<Float> otherSpecs = other.FAPs.get(i);
			
				newMask = new ArrayList<Integer>();
				newSpecs = new ArrayList<Float>();
				
				int specIndex = 0;
				int otherSpecIndex = 0;
				
				for (int j = 0; j < masks.size(); j++) {
					int maskValue = masks.get(j);
					int otherMaskValue = otherMasks.get(j);
					
					if (maskValue == 1 && otherMaskValue == 0) { /* Copy from self */
						newMask.add(1);
						newSpecs.add(specs.get(specIndex));
					} else if (maskValue == 0 && otherMaskValue == 1) { /* Copy from other */
						newMask.add(1);
						newSpecs.add(otherSpecs.get(otherSpecIndex));
					} else if (maskValue == 1 && otherMaskValue == 1) { /* Dominance */
						float val = specs.get(specIndex);
						float otherVal = otherSpecs.get(otherSpecIndex);
						
						newMask.add(1);
						if (val > otherVal) {
							newSpecs.add(val);
						} else {
							newSpecs.add(otherVal);
						}
					} else { /* Undefined */
						newMask.add(0);
					}
					
					if (maskValue == 1) {
						specIndex++;
					}
					if (otherMaskValue == 1) {
						otherSpecIndex++;
					}
				}
			} else {
				newMask = new ArrayList<Integer>(masks);
				newSpecs = new ArrayList<Float>(specs);
			}
			
			newFAPMasks.add(newMask);
			newFAPs.add(newSpecs);
		}
		
		return new FAPFile(this.FPS, newFAPMasks, newFAPs);
	}
	
	/**
	 * Scales this FAPFile to the specified duration. This is achieved by dropping
	 * FAP lines from the input file. This function can only scale down to a certain
	 * duration, it can not (yet) scale upwards.
	 * 
	 * @param targetDuration Duration to aim for in milliseconds.
	 * @return new scaled FAPFile.
	 */
	public FAPFile scale(Long targetDuration) throws Exception
	{
		 int FAPDuration = (this.FAPCount / this.FPS) * 1000;
		 
		 if (targetDuration > FAPDuration) {
			 return this; /* Can't scale upward */
		 }
		 
		 float scaleFactor = FAPDuration / targetDuration;
		
		 ArrayList<ArrayList<Integer>> newFAPMasks = new ArrayList<ArrayList<Integer>>();
		 ArrayList<ArrayList<Float>> newFAPs = new ArrayList<ArrayList<Float>>();
		 
		 int skipFactor = (int) Math.ceil(scaleFactor);
		 for (int i = 0; i < FAPs.size() - 1; i++) {
			 
			 newFAPMasks.add(new ArrayList<Integer>(FAPMasks.get(i)));
			 newFAPs.add(new ArrayList<Float>(FAPs.get(i)));
			 
			 i += skipFactor - 1; /* Skip Extra FAPs */ 
		 }
		 
		 /* Add last item */
		 newFAPMasks.add(new ArrayList<Integer>(FAPMasks.get(FAPMasks.size() - 1)));
		 newFAPs.add(new ArrayList<Float>(FAPs.get(FAPs.size() - 1)));
		 
		 return new FAPFile(this.FPS, newFAPMasks, newFAPs);
	}
}