/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.om.xface;

import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

/**
 * Representation of notifications send by the XFace server. This is basically a
 * port of Notification.cpp & Notification.h contained within the XFace
 * distribution.
 */
class Notification {
	public enum TaskStatus {
		tsUnknown, tsQueued, tsStarted, tsFinished, tsDiscarded, tsInProgress, tsInQueue, tsError, tsOMError
	};

	private String name;
	private short ownerID;
	private int taskID;
	private TaskStatus taskStatus;

	/**
	 * Creates a new Notification object based on an XML node. These nodes look
	 * like: &lt;notify name="CONNECTION_OK" ownerId="5" taskId="0"
	 * status="0"/&gt;
	 * 
	 * Each of the node attributes is accessible through the accessor functions
	 * of this class.
	 * 
	 * @param node
	 *            Node to parse notification information from.
	 */
	Notification(Node node) {
		assert node.getNodeName() == "notify"; /*
												 * Only XML communication we
												 * know of
												 */

		NamedNodeMap attributes = node.getAttributes();
		this.name = attributes.getNamedItem("name").getNodeValue();
		this.ownerID = Short.parseShort(attributes.getNamedItem("ownerId")
				.getNodeValue());
		this.taskID = Integer.parseInt(attributes.getNamedItem("taskId")
				.getNodeValue());
		this.taskStatus = taskStatusNumberToEnum(Integer.parseInt(attributes
				.getNamedItem("status").getNodeValue()));
	}

	/**
	 * Converts a tasknumber to one of the TaskNumber.* enums. TODO: Is it
	 * really necessary to do it like this? There must be a better way since the
	 * number <-> enum correspondence is 1 on 1.
	 * 
	 * @param statusNumber
	 *            Status number to convert.
	 * @return A TaskStatus.* enum constant. If the number is out of range
	 *         TaskStatus.tsOMError is returned.
	 */
	private TaskStatus taskStatusNumberToEnum(int statusNumber) {
		switch (statusNumber) {
		case 0:
			return TaskStatus.tsUnknown;
		case 1:
			return TaskStatus.tsQueued;
		case 2:
			return TaskStatus.tsStarted;
		case 3:
			return TaskStatus.tsFinished;
		case 4:
			return TaskStatus.tsDiscarded;
		case 5:
			return TaskStatus.tsInProgress;
		case 6:
			return TaskStatus.tsInQueue;
		case 7:
			return TaskStatus.tsError;
		default:
			return TaskStatus.tsOMError;
		}
	}

	/**
	 * Retrieves the name of this notification.
	 * 
	 * TODO: Consider note giving a name but converting the names to constants
	 * like Notification.CONNECTION_OK, etc. This might be worthwile since it
	 * eliminates the need for string comparisons in other places in the OM, and
	 * centralizes it to this class.
	 * 
	 * @return Name of the notification.
	 */
	public String getName() {
		return this.name;
	}

	/**
	 * Returns the ownerID (which is the 'client' number assigned by the server)
	 * of the notification. This should be the same for one OutputLink.
	 * 
	 * @return Numerical OwnerID.
	 */
	public short getOwnerID() {
		return this.ownerID;
	}

	/**
	 * Returns the number of the task the notification is associated with.
	 * 
	 * @return Task Number.
	 */
	public int getTaskID() {
		return this.taskID;
	}

	/**
	 * Retrieves the status of the task this notification is associated with.
	 * 
	 * @return TaskStatus (one of the ts* constants).
	 */
	public TaskStatus getTaskStatus() {
		return this.taskStatus;
	}

	/**
	 * Returns a string representation of the Notification. (This is nice for
	 * debugging purposes).
	 * 
	 * @return String notification.
	 */
	public String toString() {
		return "Notification { Name: '" + this.name + "', OwnerID: '"
				+ this.ownerID + "', TaskID: '" + this.taskID
				+ "', TaskStatus: '" + this.taskStatus + "' }";
	}
}