/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

/*
 * @author Based on work by Koray Balci <koraybalci@yahoo.com>
 */
package nl.utwente.ewi.hmi.deira.om.xface;

import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.logging.Logger;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Representation of tasks send to the XFace server. This is basically a port of
 * Task.cpp & Task.h contained within the XFace distribution.
 */
class Task {
	public enum TaskType {
		ttpPlay, ttpStop, ttpRewind, ttuFap, ttuWav, ttuPho, ttuAnim, ttuScript, ttlWav, ttlFap, ttlFdp
	};

	private long taskID;
	/** < Unique Identification Number of this task */
	private TaskType taskType;
	/** < Type of this task, see TaskType */
	private ArrayList<String> parameters;
	/** <! Parameters of this task */
	private static DocumentBuilderFactory documentFactory = DocumentBuilderFactory
			.newInstance();
	/** <! XML Factory for document creation */

	private static Logger log = Logger.getLogger("deira.om");

	/** < Logging instance */

	/**
	 * Creates a new Task object.
	 * 
	 * @param taskID
	 *            ID of the new task.
	 * @param taskType
	 *            Type of the task to create.
	 * @param parameters
	 *            Arguments to the new task.
	 * @throws Exception
	 *             If the the task could not be instantiated.
	 */
	Task(long taskID, TaskType taskType, String[] parameters) throws Exception {
		this.taskID = taskID;
		this.taskType = taskType;

		this.parameters = new ArrayList<String>();

		if (parameters != null) {
			for (String parameter : parameters) {
				if (!this.parameters.add(parameter)) {
					throw new Exception(
							"Could not instantiate Task: failure adding parameter '"
									+ parameter + "'");
				}
			}
		}
	}

	/**
	 * Converts a task type (see the TaskType enum) to a string representation.
	 * This is the representation as used in XFace internally.
	 * 
	 * @param taskType
	 *            TaskType to convert.
	 * @return XFace compatible String representation of the taskType.
	 */
	private String taskTypeToString(TaskType taskType) {
		switch (taskType) {
		case ttpPlay:
			return "RESUME_PLAYBACK";
		case ttpStop:
			return "STOP_PLAYBACK";
		case ttpRewind:
			return "REWIND_PLAYBACK";
		case ttuFap:
			return "UPLOAD_FAP";
		case ttuWav:
			return "UPLOAD_WAV";
		case ttuPho:
			return "UPLOAD_PHO";
		case ttuAnim:
			return "UPLOAD_ANIM";
		case ttuScript:
			return "UPLOAD_SCRIPT";
		case ttlWav:
			return "LOAD_WAV";
		case ttlFap:
			return "LOAD_FAP";
		case ttlFdp:
			return "LOAD_FDP";
		default:
			return "INTERNAL ERROR (TRACKSIDE DEIRA)";
		}
	}

	/**
	 * Converts the Task to an XML representation. This is compatible with the
	 * task representation accepted by the XFacePlayer.
	 * 
	 * @return XML representation of this task.
	 */
	public String toXML() throws Exception {
		DocumentBuilder builder = null;
		try {
			builder = Task.documentFactory.newDocumentBuilder();
		} catch (Exception e) {
			log
					.severe("Error while creating new documentbuilder for converting Task to XML");
		}
		Document document = builder.newDocument();

		/* Build static document members */
		Element root = (Element) document.createElement("task");
		document.appendChild(root);
		root.setAttribute("name", taskTypeToString(this.taskType));
		root.setAttribute("ownerId", Short.toString(TaskManager.getInstance()
				.getClientID()));
		root.setAttribute("taskId", Long.toString(this.taskID));

		/* Add all parameters */
		Iterator<String> i = this.parameters.iterator();
		while (i.hasNext()) {
			Element param = (Element) document.createElement("param");
			String value = i.next();
			param.appendChild(document.createTextNode(value));
			root.appendChild(param);
		}

		/* Prepare parsing source and result */
		Source source = new DOMSource(document);
		StringWriter writer = new StringWriter();
		Result result = new StreamResult(writer);

		Transformer transformer = TransformerFactory.newInstance()
				.newTransformer();
		transformer.transform(source, result);
		return writer.toString();
	}

	/**
	 * Converts this Task to a string representation. Note that this is only for
	 * debugging purposes.
	 * 
	 * @return String representation of this task.
	 */
	public String toString() {
		String result = "Task { Type: '" + taskTypeToString(this.taskType)
				+ "'";

		Iterator<String> i = this.parameters.iterator();
		while (i.hasNext()) {
			result += ", Param '" + i.next() + "'";
		}
		result += " }";

		return result;
	}
}