/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.om.xface;

import java.util.HashMap;
import java.util.Random;
import java.util.logging.Logger;

import nl.utwente.ewi.hmi.deira.om.VisageOutputLink;

/**
 * The TaskManagercreates and maintains task requests send to the server.
 */
class TaskManager {
	private static TaskManager instance = null;
	/** < Static instance **/
	private static Logger log = Logger.getLogger("deira.om");
	/** < Logging instance */

	private int lastTaskID;
	/** < ID of the last task */
	private VisageOutputLink outputLink;
	/** < Output link to use */
	private HashMap<Integer, Task> tasks;
	/** < Quick look-up for tasks based on ID */
	private HashMap<Integer, Notification> notifications;

	/** < Quick look-up for notifications based on ID */

	/**
	 * Retrieve a new static instance of the TaskManager.
	 * 
	 * @return TaskManager instance.
	 */
	public static TaskManager getInstance() {
		if (instance == null) {
			instance = new TaskManager();
		}
		return instance;
	}

	/**
	 * Create a new (internal) Task Manager.
	 */
	private TaskManager() {
		this.tasks = new HashMap<Integer, Task>();
		this.notifications = new HashMap<Integer, Notification>();
		this.outputLink = null;
		resetLastTaskID();
	}

	/**
	 * Resets the task ID to a new value.
	 */
	private void resetLastTaskID() {
		this.lastTaskID = new Random().nextInt(Integer.MAX_VALUE / 2);
	}

	/**
	 * Instructs the TaskManager to use the specified outputlink.
	 * 
	 * @param outputLink
	 *            OutputLink to use.
	 */
	public void setOutputLink(VisageOutputLink outputLink) {
		assert this.outputLink == null; // Sanity check, do not allow
										// overwriting of outputlink
		this.outputLink = outputLink;
	}

	/**
	 * Retrieves the internal clientID associated with this Trackside DEIRA
	 * application instance.
	 * 
	 * @return ClientID.
	 */
	public short getClientID() {
		return -1; // this.outputLink.getClientID();
	}

	/**
	 * Retrieve last notification concerning the specified task.
	 * 
	 * NOTE: This function queries the socket and reads out the first pending
	 * notification. NOTE: Blocking needs not block until a status relating to
	 * the queried task becomes available. Therefore you should always invoke
	 * this in a loop.
	 * 
	 * @param taskID
	 *            Number of the task.
	 * @param block
	 *            If true block until a new notification becomes available.
	 * @return Notification object associated with the task.
	 * @throws Exception
	 *             If the notification could not be retrieved.
	 */
	public Notification getTaskNotification(int taskID, boolean block)
			throws Exception {
		/*
		 * This will overwrite any previous taskStatus for a task ID.
		 */
		if (this.outputLink.hasNotifications() || block) {
			Notification aNotification = null; // this.outputLink.receive();
			log.info("OM: - Got a Notification:" + aNotification);
			notifications.put(new Integer(aNotification.getTaskID()),
					aNotification);
		}

		/*
		 * Retrieve the status of task that was queried.
		 */
		if (notifications.containsKey(taskID)) {
			Notification notification = notifications.get(taskID);
			return notification;
		} else {
			return null;
		}
	}

	/**
	 * Retrieve the status of a task. NOTE: Blocking needs not block until a
	 * status relating to the queried task becomes available. Therefore you
	 * should always invoke this in a loop.
	 * 
	 * @param taskID
	 *            ID to query.
	 * @param block
	 *            If true block until a new status becomes available.
	 * @return The status of the task specified.
	 * @throws Exception
	 *             If the Status could not be read.
	 */
	public nl.utwente.ewi.hmi.deira.om.xface.Notification.TaskStatus getTaskStatus(
			int taskID, boolean block) throws Exception {
		Notification notification = getTaskNotification(taskID, block);

		if (notification != null) {
			return notification.getTaskStatus();
		} else {
			return Notification.TaskStatus.tsUnknown;
		}
	}

	/**
	 * Creates a new task.
	 * 
	 * TODO: Task ID assignment might become a problem if tasks are send very
	 * quickly maybe some extra random factor should be added to minimize the
	 * chance of this occuring.
	 * 
	 * @param taskType
	 *            Type of task (see Task.TaskType).
	 * @param parameters
	 *            Parameters for the task (if any).
	 * @return The identification number of the task.
	 * @throws Exception
	 *             If the Task could not be created.
	 */
	public int newTask(Task.TaskType taskType, String[] parameters)
			throws Exception {
		int taskID = -1;
		// Retrieve a number for this task
		do {
			taskID = ++lastTaskID;
			if (taskID == Integer.MAX_VALUE - 1) {
				resetLastTaskID();
			}
		} while (tasks.containsKey(lastTaskID));

		// Create and register the task
		Task task = new Task(taskID, taskType, parameters);
		tasks.put(taskID, task);

		// Send the task over the link
		// outputLink.send(task);

		/*
		 * FIXME: The initial task status check is pretty much bullshit. It does
		 * become relevant when we get blocking to work correctly. This way we
		 * can immediately
		 */
		/*
		 * Notification.TaskStatus expectStatus; switch (taskType) { case
		 * ttpPlay: expectStatus = TaskStatus.tsStarted; break; case ttpStop:
		 * expectStatus = TaskStatus.tsFinished; break; case ttpRewind:
		 * expectStatus = TaskStatus.tsFinished; break; case ttuFap:
		 * expectStatus = TaskStatus.tsStarted; break; case ttuPho: expectStatus
		 * = TaskStatus.tsStarted; break; case ttuAnim: expectStatus =
		 * TaskStatus.tsStarted; break; case ttuScript: expectStatus =
		 * TaskStatus.tsStarted; break; case ttuWav: expectStatus =
		 * TaskStatus.tsQueued; break; default: throw new
		 * Exception("Unknown Task Created!"); }
		 * 
		 * Notification.TaskStatus initialStatus = getTaskStatus(taskID, false);
		 * 
		 * if (initialStatus != expectStatus) {
		 * log.warning("Initial Task Status does not match expectations!"); }
		 */

		/*
		 * NOTE: These are the status's to expect for each task (in the order
		 * specified). Please note that ttuWav is converted in XFace to a
		 * LOAD_WAV command somehow. At present I do not know how notifications
		 * concerning that are handled. Maybe it inherits the task number?
		 * 
		 * Specs in XFace's ApplicationBase.cpp (XFace/XFaceApp/) and
		 * ServerThread.cpp (wxFacePlayer/)
		 */
		// ttpPlay: Queued / Started / Finished
		// ttpStop: Finished
		// ttpRewind: Finished
		// ttuFap: Started / Error / Finished
		// ttuPho: Started / Error / Finished
		// ttuAnim: Started / Error / Finished
		// ttuScript: Started / Error / Finished
		// ttuWav: Queued / Error
		return taskID;
	}
}