/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.queue;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;

/**
 * An event that occurred in the race, as recognized by the RaceAnalysis module.
 * Possibly tagged with additional information by the other modules (MMM, TGM,
 * SAM, FAM).
 * 
 * version 2: added general tag system so any module can add and request tags.
 */
public abstract class Event {
	protected int id;
	/** < ID to identify this event */
	protected long createTime;
	/** < Time of creation counting from the start of the race, in milliseconds */
	protected long createSystemTime;
	/** < Time of creation, system time */
	protected long outputSystemTime;
	/** < Time of output, system time */
	protected String type;
	/** < Event type */
	protected String label;
	/** < Event label, or: subtype */
	protected float importance;
	/** < Basic importance of the event */
	protected float decay;
	/** < Importance decay per second, default 1 */

	protected HashMap<String, Object> tags = new HashMap<String, Object>();
	/** < Map for tag resolution */
	protected HashMap<String, Long> tagtimestamps = new HashMap<String, Long>();
	/** < Map for tag timestamp resolution */

	public final static float MAX_IMPORTANCE = 5.0f;
	/** < max importance */
	public final static float MIN_IMPORTANCE = 0.0f;

	/** < min importance */

	/**
	 * Constructor.
	 * 
	 * @param id
	 *            Event identifier.
	 * @param time
	 *            Event creation time counting from start of the race, in
	 *            milliseconds.
	 * @param type
	 *            Event type.
	 * @param label
	 *            Event label (or subtype).
	 * @param actors
	 *            Actors involved with the event.
	 * @param importance
	 *            Basic event importance.
	 * @param decay
	 *            Importance decay per second.
	 * @param params
	 *            Optional additional parameters.
	 */
	public Event(int id, long time, String type, String label,
			float importance, float decay) {
		this.id = id;
		this.createTime = time;
		this.createSystemTime = System.currentTimeMillis();
		this.type = type;
		this.label = label;
		this.importance = importance;
		this.decay = decay;
	}

	/**
	 * Returns the ID of this event.
	 * 
	 * @return The event identifier.
	 */
	public int getId() {
		return id;
	}

	/**
	 * Returns the time of event creation, counting from the start of the race,
	 * in milliseconds.
	 * 
	 * @return The time of creation, in milliseconds, since race start.
	 */
	public long getCreateTime() {
		return createTime;
	}

	/**
	 * Returns the time of event creation, counting from the start of the race,
	 * in milliseconds.
	 * 
	 * @return The time of creation, in milliseconds, system time.
	 */
	public long getCreateSystemTime() {
		return createSystemTime;
	}

	/**
	 * Tells the event it is being polled by the output module right now. Sets
	 * the internal output time of the event.
	 */
	public void putOut() {
		outputSystemTime = System.currentTimeMillis();
	}

	/**
	 * Returns the time of event output, in system time
	 * 
	 * @return The time of output, in milliseconds, system time.
	 */
	public long getOutputSystemTime() {
		return outputSystemTime;
	}

	/**
	 * Returns the event type.
	 * 
	 * @return The event type.
	 */
	public String getType() {
		return type;
	}

	/**
	 * Returns the event label.
	 * 
	 * @return The event label.
	 */
	public String getLabel() {
		return label;
	}

	/**
	 * Returns the current importance of this event, considering the basic
	 * importance and the importance decay.
	 * 
	 * @return Current event importance.
	 */
	public float getImportance() {
		long currentTime = System.currentTimeMillis();
		float currImportance = importance - decay
				* ((currentTime - createSystemTime) / 1000);
		if (currImportance < 0)
			currImportance = 0;
		return currImportance;
	}

	/**
	 * Lowers the basic importance of an event (never lower than 0).
	 * 
	 * @param l
	 *            the amount the basic importance (0-5) is lowered.
	 */
	public void lowerImportance(float l) {
		importance = importance - l;
		if (importance < 0)
			importance = 0;
	}

	/**
	 * Returns a String representation of this event.
	 * 
	 * @return String representation of this event.
	 */
	public String toString() {

		String s = "[" + createTime + "] " + id + ": " + type + "; " + label
				+ "; " + importance + " & " + decay + " = " + getImportance();
		s += "\n";

		Iterator<String> iter = tagtimestamps.keySet().iterator();
		while (iter.hasNext()) {
			String tag = iter.next();
			long longtime = ((Long) tagtimestamps.get(tag)).longValue();
			String time = new SimpleDateFormat("HH:mm:ss.SSS").format(new Date(
					longtime));

			s += tag + " @ " + time + "\n";
		}

		return s;
	}

	/**
	 * Tag this event with a certain object combined with an indicator for tag
	 * retrieval.
	 * 
	 * @param descr
	 *            Description to indicate the tag - used for tag retrieval.
	 * @param o
	 *            The object the event is tagged (augmented) with.
	 */
	public void tag(String descr, Object o) {
		tags.put(descr, o);
		tagtimestamps.put(descr, new Long(System.currentTimeMillis()));
	}

	public boolean hasTag(String descr) {
		return tags.containsKey(descr);
	}

	public Object getTag(String descr) {
		return tags.get(descr);
	}

	/**
	 * Tests whether two events are similar. Two events are similar if the type,
	 * the label and the actors are the same.
	 * 
	 * @param e
	 *            The race event to compare this event with.
	 * @return True if e is similar to this race event.
	 */
	public abstract boolean isSimilar(Event e);

}
