/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.queue;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;

import nl.utwente.ewi.hmi.deira.iam.riam.Participant;

/**
 * An event that occurred in the race, as recognized by the RaceAnalysis module.
 * Possibly tagged with additional information by the other modules (MMM, TGM,
 * SAM, FAM).
 * 
 * version 2: added general tag system so any module can add and request tags.
 */
public class HREvent extends Event {
	private ArrayList<Participant> actors;
	/** < Actors involved in the event */
	private ArrayList<Float> params;
	/** < Optional additional event parameters */

	public final static float PRESTART_IMPORTANCE = 5.0f;
	/** < max importance */
	public final static float PRESTART_DECAY = 0.0f;
	/** < no decay */
	public final static float START_IMPORTANCE = 4.95f;
	/** < max importance */
	public final static float START_DECAY = 0.0f;
	/** < no decay */
	public final static float BACKGROUND_IMPORTANCE = 0.05f;
	/** < low importance */
	public final static float BACKGROUND_DECAY = 0.0f;

	/** < no decay */

	/**
	 * Constructor.
	 * 
	 * @param id
	 *            Event identifier.
	 * @param time
	 *            Event creation time counting from start of the race, in
	 *            milliseconds.
	 * @param type
	 *            Event type.
	 * @param label
	 *            Event label (or subtype).
	 * @param actors
	 *            Actors involved with the event.
	 * @param importance
	 *            Basic event importance.
	 * @param decay
	 *            Importance decay per second.
	 * @param params
	 *            Optional additional parameters.
	 */
	public HREvent(int id, long time, String type, String label,
			ArrayList<Participant> actors, float importance, float decay,
			ArrayList<Float> params) {
		super(id, time, type, label, importance, decay);
		this.actors = actors;
		this.params = params;
	}

	/**
	 * Returns the actors involved with the event.
	 * 
	 * @return The actors involved.
	 */
	public ArrayList<Participant> getActors() {
		return actors;
	}

	/**
	 * Setting new actors for this event. Used by the EventQueue to update
	 * otherwise obsolete events.
	 * 
	 * @param actors
	 *            The new actors for this event.
	 */
	public void setActors(ArrayList<Participant> actors) {
		this.actors = actors;
	}

	/**
	 * Returns any additional parameters for this event.
	 * 
	 * @return Additional event parameters.
	 */
	public ArrayList<Float> getParams() {
		return params;
	}

	/**
	 * Returns a String representation of this event.
	 * 
	 * @return String representation of this event.
	 * 
	 *         TODO add params representation ?
	 */
	public String toString() {
		String as = "";
		String ps = "";
		if (actors != null)
			as = actors.toString();
		if (params != null)
			ps = params.toString();

		String s = "[" + createTime + "] " + id + ": " + type + "; " + label
				+ "; " + importance + " & " + decay + " = " + getImportance()
				+ "; " + as + "; " + ps;
		s += "\n";

		Iterator<String> iter = tagtimestamps.keySet().iterator();
		while (iter.hasNext()) {
			String tag = iter.next();
			long longtime = ((Long) tagtimestamps.get(tag)).longValue();
			String time = new SimpleDateFormat("HH:mm:ss.SSS").format(new Date(
					longtime));

			s += tag + " @ " + time + "\n";
		}

		return s;
		// private long createTime; // real time, in milliseconds
		// private ArrayList<Float> params;
	}

	/**
	 * Tests whether two events are similar. Two events are similar if the type,
	 * the label and the actors are the same.
	 * 
	 * @param e
	 *            The race event to compare this event with.
	 * @return True if e is similar to this race event.
	 */
	public boolean isSimilar(Event e) {
		if (!(e instanceof HREvent))
			return false;
		HREvent hre = (HREvent) e;
		if (!(getType().equals(hre.getType())))
			return false;
		if (!(getLabel().equals(hre.getLabel())))
			return false;
		if ((getActors() == null) && (hre.getActors() == null))
			return true;
		if (!(getActors().equals(hre.getActors())))
			return false;
		return true;
	}

}
