/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.sam;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.logging.Logger;

import nl.utwente.ewi.hmi.deira.generic.Module;
import nl.utwente.ewi.hmi.deira.mmm.MMM;
import nl.utwente.ewi.hmi.deira.queue.Event;
import nl.utwente.ewi.hmi.deira.queue.EventQueue;
import nl.utwente.ewi.hmi.deira.tgm.TextFragment;

/**
 * Class that facilitates generating Speech.
 */
public class SpeechAdaptor extends Thread implements Module {

	/** The reference to the Event Queue */
	private EventQueue queue;

	/** The reference to the Event Queue */
	private MMM mmm;

	/** The variable that dictates whether the thread runs */
	private boolean work;

	/** < Logging instance */
	private static Logger log = Logger.getLogger("deira.sam");

	/**
	 * Empty constructor for the class SpeechAdaptor.
	 */
	public SpeechAdaptor() {
	}

	/**
	 * Constructor for the class SpeechAdaptor. Initializes the links with the
	 * MMM, the EQ and the SDB. The SDB is created and filled.
	 * 
	 * @param mmm
	 *            The MMM object the SpeechAdaptor should be linked to.
	 * @param queue
	 *            The EQ object the SpeechAdaptor should be linked to.
	 */
	public SpeechAdaptor(MMM mmm, EventQueue queue) {
		super("DEIRA SAM");

		// Get link to Event Queue
		this.queue = queue;
		this.mmm = mmm;

		work = true;
	}

	/**
	 * Method to close down the SpeechAdaptor thread.
	 */
	public void close() {
		work = false;
	}

	/**
	 * Main execution of the Speech Adaptor. TextFragments are scanned for
	 * occurrences of elements that need to be replaced for proper pronunciation
	 * Currently only a very specific set of items are replaced for one specific
	 * voice. In the future (for SAPI 5.3 f.i.), this could be the place to
	 * change the intonation and such at specific points.
	 */
	public void run() {
		while (work) {
			try {
				this.queue.waitForEvent();
			} catch (Exception e) {
				log.severe("SAM: Error waiting for event! " + e);
			}

			// Get Event from EventQueue
			Event currentevent = this.queue.peek();

			if (currentevent != null) {
				log.info("SAM: Event grabbed (" + currentevent.getId() + ")");

				TextFragment textfragment = (TextFragment) currentevent
						.getTag("TextFragment");
				ArrayList<String> texts = textfragment.getText();
				int index = 0;
				for (String t : texts) {
					t = t.replace("`", "'"); // TODO Nasty hack due to syntax of
					// production rules
					t = t.replace(";", ","); // TODO Nasty hack due to syntax of
					// production rules
					t = t.replace("  ", " ");
					t = t.replace("   ", " ");
					t = t.replace("Thijs", "Thais");
					t = t.replace("Franois", " Franswaah ");
					// t = t.replace("lead", "leed");
					t = t.replace("Azure", "A zuor");
					t = t.replace("Eben", "Eaben");
					// t = t.replace("Carmine",
					// "<pron sym=\"k ah r m ay n\">Carmine</pron>");

					t = t.replace("with a burst to speed",
							", with a burst to speed");
					texts.set(index, t);
					index++;
				}
				textfragment.setText(texts);
				currentevent.tag("TextFragment", textfragment);

				HashMap<String, Integer> speechparameters = this
						.getSpeechCharacteristicsForEvent(currentevent);

				currentevent.tag("SpeechParameters", speechparameters);

				log.info("SAM: Event processed (" + currentevent.getId() + ")");
				this.queue.update(currentevent);
			} else {
				log.severe("SAM: Woken while nothing was available!");
			}

		}
	}

	/**
	 * Retrieves the speech characteristics for a specific event.
	 * 
	 * @param event
	 *            Event to retrieve the speech characteristics for.
	 */
	private HashMap<String, Integer> getSpeechCharacteristicsForEvent(
			Event event) {
		HashMap<String, Integer> result = new HashMap<String, Integer>();

		// EventEmotion emotion = (EventEmotion) event.getTag("emotion");
		double tension = this.mmm.getEmotionalState().getTension() * 10;

		int speed = (int) tension;
		int pitch = (int) tension;
		int volume = (int) tension;

		if (volume < 0)
			volume = 0;
		if (speed < 0)
			speed = 0;
		if (pitch < 0)
			pitch = 0;

		result.put("speed", speed);
		result.put("pitch", pitch);
		result.put("volume", volume);

		return result;
	}

	public String getModuleName() {
		return "SAM";
	}

}