/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.tgm;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.StringTokenizer;
import java.util.logging.Logger;

/**
 * Utility class that contains functionality to create, parse and evaluate
 * conditions. Supports a single operator out of the set {==, !=, <, >} Lefthand
 * side is the name of a DEIRA parameter. Righthand side is a set names of DEIRA
 * parameters which are used to check the condition.
 */
// TODO: Fit this class in a more general language framework
public class Condition {

	/** The lefthand side of the condition */
	private String lefthand;

	/** The operator of the condition */
	private String operator;

	/** The righthand values used for evaluation */
	private ArrayList<String> values;

	/** The definition of the equality operator */
	final static String EQUALITY = "==";

	/** The definition of the inequality operator */
	final static String INEQUALITY = "!=";

	/** The definition of the 'greater then' operator */
	final static String GREATERTHEN = ">";

	/** The definition of the 'smaller then' operator */
	final static String SMALLERTHEN = "<";

	/** < Logging instance */
	private static Logger log = Logger.getLogger("deira.tgm");

	/**
	 * Constructor for the class Condition. Assigns the parameters to the
	 * instance variables.
	 * 
	 * @param lefthand
	 *            The lefthand side of the condition.
	 * @param operator
	 *            The operator of the condition.
	 * @param values
	 *            The righthand values of the condition.
	 */
	public Condition(String lefthand, String operator, ArrayList<String> values) {

		this.lefthand = lefthand;
		this.operator = operator;
		this.values = values;
	}

	/**
	 * Constructor for the class Condition. Calls parse(parseable)
	 * 
	 * @param parseable
	 *            The string to parse.
	 */
	public Condition(String parseable) {
		this.values = new ArrayList<String>();
		this.lefthand = "";
		this.operator = "";

		this.parse(parseable);

	}

	/**
	 * Parses a string representation and assigns its processed contents to the
	 * instance variables.
	 * 
	 * @param parseable
	 *            The string to parse. Defined as [identifier] [operator]
	 *            [identifier] (,[identifier])*
	 */
	private void parse(String parseable) {
		// Determine operator
		if (parseable.indexOf(Condition.EQUALITY) != -1) {
			this.operator = Condition.EQUALITY;
		}

		else if (parseable.indexOf(Condition.INEQUALITY) != -1) {
			this.operator = Condition.INEQUALITY;
		}

		else if (parseable.indexOf(Condition.GREATERTHEN) != -1) {
			this.operator = Condition.GREATERTHEN;
		}

		else if (parseable.indexOf(Condition.SMALLERTHEN) != -1) {
			this.operator = Condition.SMALLERTHEN;
		}

		// Determine lefthand side
		this.lefthand = parseable
				.substring(0, parseable.indexOf(this.operator));

		// Determine righthand side
		String righthand = parseable.substring(parseable.indexOf(this.operator)
				+ this.operator.length());
		StringTokenizer righthandtk = new StringTokenizer(righthand, "|");

		while (righthandtk.hasMoreTokens()) {
			String currenttoken = righthandtk.nextToken();
			this.values.add(currenttoken);
		}
	}

	/**
	 * Evaluates whether the condition is true, given the supplied parameter
	 * list.
	 * 
	 * @param parameters
	 *            The parameter array that contains the values associated with
	 *            the DEIRA parameters.
	 * @return The value the condition evaluates to.
	 */
	public boolean evaluate(HashMap<String, String> parameters) {
		if (parameters == null) {
			log.severe("TGM.C.ev: no parameters provided!");
			return false;
		}
		String paramoutput = parameters.get(this.lefthand);
		if (paramoutput == null) {
			// log.info("TGM.C.ev: parameters not present");
			return false;
		}

		int paramoutputint;
		int firstvalueint;

		try {
			paramoutputint = Integer.parseInt(paramoutput);
		} catch (NumberFormatException e) {
			if (paramoutput.equals("null"))
				paramoutputint = 0;
			else
				paramoutputint = 88;
		}
		try {
			firstvalueint = Integer.parseInt(values.get(0));
		} catch (NumberFormatException e) {
			firstvalueint = 88;
		}

		// log.info("TGM.C.ev: paramoutput: "+paramoutput);

		if (this.operator.equals(Condition.EQUALITY)
				&& values.contains(paramoutput)) {
			return true;
		} else if (this.operator.equals(Condition.INEQUALITY)
				&& !values.contains(paramoutput)) {
			return true;
		} else if (this.operator.equals(Condition.GREATERTHEN)
				&& paramoutputint > firstvalueint) {
			return true;
		} else if (this.operator.equals(Condition.SMALLERTHEN)
				&& paramoutputint < firstvalueint) {
			return true;
		}

		else {
			log.finest("TGM.C.ev: Condition says no for lefthand ["
					+ this.lefthand + "]");
			return false;
		}

	}

}
