/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.tgm;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.StringTokenizer;
import java.util.logging.Logger;

/**
 * The main responsibility of a Grammar is the processing of a RaceEvent to a
 * collection of sentences (TextFragments). It does so via a set of production
 * rules.
 */

public class Grammar

{
	/** The rules that define this grammar */
	private HashMap<String, ArrayList<ProductionRule>> allrules = new HashMap<String, ArrayList<ProductionRule>>();

	private String grammarFile = "grammar.txt";

	/** < Logging instance */
	private static Logger log = Logger.getLogger("deira.tgm");

	/**
	 * Constructor for the class Grammar. Adds the default production ruleset to
	 * the grammar.
	 */
	public Grammar() {
		super();
		this.addRules();
	}

	/**
	 * Constructor for the class Grammar. Adds the specified ruleset to the
	 * grammar.
	 * 
	 * @param grammarFile
	 */
	public Grammar(String grammarFile) {
		super();
		this.grammarFile = grammarFile;
		this.addRules();
	}

	private void addRules() {
		// From file
		try {
			BufferedReader file = new BufferedReader(new FileReader(
					this.grammarFile));
			String line;
			while ((line = file.readLine()) != null) {
				ProductionRule newRule = parseLine(line);
				if (newRule != null) {
					this.add(newRule);
				}
			}
		} catch (FileNotFoundException e1) {
			log.severe("Cannot find grammar file!");
		} catch (IOException e2) {
			log.severe("Error reading grammar file!");
		}

	}

	/**
	 * Method to parse a string representation of a productionrule
	 * 
	 * @return
	 */
	private ProductionRule parseLine(String originalLine) {
		String line = originalLine.trim();

		// =-- Check comment
		if (line.length() > 0 && !line.substring(0, 2).equals("//")) {
			int semicolonindex;

			// =-- Extract rulename string
			semicolonindex = line.indexOf(';');
			String rulestring = line.substring(0, semicolonindex).trim();
			line = line.substring(line.indexOf(';') + 1);

			// =-- Extract condition string
			semicolonindex = line.indexOf(';');
			String conditionstring = line.substring(0, semicolonindex).trim();
			line = line.substring(line.indexOf(';') + 1);

			// =-- Extract emotional bounds string
			semicolonindex = line.indexOf(';');
			String boundsstring = line.substring(0, semicolonindex).trim();
			line = line.substring(line.indexOf(';') + 1);

			// =-- Extract rule righthand string
			String righthandstring = line.trim();

			String ruleName = "";
			ArrayList<Condition> conditions = new ArrayList<Condition>();
			ArrayList<EmotionalBoundPair> emotionBounds = new ArrayList<EmotionalBoundPair>();
			ArrayList<RuleToken> rightHand = new ArrayList<RuleToken>();

			try {
				// =-- Determine rulename
				ruleName = rulestring.substring(rulestring.indexOf('[') + 1,
						rulestring.indexOf(']'));
			} catch (Exception e) {
				log.info("Erroneous rulename found in: " + originalLine);
				e.printStackTrace();
			}

			// =-- Determine conditions
			try {
				conditionstring = conditionstring.substring(conditionstring
						.indexOf('(') + 1, conditionstring.indexOf(')'));
				StringTokenizer conditionTokenizer = new StringTokenizer(
						conditionstring, ",");
				while (conditionTokenizer.hasMoreTokens()) {
					String currentcondition = conditionTokenizer.nextToken()
							.trim();
					Condition newCondition = new Condition(currentcondition);
					conditions.add(newCondition);
				}
			} catch (StringIndexOutOfBoundsException e) {
			}

			// =-- Determine emotional bounds
			try {
				boundsstring = boundsstring.substring(
						boundsstring.indexOf('(') + 1, boundsstring
								.indexOf(')'));
				StringTokenizer boundsTokenizer = new StringTokenizer(
						boundsstring, ",");
				while (boundsTokenizer.hasMoreTokens()) {
					String currentbounds = boundsTokenizer.nextToken().trim();
					EmotionalBoundPair newBounds = new EmotionalBoundPair(
							currentbounds);
					emotionBounds.add(newBounds);
				}
			} catch (StringIndexOutOfBoundsException e) {
			}

			// =-- Determine rule righthand tokens
			rightHand = parseRightHand(righthandstring);

			ProductionRule newRule = new ProductionRule(ruleName, conditions,
					emotionBounds, rightHand, this);

			return newRule;
		} else {
			return null; // <--Comment
		}
	}

	private ArrayList<RuleToken> parseRightHand(String righthandstring) {
		StringTokenizer tokenizer = new StringTokenizer(righthandstring, ",");

		ArrayList<RuleToken> righthandtokens = new ArrayList<RuleToken>();

		while (tokenizer.hasMoreTokens()) {
			String toParse = tokenizer.nextToken().trim();
			String text = "";
			int type = 0;

			// Production token
			if (toParse.charAt(0) == '[') {
				type = RuleToken.PRODUCTION;

				char lastchar = toParse.charAt(toParse.length() - 1);
				if (lastchar == ']') {
					text = toParse.substring(1, toParse.length() - 1);
				}

				// =-- Leave room for Kleene star and such
				else {
					text = toParse.substring(1, toParse.length() - 2);
				}
			}

			// Parameter token
			else if (toParse.charAt(0) == '{') {
				type = RuleToken.PARAMETER;
				text = toParse.substring(1, toParse.length() - 1);
			}

			// Literal
			else if (toParse.charAt(0) == '\'') {
				type = RuleToken.LITERAL;
				text = toParse.substring(1, toParse.length() - 1);
			}

			righthandtokens.add(new RuleToken(text, type));

		}
		return righthandtokens;
	}

	/**
	 * Method to retrieve the production rules with the specified name.
	 * 
	 * @param rulename
	 *            The lefthand side the production rules must have.
	 * @return The production rules with the specified lefthand side.
	 */
	public ArrayList<ProductionRule> getProductionRules(String rulename) {
		// =-- Prepare returnvariable
		ArrayList<ProductionRule> rulesfound;

		// =-- Return all production rules with lefthand side equalling
		// currenttext
		rulesfound = (ArrayList<ProductionRule>) allrules.get(rulename);

		return rulesfound;
	}

	/**
	 * Method to add a production rule to the ruleset of this grammar. Creates a
	 * new vector of production rules for rules with a lefthand side that is not
	 * yet in the ruleset. If a rule with the lefthand side of the rule to add
	 * already exists in the ruleset, the new rule is added to the vector that
	 * rule is contained in.
	 * 
	 * @param rule
	 *            The rule to add to the ruleset.
	 */
	public void add(ProductionRule rule) {
		String rulename = rule.getLefthand();
		ArrayList<ProductionRule> rulevariations = allrules.get(rulename);
		if (rulevariations != null) {
			rulevariations.add(rule);
		} else {
			ArrayList<ProductionRule> newrulevariations = new ArrayList<ProductionRule>();
			newrulevariations.add(rule);
			allrules.put(rulename, newrulevariations);
		}
	}
}