/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.tgm;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;

import nl.utwente.ewi.hmi.deira.db.PersonalityDB;
import nl.utwente.ewi.hmi.deira.mmm.MMM;
import nl.utwente.ewi.hmi.deira.queue.Event;
import nl.utwente.ewi.hmi.deira.queue.EventQueue;
import nl.utwente.ewi.hmi.deira.queue.HREvent;

public class HRTextGenerator extends TextGenerator {

	public HRTextGenerator(PersonalityDB pdb, MMM mmm, EventQueue queue) {
		super(pdb, mmm, queue);

		this.setGrammar(null);

		String[] participants = nl.utwente.ewi.hmi.deira.db.BackgroundInformation.PARTICIPANT;
		for (String participant : participants) {

			String ruleName = participant;
			ArrayList<Condition> conditions = new ArrayList<Condition>();
			ArrayList<EmotionalBoundPair> emotionBounds = new ArrayList<EmotionalBoundPair>();
			ArrayList<RuleToken> rightHand = new ArrayList<RuleToken>();

			RuleToken nameToken = new RuleToken(participant, RuleToken.LITERAL);

			rightHand.add(nameToken);

			ProductionRule newRule = new ProductionRule(ruleName, conditions,
					emotionBounds, rightHand, this.grammar);
			grammar.add(newRule);
		}

	}

	/**
	 * Converts the data that is accessible within the event to parameters that
	 * are usable in the text generation process.
	 * 
	 * @param current
	 * @return a map of parameter => value
	 */
	protected HashMap<String, String> prepareParameters(Event p_event) {
		if (!(p_event instanceof HREvent))
			return null;
		HREvent event = (HREvent) p_event;

		// =-- Prepare outputvariable
		HashMap<String, String> parameters = new HashMap<String, String>();

		// =-- Actor name {ACTORx} and position {POSITIONx}
		ArrayList<nl.utwente.ewi.hmi.deira.iam.riam.Participant> actors = event
				.getActors();
		String[] positions = { "ERR_0th", "1st", "2nd", "3rd", "4th", "5th" };

		if (actors != null) {
			int i = 1;
			int random = ((int) Math.random() * 4) + 1;
			if (random < 1 | random > 4) {
				random = 1;
			}
			for (nl.utwente.ewi.hmi.deira.iam.riam.Participant actor : actors) {
				int actor_rank = actor.getCurrentState().getRank();
				parameters.put("ACTOR" + i, actor.getName());
				parameters.put("ACTORRANKED" + actor_rank, actor.getName());
				parameters.put("RANK" + i, positions[actor_rank]);

				// parameters.put("OT-RANK"+i, positions[4]);
				// //positions[actor.getCurrentState().getRank()+1]);
				parameters.put("SU-POSITION" + i, positions[actor_rank - 1]);
				parameters.put("SD-POSITION" + i, positions[actor_rank + 1]);
				if (i == random) {
					parameters.put("RANDOMACTOR", actor.getName());
				}
				i++;

			}
		}

		// log.info("TGM.G.pP: preparing parameters for event type:
		// "+event.getType());
		String type = event.getType();

		if (type.equals("TIME")) {
			// =-- Time {MINUTE},{DEC_SECONDS},{SECONDS},{FIFTHS}

			Calendar timeobject = Calendar.getInstance();
			timeobject.setTimeInMillis(event.getCreateTime());

			// log.info("TGM.G.pP: CreateTime="+event.getCreateTime());

			// Calculate time-parts.
			String minute = "" + timeobject.get(Calendar.MINUTE);
			int totalsecondsint = timeobject.get(Calendar.SECOND);

			String decseconds;
			String seconds;
			int decsecondsint = totalsecondsint / 10;
			// No decaseconds
			if (decsecondsint == 0) {
				decseconds = "null";
				seconds = "" + totalsecondsint;
			}
			// Teens
			else if (totalsecondsint > 10 && totalsecondsint < 20) {
				decseconds = "" + totalsecondsint;
				seconds = "null";
			}
			// Above 20
			else {
				decseconds = decsecondsint + "0";
				seconds = "" + (totalsecondsint - decsecondsint * 10);
			}

			if (minute.equals("0")) {
				minute = "null";
			}
			if (seconds.equals("0")) {
				seconds = "null";
			}
			String fifths = "" + timeobject.get(Calendar.MILLISECOND) / 200;
			if (fifths.equals("0")) {
				fifths = "null";
			}

			parameters.put("MINUTE", minute);
			parameters.put("DEC_SECONDS", decseconds);
			parameters.put("SECONDS", seconds);
			parameters.put("FIFTHS", fifths);

			// =-- Pole {POLE}
			parameters.put("POLE", event.getLabel());
		}

		else if (type.equals("POSITION")) {
			// =-- Track location {LOCATION}
			parameters.put("LOCATION", event.getLabel());
		}

		else if (type.equals("SPEED")) {
			// SPEED
			parameters.put("SPEEDTYPE", event.getLabel());
		}

		else if (type.equals("GAP")) {
			// =-- GAP
			parameters.put("GAPTYPE", event.getLabel());

			// Furlongs, yards or lengths
			float meters = Math.abs(actors.get(0).getCurrentState()
					.getPosition()
					- actors.get(1).getCurrentState().getPosition());
			float feet = 3.28f * meters;

			int yards = (int) (meters / 0.9144f);
			int furlongs = (int) (meters / 201.168f);
			int lengths = (int) (feet / 13.5f);
			int dozens = 12 * yards;
			int quarterlengths = (int) (lengths / 4);
			int halflengths = quarterlengths * 2;
			int threeqlengths = quarterlengths * 3;

			// Bounds of numbers that can be expressed
			int MAXNUMBER = 60;
			int MINNUMBER = 1;

			if (yards < MAXNUMBER && yards >= MINNUMBER) {
				parameters.put("GAP-SIZE-Y", "" + yards);
			}
			if (furlongs < MAXNUMBER && furlongs >= MINNUMBER) {
				parameters.put("GAP-SIZE-F", "" + furlongs);
			}
			if (lengths < MAXNUMBER && lengths >= MINNUMBER) {
				parameters.put("GAP-SIZE-L", "" + lengths);
			}
			if (dozens < MAXNUMBER && dozens >= MINNUMBER) {
				parameters.put("GAP-SIZE-D", "" + dozens);
			}
			if (quarterlengths < MAXNUMBER && quarterlengths >= MINNUMBER) {
				parameters.put("GAP-SIZE-1QL", "" + quarterlengths);
			}
			if (halflengths < MAXNUMBER && halflengths >= MINNUMBER) {
				parameters.put("GAP-SIZE-HL", "" + halflengths);
			}
			if (threeqlengths < MAXNUMBER && threeqlengths >= MINNUMBER) {
				parameters.put("GAP-SIZE-3QL", "" + threeqlengths);
			}
		}

		else if (type.equals("FINISH")) {
			// =-- FINISH
			int finished_as = Integer.parseInt(event.getLabel());
			parameters.put("F_RANK", positions[finished_as]);
		}

		// =-- Return parameters
		return parameters;

	}

	/**
	 * Processes an event to sentences
	 * 
	 * @param event
	 * @return the produced sentences
	 */
	@SuppressWarnings("unchecked")
	protected ArrayList<TextFragment> processEvent(Event event) {
		// =-- Prepare outputvariable
		ArrayList<TextFragment> textfragments = new ArrayList<TextFragment>();

		// =-- Determine Input for the production rules
		// Rulename (==type)
		String rulename = event.getType();

		// Parameters
		HashMap<String, String> parameters = (HashMap<String, String>) event
				.getTag("textparams");

		// =-- Process input with proper production rules.
		ArrayList<ProductionRule> applicablerules = this.grammar
				.getProductionRules(rulename);
		if (applicablerules != null) {
			for (ProductionRule currentrule : applicablerules) {
				ArrayList<TextFragment> currentfragments = currentrule
						.produce(parameters);
				if (currentfragments != null)
					textfragments.addAll(currentfragments);
			}

			// =-- Return TextFragments
			return textfragments;
		} else {
			log.severe("TGM.G.pE(): no rules applicable for " + rulename);
			return null;
		}
	}

	public String getModuleName() {
		return "HRTGM";
	}

}
