/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.tgm;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.logging.Logger;

/**
 * Class for containing production rules and working with them.
 */
public class ProductionRule {
	/** The name of this production rule */
	private String ruleName;

	/** The conditions that must be met before this production rule fires */
	private ArrayList<Condition> conditions;

	/** The conditions that must be met before this production rule fires */
	private ArrayList<EmotionalBoundPair> emotionBounds;

	/** The tokens this production rule expands to */
	private ArrayList<RuleToken> righthand = new ArrayList<RuleToken>();

	/** The grammar this production rule is a part of */
	private Grammar grammar;

	/** < Logging instance */
	private static Logger log = Logger.getLogger("deira.tgm");

	public ProductionRule(String ruleName, ArrayList<Condition> conditions,
			ArrayList<EmotionalBoundPair> emotionBounds,
			ArrayList<RuleToken> righthand, Grammar grammar) {
		super();
		this.conditions = conditions;
		this.emotionBounds = emotionBounds;
		this.grammar = grammar;
		this.righthand = righthand;
		this.ruleName = ruleName;
	}

	/**
	 * Method for the production of a TextFragment based on the tokens of the
	 * rule and the DEIRA parameters specified. Contains recursive calls.
	 * 
	 * @param parameters
	 *            The DEIRA parameters that can be used for substitution.
	 * @return The TextFragment containing the production of the rule combined
	 *         with the DEIRA parameters.
	 */
	public ArrayList<TextFragment> produce(HashMap<String, String> parameters) {
		// log.info("TGM.PR.pr: entering production for " + this.lefthand);

		// =-- Initialize return ArrayList
		ArrayList<TextFragment> produced = new ArrayList<TextFragment>();
		if (this.conditions != null) {
			// log.info("TGM.PR.pr: Evaluating conditions for " +
			// this.lefthand);
			if (!this.evaluate(parameters))
				return null;
		}
		// =-- Iterate over righthand tokens
		for (RuleToken currenttoken : this.righthand) {
			int currenttype = currenttoken.getType();
			String currenttext = currenttoken.getText();
			switch (currenttype) {
			// =-- Currenttoken is a LITERAL: append to all output texts
			case RuleToken.LITERAL:
				if (currenttext != "") {
					produced = appendLiteralToProduced(currenttext, produced);
				}
				break;

			// =-- Currenttoken is a PARAMETER
			case RuleToken.PARAMETER:
				// Get value for parameter from parameter map
				if (parameters.containsKey(currenttext)) {
					// Get productions for parameter value
					ArrayList<ProductionRule> rules = this.grammar
							.getProductionRules(parameters.get(currenttext));

					// If no production rule avaliable --> exception
					if (rules == null || rules.size() == 0) {
						// Number hack
						String value = parameters.get(currenttext);
						boolean valueisint = false;
						int intvalue = -1;
						try {
							intvalue = Integer.parseInt(value);
							valueisint = true;
						} catch (NumberFormatException nfe) {
						}

						if (valueisint) {
							int deca = intvalue / 10;
							int single = intvalue - (10 * deca);

							ArrayList<TextFragment> decaproduction;
							ArrayList<TextFragment> singleproduction;

							ArrayList<ProductionRule> decarules = this.grammar
									.getProductionRules("" + deca + "0");
							if (decarules != null && decarules.size() > 0) {
								decaproduction = decarules.get(0).produce(
										parameters);
							} else {
								log
										.severe("TGM.PR.p: No production rules found for: ["
												+ deca + "0]");
								break;
							}

							if (single != 0) {
								ArrayList<ProductionRule> singlerules = this.grammar
										.getProductionRules("" + single);
								if (singlerules != null
										&& singlerules.size() > 0) {
									singleproduction = singlerules.get(0)
											.produce(parameters);
									decaproduction.addAll(singleproduction);
								} else {
									log
											.severe("TGM.PR.p: No production rules found for: ["
													+ single + "]");
									break;
								}
							}
							produced = decaproduction;
							break;
						}

						else { 
							produced = appendLiteralToProduced(value, produced);
							break;
						}
					}
					// Fire each production rule with parameter list
					// Generate new empty production-vector
					ArrayList<TextFragment> productioncombiner = new ArrayList<TextFragment>();
					ArrayList<TextFragment> originalproduced = new ArrayList<TextFragment>(
							produced);
					for (ProductionRule currentrule : rules) {
						ArrayList<TextFragment> currentsentencelist = currentrule
								.produce(parameters);

						produced = this.appendPartialSentencesToProduced(
								currentsentencelist, produced);
						productioncombiner.addAll(produced);
						produced = originalproduced;
					}
					produced = productioncombiner;
				}
				// Parameter not provided!
				else {
					log.severe("TGM.PR.p: Parameter not provided: ["
							+ currenttext + "]");
				}
				break;

			// =-- Currenttoken is a PRODUCTION TOKEN:
			case RuleToken.PRODUCTION:
				// Find production rules for token
				ArrayList<ProductionRule> rules = this.grammar
						.getProductionRules(currenttext);

				// If no production rule avaliable --> exception
				if (rules == null || rules.size() == 0) {
					log.severe("TGM.PR.p: No production rules found for: "
							+ currenttext);
					break;
				}

				// Fire each production rule with parameter list
				// Generate new empty production-vector
				ArrayList<TextFragment> productioncombiner = new ArrayList<TextFragment>();
				ArrayList<TextFragment> originalproduced = new ArrayList<TextFragment>(
						produced);

				boolean successfulproduction = false;
				for (ProductionRule currentrule : rules) {
					ArrayList<TextFragment> currentfragmentlist = currentrule
							.produce(parameters);
					if (currentfragmentlist != null) {
						successfulproduction = true;
						produced = this.appendPartialSentencesToProduced(
								currentfragmentlist, produced);
						productioncombiner.addAll(produced);
						produced = originalproduced;
					}
				}
				if (!successfulproduction) {
					return null;
				}
				produced = productioncombiner;

				break;

			default:
				break;
			}
		}

		return produced;
	}

	/**
	 * Method to evaluate whether this rule should fire. Checks all Conditions
	 * in Condition for evaluation to true
	 * 
	 * @param parameters
	 *            The DEIRA parameters that could be necessary for evaluation.
	 * @return Whether this rule should fire.
	 */
	private boolean evaluate(HashMap<String, String> parameters) {
		boolean allconditionsmet = true;

		for (Condition currentcond : this.conditions) {
			allconditionsmet = (allconditionsmet & currentcond
					.evaluate(parameters));
		}

		return allconditionsmet;
	}

	/**
	 * Helper method to add a literal token to a vector of produced fragments.
	 * The literal is appended to every textfragment in producedvector. If
	 * producedvector is empty, the literal is put into a new textfragment,
	 * which is added to producedvector.
	 * 
	 * @param toAdd
	 *            The literal to add to producedvector.
	 * @param producedvector
	 *            The original vector of produced TextFragments.
	 * @return A new ArrayList of TextFragments with the literal appended to the
	 *         TextFragments.
	 */
	private ArrayList<TextFragment> appendLiteralToProduced(String toAdd,
			ArrayList<TextFragment> producedvector) {
		ArrayList<TextFragment> returnvector = new ArrayList<TextFragment>();

		// =-- If production vector == empty, add toAdd as new TextFragment
		if (producedvector.size() == 0) {
			TextFragment newSentence = new TextFragment(this.emotionBounds,
					toAdd);
			returnvector.add(newSentence);
		}

		// =-- Iterate over productionvector
		else
			for (TextFragment currentfragment : producedvector) {
				// =-- Add text to clone of current textfragment
				TextFragment cloned = currentfragment.copy();
				cloned.addWord(toAdd);

				// =-- Check if bounds have been narrowed and update bounds
				// where necessary
				checkAndUpdateNarrowedBounds(cloned);

				returnvector.add(cloned);
			}

		return returnvector;
	}

	private void checkAndUpdateNarrowedBounds(TextFragment cloned) {
		ArrayList<EmotionalBoundPair> clonedBounds = cloned
				.getEmotionalBounds();
		HashMap<String, EmotionalBoundPair> clonedBoundsMap = new HashMap<String, EmotionalBoundPair>();
		for (EmotionalBoundPair currentpair : clonedBounds) {
			clonedBoundsMap.put(currentpair.getEmotionName(), currentpair);
		}

		for (EmotionalBoundPair currentEmotion : emotionBounds) {
			EmotionalBoundPair clonedEmoPair = clonedBoundsMap
					.get(currentEmotion.getEmotionName());
			// =-- New restriction!
			if (clonedEmoPair == null) {
				cloned.addEmotionalBound(currentEmotion);
			} else {
				if (currentEmotion.getLowerBound() > clonedEmoPair
						.getLowerBound()) {
					clonedEmoPair.setLowerBound(currentEmotion.getLowerBound());
				}
				if (currentEmotion.getUpperBound() > clonedEmoPair
						.getUpperBound()) {
					clonedEmoPair.setUpperBound(currentEmotion.getUpperBound());
				}
			}
		}
	}

	/**
	 * Helper method to add a partial sentence to a vector of produced
	 * fragments. The partial sentence is appended to every textfragment in
	 * producedvector. If producedvector is empty, the partial sentence is put
	 * into a new textfragment, which is added to producedvector.
	 * 
	 * @param toAdd
	 *            The partial sentence to add to producedvector.
	 * @param producedvector
	 *            The original vector of produced TextFragments.
	 * @return A new ArrayList of TextFragments with the partial sentence
	 *         appended to the TextFragments.
	 */
	private ArrayList<TextFragment> appendPartialSentencesToProduced(
			ArrayList<TextFragment> toAdd,
			ArrayList<TextFragment> producedvector) {
		ArrayList<TextFragment> returnvector = new ArrayList<TextFragment>();

		// =-- If production vector == empty, add toAdd as new TextFragment
		if (producedvector.size() == 0) {
			for (TextFragment currentToAdd : toAdd) {
				returnvector.add(currentToAdd);
			}
		}

		else
			for (TextFragment currentToAdd : toAdd) {
				// =-- Iterate over productionvector
				for (TextFragment currentfragment : producedvector) {
					// =-- Add text to clone of current sentence

					TextFragment cloned = currentfragment.copy();
					cloned.addWords(currentToAdd.getText());

					// =-- Check if bounds have been narrowed and update bounds
					// where necessary
					checkAndUpdateNarrowedBounds(cloned);

					returnvector.add(cloned);

				}
			}
		return returnvector;
	}

	/**
	 * @return Returns the lefthand.
	 */
	public String getLefthand() {
		return ruleName;
	}
}