/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.tgm;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.logging.Logger;

import nl.utwente.ewi.hmi.deira.db.PersonalityDB;
import nl.utwente.ewi.hmi.deira.generic.Module;
import nl.utwente.ewi.hmi.deira.mmm.EmotionalState;
import nl.utwente.ewi.hmi.deira.mmm.MMM;
import nl.utwente.ewi.hmi.deira.queue.Event;
import nl.utwente.ewi.hmi.deira.queue.EventQueue;

/**
 * Class for generation of Text for an Event. Initiates production of sentences
 * for an event and handles selection between produced sentences.
 */
public abstract class TextGenerator extends Thread implements Module {
	/** The mental model module that is used for emotional input */
	protected MMM mmm;

	/** The grammar that is used for production of text */
	protected Grammar grammar;

	/** The event queue out of which events are grabbed */
	protected EventQueue queue;

	/** The history of selected text */
	protected TextHistory history;

	/** The personality database */
	protected PersonalityDB pdb;

	/** Parameter that specifies whether the thread should be running or not */
	private boolean work = true;

	/** < Logging instance */
	protected static Logger log = Logger.getLogger("deira.tgm");

	/** The history-file of all generated TextFragments */
	private BufferedWriter sentenceCollector;
	private boolean logallsentences = false;

	/**
	 * Constructor for the class TextGenerator. Parameters are assigned to the
	 * instance variables, except for pdb. pdb is used to retrieve a grammar to
	 * assign to the instance variable. this.history is initialized to a new
	 * TextHistory
	 * 
	 * @param pdb
	 *            The personality database containing the grammar.
	 * @param mmm
	 *            See instance variable specification.
	 * @param queue
	 *            See instance variable specification.
	 */
	public TextGenerator(nl.utwente.ewi.hmi.deira.db.PersonalityDB pdb,
			MMM mmm, EventQueue queue) {
		super("DEIRA TGM");

		// Get link to Event Queue, and register for update notifications
		this.queue = queue;

		// Get link to MMM
		this.mmm = mmm;

		// Get link to PDB
		this.pdb = pdb;

		// Initialize TextHistory
		this.history = new TextHistory();

		// Open log-file for all generated sentences
		if (this.logallsentences) {
			try {
				// Create file
				FileWriter fstream = new FileWriter(
						"all_generated_sentences.txt");
				sentenceCollector = new BufferedWriter(fstream);
				sentenceCollector.write("Initialized in constructor TGM\n");
			} catch (Exception e) {
				log.severe("Exception while opening log-file: "
						+ e.getMessage());
			}
		}
	}

	/**
	 * Method to destroy this thread. Sets this.work to false.
	 */
	public void close() {
		this.work = false;
	}

	/**
	 * Method containing the main execution loop of the class. Grabs an event
	 * from the queue, passes this to the grammar which returns a set of
	 * sentences. Out of these sentences, a single sentence is selected and
	 * tagged to the event that has been grabbed. The TextHistory is updated
	 * with the selected sentence.
	 */
	public void run() {
		while (work) {
			try {
				this.queue.waitForEvent();
			} catch (Exception e) {
				log.severe("TGM: Error waiting for event! " + e);
			}

			// Get Event from EventQueue
			Event currentevent = this.queue.peek();

			if (currentevent != null) {
				log.info("TGM: Event grabbed (" + currentevent.getId() + ")");

				// Prepare text parameters
				HashMap<String, String> params = this
						.prepareParameters(currentevent);
				currentevent.tag("textparams", params);

				// Give Event to Grammar
				ArrayList<TextFragment> textfragments = this
						.processEvent(currentevent);

				if (textfragments != null && textfragments.size() > 0) {
					// =-- Print all generated TextFragments to a logfile
					if (this.logallsentences) {
						this.printGeneratedTextFragments(textfragments);
					}

					// Select Text Fragment based on history and tension
					TextFragment textfragment = this
							.selectTextFragment(textfragments);

					// Put Text Fragment in event
					currentevent.tag("TextFragment", textfragment);

					this.history.addTextFragment(textfragment);
					log.info("TGM: Event processed (" + currentevent.getId()
							+ ")");
					queue.update(currentevent);
				} else {
					log.info("TGM: Erroneous event!! (" + currentevent.getId()
							+ ")");
				}
			} else {
				log.severe("TGM: Woken while nothing was available!");
			}
		}

		// Close the log-file
		if (this.logallsentences) {
			try {
				sentenceCollector.close();
			} catch (Exception e) {
				log.severe("Exception while closing log-file: "
						+ e.getMessage());
			}
		}
	}

	private void printGeneratedTextFragments(
			ArrayList<TextFragment> textfragments) {
		Iterator<TextFragment> it = textfragments.iterator();
		try {
			sentenceCollector
					.write("---------- new call of selectTextFragment() method in TGM ----------\n");
			while (it.hasNext()) {
				TextFragment tf = (TextFragment) it.next();
				String result = tf.toString().replaceAll("[\\[\\]]+", " ")
						.trim();
				sentenceCollector.write(result + "\n");
			}
		} catch (Exception e) {
			log.severe("Exception while printing to log-file: "
					+ e.getMessage());
		}

	}

	/**
	 * Prepares the parameters of an event for processing to text
	 * 
	 * @param currentevent
	 * @return a map of parameter => value
	 */
	protected abstract HashMap<String, String> prepareParameters(
			Event currentevent);

	/**
	 * Processes an event to sentences
	 * 
	 * @param currentevent
	 * @return the produced sentences
	 */
	protected abstract ArrayList<TextFragment> processEvent(Event currentevent);

	/**
	 * Method to choose a sentence
	 * 
	 * @param textfragments
	 *            The set of Text Fragments to choose from.
	 * @return the selected sentence
	 */
	protected TextFragment selectTextFragment(
			ArrayList<TextFragment> textfragments) {
		// =-- Check input param validity
		if (textfragments == null | textfragments.size() == 0) {
			log.severe("TGM.TG.sTF: TextFragment List is empty or null!!");
			return null;
		}

		// =-- Prepare scoring variables
		double[] scores = new double[textfragments.size()];
		double highestscore = -88888;
		int highestscoreindex = -1;

		// =-- Get emotional state of agent
		EmotionalState feeling = this.mmm.getEmotionalState();

		// =-- Iterate over textfragments
		int i = 0;
		fragmentloop: for (TextFragment currentfragment : textfragments) {

			emotionsloop: for (EmotionalBoundPair emotionalBoundPair : currentfragment
					.getEmotionalBounds()) {
				String emotionName = emotionalBoundPair.getEmotionName();
				double currentEmotionValue = feeling
						.getEmotionalLevel(emotionName);

				if (currentEmotionValue == -1) {
					continue emotionsloop;
				}
				if (currentEmotionValue < emotionalBoundPair.getLowerBound()) {
					i++;
					continue fragmentloop;
				}
				if (currentEmotionValue > emotionalBoundPair.getUpperBound()) {
					i++;
					continue fragmentloop;
				}

			}

			// Calculate history penalties
			double randomness = 3 * Math.random();
			scores[i] = 100 - this.history.calculatePenalty(currentfragment)
					- randomness;

			// Track highest score
			if (scores[i] > highestscore) {
				highestscoreindex = i;
				highestscore = scores[i];
			} else if (scores[i] == highestscore) {
				// flip a coin
				boolean heads = (Math.round(Math.random()) == 1);
				if (heads) {
					highestscoreindex = i;
					highestscore = scores[i];
				}
			}
			i++;
		}

		// =-- Select and return best fragment
		TextFragment selectedTextFragment = (TextFragment) textfragments
				.get(highestscoreindex);
		log.info("TGM.TG.sTF: Selected fragment ["
				+ selectedTextFragment.toString() + "] --> EMO: "
				+ this.mmm.getEmotionalState().getDominantEmotion());
		return selectedTextFragment;

	}

	public void setGrammar(Grammar grammar) {
		if (grammar == null) {
			// Get grammar from DB
			this.grammar = pdb.getGrammar();
		} else {
			this.grammar = grammar;
		}
	}

	public String getModuleName() {
		return "TGM";
	}

}
