/* 
 * Copyright (C) 2009 HMI Group, University of Twente.
 *
 * @author Franois L.A. Knoppel <[alias] francois [on server at] knoppel.nl>
 * @author Almer S. Tigelaar <[alias] almer [on server at] gnome.org>
 * @author Danny Oude Bos <[alias] d.oudebos [on server at] utwente.nl>
 * @author Thijs Alofs <[alias] t.alofs [on server at] alumnus.utwente.nl>
 *
 * This file is part of the DEIRA project, which is free software; you can redistribute it
 * and/or modify it under the terms of the GNU "Lesser" General Public License v3 as published by the 
 * Free Software Foundation.
 */

package nl.utwente.ewi.hmi.deira.tgm;

import java.util.ArrayList;
import java.util.logging.Logger;

/**
 * TextHistory maintains a history for TextFragments and is responsible for
 * calculating a penalty for these TextFragments, based on number of times
 * selected and time of selection.
 */

public class TextHistory {

	/** The phrases for which penalties have been calculated */
	private ArrayList<String> phrases;

	/** The calculated penalties */
	private ArrayList<Float> penalties;

	/** The last update of scores in millis */
	private long lastUpdate;

	/** The decay factor used for the penalties */
	private final float decay = 0.0002f;

	/** The amount of penalty added when a phrase is added to the history */
	private final float countpenalty = 6.0f;

	/** The phrases that should always stay at 0 penalty */
	private ArrayList<String> ignorelist;

	/** < Logging instance */
	private static Logger log = Logger.getLogger("deira.tgm");

	/**
	 * Constructor for the class TextHistory. Initializes phrases, penalties and
	 * ignorelist to new ArrayLists. The ignorelist is filled with several
	 * phrases.
	 */
	public TextHistory() {
		this.phrases = new ArrayList<String>();
		this.penalties = new ArrayList<Float>();
		this.ignorelist = new ArrayList<String>();

		ignorelist.add("Carmine");
		ignorelist.add("Eben");
		ignorelist.add("Azure");
		ignorelist.add("Topaz");
		ignorelist.add("");
		ignorelist.add("and");
		ignorelist.add("now");
		ignorelist.add("then");
		ignorelist.add("now");

	}

	/**
	 * Method to calculate a penalty for an entire TextFragment. This penalty is
	 * the total of the penalties of the phrases the TextFragment is comprised
	 * of.
	 * 
	 * @param fragment
	 *            The fragment for which a penalty is to be calculated.
	 * @return The calculated penalty.
	 */

	public float calculatePenalty(TextFragment fragment) {
		float penalty = 0;
		updateScores();
		if (fragment == null) {
			log.severe("TGM.TH.cP: null!");
			return -88;
		}

		for (String currentphrase : fragment.getText()) {
			penalty += phrasePenalty(currentphrase);
		}
		return penalty;
	}

	/**
	 * Method to retrieve a penalty for a single phrase.
	 * 
	 * @param phrase
	 *            The phrase for which a penalty is to be retrieved.
	 * @return The retrieved penalty.
	 */
	private float phrasePenalty(String phrase) {
		float penalty = 0;
		int index = phrases.indexOf(phrase);
		if (index != -1)
			penalty = penalties.get(index);
		return penalty;
	}

	/**
	 * Method to add a TextFragment to the history. updateScores() is called,
	 * and for every phrase in the TextFragment, the penalty list is updated.
	 * Phrases that are not present in the history yet are added to the
	 * phraselist. For phrases that are present in the history, the penalty is
	 * augmented with the standard score. The index of the penalty of a phrase
	 * in this.penalties is the index of the phrase in this.phrases.
	 * 
	 * @param textfragment
	 */
	public void addTextFragment(TextFragment textfragment) {
		updateScores();
		ArrayList<String> fragmentphrases = textfragment.getText();
		for (String currentphrase : fragmentphrases) {
			if (this.ignorelist.contains(currentphrase)) {
				continue;
			}
			int index = phrases.indexOf(currentphrase);
			if (index == -1) {
				// Add new phrase to history
				phrases.add(currentphrase);
				penalties.add(new Float(countpenalty));
			} else {
				float currentvalue = penalties.get(index).floatValue();
				// if (currentvalue < 0) currentvalue = 0;
				penalties.set(index, new Float(currentvalue + countpenalty));
			}
		}
	}

	/**
	 * Method to update the penalties in the history based on passed time and
	 * the decayfactor.
	 */
	private void updateScores() {
		long currenttime = System.currentTimeMillis();

		int i = 0;
		for (Float currentpenalty : this.penalties) {
			currentpenalty = currentpenalty
					- (this.decay * (currenttime - this.lastUpdate));
			if (currentpenalty < 0)
				currentpenalty = new Float(0);
			this.penalties.set(i, currentpenalty);
			i++;
		}
		this.lastUpdate = currenttime;

	}

	/**
	 * Method to output the penalties that have been calculated.
	 */
	public void printPenalties() {
		int i = 0;
		log.info("TGM.TH: Word penalties___________________________");
		for (Float penalty : penalties) {
			log.info(this.phrases.get(i) + ": " + penalty.toString());
			i++;
		}

	}

}
